; =====================================================================================================================================================================================================
; Sensormodul2/Sensormodul3 zur Erfassung von Temperaturen, Luftdruck, Luftfeuchtigkeit und Alarmen mit ATmega8
; Scott-Falk Hhn (s.huehn@soemtron.de)
; Version 2.11, letzte Bearbeitung am 28.03.2017
; =====================================================================================================================================================================================================
;
; Dieses Sensormodul kommuniziert mit bis zu 24 Temperatursensoren sowie bis zu 3 Luftfeuchtigkeitssensoren, die an bis zu 8 1-Wire-Busse angeschlossen werden knnen. Verwendbar sind folgende
; Temperatursensoren: DS1820, DS18S20, DS18B20, DS1822, DS1920. Die Luftfeuchtigkeit wird mit einem HIH-4000 ermittelt und ber einen DS2438 an den 1-Wire angeschlossen. Weiterhin wird der Luftdruck
; ber einen MPXA6115A direkt auf dem Sensormodul gemessen. Zustzlich sind noch 4 Alarmeingnge vorhanden. Ein LCD mit HD44780 Controller (4 Zeilen, 20 Zeichen) zeigt die Werte aller Sensoren und
; weitere Daten an. Mit einem Taster knnen verschiedene Anzeige-Modi ausgewhlt und einige Parameter konfiguriert werden. Die ermittelten Daten werden folgendermaen ber RS232 mit 9600 Baud im
; ASCII Format gesendet:
;
; * Sensorbelegung:
;
; 33000200#     (Beispiel: 3 Sensoren an Bus1, 3 Sensoren an Bus2 und 2 Sensoren an Bus6)
;         - Enter-Zeichen CR (13)
; --------  Anzahl der Sensoren an jedem 1-Wire Bus, von Bus1 bis Bus8; bei mehr als 9 Sensoren an einem Bus wird ein Buchstabe ausgegeben: aus 10 wird "A", aus 11 wird "B" usw.
;
; Diese Meldung wird nur einmal in der Startphase des Sensormoduls gesendet (ca. 10,5 Sekunden nach dem Einschalten).
;
; * Temperaturwerte:
;
; 2: 23.5#      (Temperatur 2 = 23.5C)
;        - Enter-Zeichen CR (13)
;   -----  Temperaturwert im Bereich -55.0 bis 125.0 (Das Minuszeichen steht direkt vor der ersten Ziffer)
;  -       Trennzeichen (Doppelpunkt)
; -        Sensornummer im Bereich 1-8, a-h (Sensor 9-16), i-p (Sensor 17-24), q-s (Temperatur vom Feuchtesensor 1-3)
;
; * Luftdruckwert:
;
; w: 1015#      (Luftdruck = 1015hPa)
;        - Enter-Zeichen CR (13)
;    ----  Luftdruckwert im Bereich 200-1150
;   -      Leerzeichen
;  -       Trennzeichen (Doppelpunkt)
; -        Sensorkennung "w"
;
; * Luftfeuchtigkeitswerte:
;
; x:   38#      (Luftfeuchtigkeit = 38% RH)
;        - Enter-Zeichen CR (13)
;   -----  Luftfeuchtigkeitswert im Bereich 0 bis 100
;  -       Trennzeichen (Doppelpunkt)
; -        Sensornummer im Bereich x-z (Sensor 1-3)
;
; Die Sensordaten werden fortlaufend in einem einstellbaren Intervall zwischen 100ms und 1000ms gesendet. Zustzlich wird noch ein fester Zeitraum von 1s fr die Messung eingefgt. Ein kompletter
; Zyklus bei einem voll ausgebauten System (24 + 1 + 3 Sensoren, 24 + 1 + 3 + 3 Werte) je nach Intervall-Einstellung zwischen 4,1s und 32s.
;
; * Alarmmeldungen:
;
; C:1#          (Alarm 3 ein)
;    - Enter-Zeichen CR (13)
;   -  Alarmwert 0 =Alarm aus (Kontakt geffnet)
;                1 =Alarm ein (Kontakt geschlossen)
;  -   Trennzeichen (Doppelpunkt)
; -    Alarmnummer, A entspricht Alarm 1, B entspricht Alarm 2 usw.
;
; Eine Alarmmeldung wird bei jeder Statusnderung eines Alarmeingangs gesendet. Auerdem wird zu Beginn eines neuen Zyklus der aktuelle Status aller 4 Alarme gesendet.
;
; =====================================================================================================================================================================================================
;
; Einstellung der Konfigurationsbits (Fuse-Bits):
;
; EESAVE                         = 0     (EEPROM beim Schreiben des Flash-Speichers nicht lschen)
; CKOPT                          = 1
; CKSEL3, CKSEL2, CKSEL1, CKSEL0 = 1111  (Quarz 3-8 MHz)
; SUT1, SUT0                     = 11    (Quarz, langsam ansteigende Betriebsspannung)
; BODLEVEL, BODEN                = 00    (Brown-Out ein, 4,0 V)
;
; High = 0xD1
; Low  = 0x3F
;
; =====================================================================================================================================================================================================
;
; ----------------------
; Belegung der I/O-Ports
; ----------------------
;
; Sensormodul2                                  Sensormodul3
; ------------                                  ------------
;
; PortB0: Eingang Alarm 1 mit Pull-up, offen= Alarm aus, Low= Alarm ein
; PortB1: Eingang Alarm 2 mit Pull-up, offen= Alarm aus, Low= Alarm ein
; PortB2: Ausgang LCD, Anschluss D4
; PortB3: Ausgang LCD, Anschluss D5 (MOSI)
; PortB4: Ausgang LCD, Anschluss D6 (MISO)
; PortB5: Ausgang LCD, Anschluss D7 (SCK)
;
; PortC0: Ein/Ausgang Sensorbus 1
; PortC1: Ein/Ausgang Sensorbus 2
; PortC2: Ein/Ausgang Sensorbus 3
; PortC3: Ein/Ausgang Sensorbus 4
; PortC4: Ausgang LCD, Anschluss RS             Eingang analog fr Luftdrucksensor
; PortC5: Ausgang LCD, Anschluss E 
;
; PortD0: Eingang Taster mit Pull-up            Ausgang LCD, Anschuss RS und Taster-Eingang (Doppelfunktion)
; PortD1: Ausgang RS-232, 9600 Baud
; PortD2: Eingang Alarm 3 mit Pull-up, offen= Alarm aus, Low= Alarm ein
; PortD3: Eingang Alarm 4 mit Pull-up, offen= Alarm aus, Low= Alarm ein
; PortD4: Ein/Ausgang Sensorbus 5
; PortD5: Ein/Ausgang Sensorbus 6
; PortD6: Ein/Ausgang Sensorbus 7
; PortD7: Ein/Ausgang Sensorbus 8
;
; ------------------
; Belegung der Timer
; ------------------
;
; Timer0: nicht genutzt
; Timer1: Interrupt alle 50ms
; Timer2: nicht genutzt
;
; -----------------------------------------
; Hardware- und Software-Version definieren
; -----------------------------------------
;
;#define sm2                     ;Software fr Sensormodul2 erzeugen
#define sm3                     ;Software fr Sensormodul3 erzeugen
#define version " v2.11"        ;Versionsnummer (6 Zeichen)
;
; -----------------------------------
; Konstanten und Registerdefinitionen
; -----------------------------------
;
;.include "m8def.inc"           ;ist beim Atmel Studio ab Version 6 nicht mehr notwendig
;
.equ    clock=  4000000         ;Controller Taktfrequenz = 4,0MHz
.equ    tim1fr= 20              ;Timer1 Frequenz = 20Hz (50ms)
.equ    baud=   9600            ;RS-232 Baudrate = 9600
.equ    cr=     13              ;ASCII-Code fr CR
.equ    keytio= 40              ;Langer Tastendruck = 2s (einstellbar in 50ms-Einheiten)
.equ    keytix= 200             ;Erweiterter Tastendruck = 10s (einstellbar in 50ms-Einheiten)
.equ    almtim= 60              ;Alarmanzeige-Zeit = 3s (einstellbar in 50ms-Einheiten)
.equ    setrtn= 1200            ;Rckkehr aus dem Setup-Men = 60s (einstellbar in 50ms-Einheiten)
;
.def    isreg=  r13             ;Zwischenspeicher fr SREG bei Interruptbearbeitung
.def    itemp1= r14             ;Zwischenspeicher 1 fr Interruptroutine
.def    itemp2= r24             ;Zwischenspeicher 2 fr Interruptroutine
.def    flags=  r25             ;Verschiedene Flags
                                ;Bit0:  Timer-Flag, wird bei jedem Timer1-Interrupt gesetzt und vom Hauptprogramm nach Bearbeitung wieder gelscht
                                ;Bit1:  nderungs-Flag, zeigt an, dass die Temperaturanzeige aktualisiert werden muss
                                ;Bit2:  Anzeige-Flag, sperrt die automatische Wechsel-Anzeige bei Benutzereingriff
                                ;Bit3:  Tastenflag 1 - vorheriger Tasterstatus, wird vom Timer1-Interrupt verwaltet, enthlt Taster-Status vom vorherigen Timer1-Interrupt
                                ;Bit4:  Tastenflag 2 - kurzer Tastendruck, wird vom Interrupt gesetzt und gelscht
                                ;Bit5:  Tastenflag 3 - kurzer Tastendruck quittiert, wird vom Hauptprogramm gesetzt und vom Interrupt wieder gelscht
                                ;Bit6:  Tastenflag 4 - langer Tastendruck, wird vom Interrupt gesetzt und gelscht
                                ;Bit7:  Tastenflag 5 - langer Tastendruck quittiert, wird vom Hauptprogramm gesetzt und vom Interrupt wieder gelscht
;
; --------------
; SRAM Variablen
; --------------
;
.dseg
.org    0x60
;
rsbuff: .byte   8               ;RS-232 Puffer fr 8 Textzeichen (ASCII)
adrtab: .byte   8               ;LCD-Adress-Tabelle fr Messwert-Positionen
alstat: .byte   4               ;Alarmstatus fr 4 Alarme, wird vollstndig vom Timer1-Interrupt gesteuert:
                                ;Bit0:  aktuell gelesener Alarmstatus
                                ;Bit1:  letzter stabiler Alarmstatus
                                ;Bit2:  Bearbeitungs-Flag (zeigt an, dass Statuswechsel schon bearbeitet wird)
                                ;Bit3:  Melde-Flag, Info fr Hauptprogramm zum sofortigen Senden des Statuswechsels, wird dann vom Hauptprogramm gelscht
                                ;Bit4:  Melde-Status, Info fr das Hauptprogramm ber den zu meldenden Status
acount: .byte   4               ;Alarmzeitzhler fr 4 Alarme, wird vom Interrupt bei Statusnderung auf den Wert von <altime> gesetzt, auf 0 gezhlt, danach wird Melde-Flag gesetzt
tsenid: .byte   24 * 8          ;8-Byte-ROM-IDs aller Temperatursensoren
tsentb: .byte   24              ;Sensor-Tabelle, enthlt fr jeden Temperatursensor die Nummer des 1-Wire-Busses (0-7), an dem er angeschlossen ist, fr nicht vorhandene Sensoren wird 0xff
                                ;eingetragen
hsenid: .byte   3 * 8           ;8-Byte-ROM-IDs aller Luftfeuchtigkeitssensoren
hsentb: .byte   3               ;Sensor-Tabelle, enthlt fr jeden Luftfeuchtigkeitssensor die Nummer des 1-Wire-Busses (0-7), an dem er angeschlossen ist, fr nicht vorhandene Sensoren wird 0xff
                                ;eingetragen
temval: .byte   24 * 5          ;Temperaturwerte mit je 5 ASCII-Zeichen
senflg: .byte   24              ;Sensor-Flag fr alle Temperatursensoren (jeweils 1 Byte)
                                ;Bit0:  letzter Temperaturwert war 85,0C
senval: .byte   7 * 5           ;Luftdruckwert sowie 3 Temperatur- und 3 Feuchtigkeitswerte mit je 5 ASCII-Zeichen
senmap: .byte   8               ;ermittelte Sensorbelegung
prbuff: .byte   8 * 2           ;Puffer fr Drucksensordaten, enthlt die letzten 8 Messwerte zur Mittelwertberechnung
htbuff: .byte   8 * 2 * 3       ;Puffer fr Temperaturdaten der Luftfeuchtigkeitssensoren, enthlt die letzten 8 Messwerte zur Mittelwertberechnung
h1buff: .byte   8 * 2 * 3       ;Puffer fr Luftfeuchtigkeitssensordaten, enthlt die letzten 8 Messwerte der Sensorbetriebsspannung fr die Mittelwertberechnung der 3 Sensoren
h2buff: .byte   8 * 2 * 3       ;Puffer fr Luftfeuchtigkeitssensordaten, enthlt die letzten 8 Messwerte der Sensorausgangsspannung fr die Mittelwertberechnung der 3 Sensoren
htmval: .byte   3               ;letzter Temperaturwert der 3 Luftfeuchtigkeitssensoren (nur H-Byte)
owbuff: .byte   8               ;1-Wire-Puffer fr empfangene 64-Bit-ROM-ID und fr den Empfang des Scratchpad
temsen: .byte   1               ;ermittelte Anzahl der Temperatursensoren (0 bis 24)
prssen: .byte   1               ;ermittelte Anzahl der Luftdrucksensoren (0 bis 1)
humsen: .byte   1               ;ermittelte Anzahl der Luftfeuchtigkeitssensoren (0 bis 3)
sencnt: .byte   1               ;Sensorzhler, zeigt den aktuell bearbeiteten Sensor an, ist whrend der Mess-Phase 0, bei Temperaturen im Bereich von 1 bis 24 (abhngig von <temsen>), bei
                                ;Luftdruckmessung 25 und bei Luftfeuchtigkeitsmessung im Bereich von 26 bis 31 (abhngig von <humsen>)
intcnt: .byte   1               ;Interruptzhler fr das Mess-Timing
modflg: .byte   1               ;Modus-Flag (wird im EEPROM gespeichert):
                                ;=0:    Sensordaten werden aus EEPROM gelesen
                                ;=1:    Sensordaten werden aus EEPROM gelesen, zustzlich wird eine Sensorsuche durchgefhrt (Update-Funktion)
                                ;>1:    Sensordaten werden durch Suche ermittelt
intval: .byte   1               ;Sende-Intervall fr Temperaturdaten in 50ms-Schritten (wird im EEPROM gespeichert), Werte-Bereich von 2 bis 20 (100ms bis 1000ms)
altime: .byte   1               ;Alarm-Zeit in 50ms-Schritten (wird im EEPROM gespeichert), Wertebereich von 2 bis 20 (100ms bis 1000ms)
lang:   .byte   1               ;Men-Sprache (wird im EEPROM gespeichert):
                                ;=0:    Englisch
                                ;>0:    Deutsch
dimode: .byte   1               ;Anzeige-Modus:
                                ;0:     Anzeige der Temperatursensoren 1-8 (Sensornummer 1-8)
                                ;1:     Anzeige der Temperatursensoren a-h (Sensornummer 9-16)
                                ;2:     Anzeige der Temperatursensoren i-p (Sensornummer 17-24)
                                ;3:     Anzeige der Luftdruck- und Feuchtigkeitssensoren (Sensornummer 25-28)
                                ;4:     Anzeige der Alarme
                                ;5:     Statusanzeige (Versionsnummer, Sensoren, Belegung, Modus)
                                ;6:     Anzeige der ROM-Codes der Sensoren 1-4
                                ;7:     Anzeige der ROM-Codes der Sensoren 5-8
                                ;...
                                ;12:    Anzeige der ROM-Codes der Sensoren 25-27
semode: .byte   1               ;Einstellungs-Modus (Setup):
                                ;0:     aus
                                ;1:     Modus: Direkt, Speicher
                                ;2:     Sende-Intervall: 100, 250, 500, 750, 1000ms
                                ;3:     Alarmzeit: 250, 500, 750, 1000ms
                                ;4:     Sprache: Englisch/Deutsch
keycnt: .byte   1               ;Tasterzhler, wird bei Tastendruck auf den Wert <keytio> gesetzt und auf 0 gezhlt, dient zur Erkennung des langen Tastendrucks
intc16: .byte   2               ;16-Bit Interruptzhler, wird bei jedem Timer1-Interrupt (alle 50ms) um 1 vermindert und bei 0 gestoppt, fr lange Warteschleifen und automatische Rckkehr vom Setup-
                                ;in den Automatik-Modus
dezval: .byte   4               ;Zwischenspeicher fr die Umrechnung von 16-Bit-Hex-Werten in 4 Dezimal-Stellen
flags2: .byte   1               ;Verschiedene Flags 2:
                                ;Bit0:  Merker fr kurzen Tastendruck, wird fr Einstellungs-Modus bentigt, um das ffnen des Tasters zu erkennen
lastal: .byte   1               ;zuletzt ausgelster oder zurckgesetzter Alarm im Bereich 0-3, wird fr die kurzzeitige Alarm-Anzeige im Automatik-Modus bentigt
tseni2: .byte   24 * 8          ;8-Byte-ROM-IDs aller Temperatursensoren, hier werden bei der Update-Funktion alle ermittelten Sensoren gespeichert und anschlieend mit <tsenid> verglichen
tsent2: .byte   24              ;Sensor-Tabelle fr Update-Funktion, enthlt fr jeden neuen Temperatursensor die Nummer des 1-Wire-Busses (0-7), an dem er angeschlossen ist; bei bereits vorhandenen
                                ;Sensoren wird 0xff eingetragen
hseni2: .byte   3 * 8           ;8-Byte-ROM-IDs aller Luftfeuchtigkeitssensoren, hier werden bei der Update-Funktion alle ermittelten Sensoren gespeichert und anschlieend mit <hsenid> verglichen
hsent2: .byte   3               ;Sensor-Tabelle fr Update-Funktion, enthlt fr jeden neuen Luftfeuchtigkeitssensor die Nummer des 1-Wire-Busses (0-7), an dem er angeschlossen ist; bei bereits
                                ;vorhandenen Sensoren wird 0xff eingetragen
temse2: .byte   1               ;ermittelte Anzahl der neuen Temperatursensoren (0 bis 24)
prsse2: .byte   1               ;ermittelte Anzahl der neuen Luftdrucksensoren (0 bis 1)
humse2: .byte   1               ;ermittelte Anzahl der neuen Luftfeuchtigkeitssensoren (0 bis 3)
extkey: .byte   1               ;Zhler fr die Erkennung eines erweiterten Tastendrucks (5s), wird fr den Start der Update-Funktion bentigt
;
; =====================================================================================================================================================================================================
; Programmbereich
; =====================================================================================================================================================================================================
;
.cseg
.org    0
;
; Interrupt-Vektoren
;
reset:  rjmp    start                                   ;Programm-Start
intr0:  reti                                            ;External Interrupt Request 0, nicht genutzt
intr1:  reti                                            ;External Interrupt Request 1, nicht genutzt
t2comp: reti                                            ;Timer/Counter2 Compare Match, nicht genutzt
t2ovf:  reti                                            ;Timer/Counter2 Overflow, nicht genutzt
t1capt: reti                                            ;Timer/Counter1 Capture Event, nicht genutzt
t1coma: rjmp    timer1                                  ;Timer/Counter1 Compare Match A
t1comb: reti                                            ;Timer/Counter1 Compare Match B, nicht genutzt
t1ovf:  reti                                            ;Timer/Counter1 Overflow, nicht genutzt
t0ovf:  reti                                            ;Timer/Counter0 Overflow, nicht genutzt
spistc: reti                                            ;SPI, STC Serial Transfer Complete, nicht genutzt
usrxc:  reti                                            ;RXC USART, Rx Complete, nicht genutzt
usudre: reti                                            ;UDRE USART Data Register Empty, nicht genutzt
ustxc:  reti                                            ;TXC USART, Tx Complete, nicht genutzt
adc_cc: reti                                            ;ADC Conversion Complete, nicht genutzt
ee_rdy: reti                                            ;EEPROM Ready, nicht genutzt
anacom: reti                                            ;Analog Comparator, nicht genutzt
twi:    reti                                            ;Two-wire Serial Interface, nicht genutzt
spmrdy: reti                                            ;Store Program Memory Ready, nicht genutzt
;
; Initialisierung SRAM
;
start:  cli                                             ;Interrupt ausschalten
        ldi     zl, low(SRAM_START)
        ldi     zh, high(SRAM_START)                    ;Anfangsadresse SRAM
        clr     r16
sramcl: st      z+, r16                                 ;Speicherzelle lschen
        cpi     zl, low(SRAM_START + SRAM_SIZE)         ;letzte Speicherzelle (Low-Adresse) erreicht?
        brne    sramcl                                  ;nein -> Schleife
        cpi     zh, high(SRAM_START + SRAM_SIZE)        ;letzte Speicherzelle (High-Adresse) erreicht?
        brne    sramcl                                  ;nein -> Schleife
;
        ldi     r16, low(RAMEND)
        ldi     r17, high(RAMEND)
        out     spl, r16
        out     sph, r17                                ;Stackpointer setzen
;
; Initialisierung I/O-Ports fr Sensormodul2
;
#if     sm2                                             ;--- Sensormodul2 ---
;
        ldi     r16, 0b00111100
        out     ddrb, r16                               ;PortB, B5-B2 Ausgang, B1-B0 Eingang
        ldi     r16, 0b00000011
        out     portb, r16                              ;PortB, B1-B0 Pull-up einschalten
        ldi     r16, 0b00111111
        out     ddrc, r16                               ;PortC, C5-C4 Ausgang, C3-C0 Ausgang (wird spter auf Eingang gesetzt)
        clr     r16
        out     portc, r16                              ;PortC, kein Pull-up, Ausgnge auf Low
        ldi     r16, 0b11110010
        out     ddrd, r16                               ;PortD, D1 Ausgang, D7-D4 Ausgang (wird spter auf Eingang gesetzt)
        ldi     r16, 0b00001111
        out     portd, r16                              ;PortD, Pull-up fr Alarm-Eingnge und Taster
#endif
;
; Initialisierung I/O-Ports und A/D-Wandler fr Sensormodul3
;
#if     sm3                                             ;--- Sensormodul3 ---
;
        ldi     r16, 0b00111100
        out     ddrb, r16                               ;PortB, B5-B2 Ausgang, B1-B0 Eingang
        ldi     r16, 0b00000011
        out     portb, r16                              ;PortB, B1-B0 Pull-up einschalten
        ldi     r16, 0b00101111
        out     ddrc, r16                               ;PortC, C5 Ausgang, C3-C0 Ausgang (wird spter auf Eingang gesetzt)
        clr     r16
        out     portc, r16                              ;PortC, kein Pull-up, Ausgnge auf Low
        ldi     r16, 0b11110011
        out     ddrd, r16                               ;PortD, D0-D1 Ausgang, D7-D4 Ausgang (wird spter auf Eingang gesetzt)
        ldi     r16, 0b00001110
        out     portd, r16                              ;PortD, Pull-up fr Alarm-Eingnge
;
        ldi     r16, 0b0100 << mux0
        out     admux, r16                              ;PortC4 an ADC schalten, Referenz an VREF (5V)
        ldi     r16, (1 << aden) | (1 << adsc) | (1 << adfr) | (0b110 << adps0)
        out     adcsra, r16                             ;ADC einschalten, Conversion starten, Free Running Mode, Vorteiler=64
#endif
;
        ldi     r16, low((clock / (16 * baud)) - 1)
        ldi     r17, high((clock / (16 * baud)) - 1)
        out     ubrrh, r17
        out     ubrrl, r16                              ;USART Baudratenregister auf 9600 Baud setzen
        ldi     r16, (1 << ursel) | (1 << ucsz1) | (1 << ucsz0)
        out     ucsrc, r16                              ;8 Datenbits, 1 Stoppbit
        ldi     r16, 1 << txen
        out     ucsrb, r16                              ;Sender aktivieren
;
        ldi     r16, low((clock / (8 * 20)) - 1)
        ldi     r17, high((clock / (8 * 20)) - 1)
        out     ocr1ah, r17
        out     ocr1al, r16                             ;Timer1, Output Compare A fr 20Hz (50ms) setzen
        ldi     r16, (1 << wgm12) | (1 << cs11)
        out     tccr1b, r16                             ;Timer1, CTC-Modus, Vorteiler=8
        ldi     r16, 1 << ocie1a
        out     timsk, r16                              ;OC1A-Interrupt aktivieren
;
        clr     flags                                   ;alle Flags lschen
        rcall   w050ms                                  ;50 ms warten
;
; Temperaturwerte initialisieren
;
        ldi     zl, low(temval)
        ldi     zh, high(temval)                        ;Zeiger auf Temperaturwerte
        ldi     r16, ' '                                ;Leerzeichen
        ldi     r17, 24 * 5                             ;24 Werte mit je 5 Zeichen
cltval: st      z+, r16                                 ;Leerzeichen schreiben
        dec     r17                                     ;alle Werte bearbeitet?
        brne    cltval                                  ;nein -> Schleife
;
; Luftdruck- und Feuchtigkeitswerte initialisieren
;
        ldi     zl, low(senval)
        ldi     zh, high(senval)                        ;Zeiger auf Druck- und Feuchtigkeitswerte
        ldi     r16, ' '                                ;Leerzeichen
        ldi     r17, 7 * 5                              ;7 Werte mit je 5 Zeichen
clsval: st      z+, r16                                 ;Leerzeichen schreiben
        dec     r17                                     ;alle Werte bearbeitet?
        brne    clsval                                  ;nein -> Schleife
;
; Variablen aus EEPROM auslesen
;
        ldi     r16, low(emodfl)
        ldi     r17, high(emodfl)                       ;EEPROM-Adresse Modus-Flag
        rcall   eeread                                  ;EEPROM lesen
        sts     modflg, r18                             ;Modus-Flag speichern
;
        ldi     r16, low(eintva)
        ldi     r17, high(eintva)                       ;EEPROM-Adresse Sende-Intervall
        rcall   eeread                                  ;EEPROM lesen
        cpi     r18, 21                                 ;Sende-Intervall>20?
        brcs    reava1                                  ;nein -> weiter
        ldi     r18, 20                                 ;sonst auf 20 setzen
reava1: sts     intval, r18                             ;Sende-Intervall speichern
;
        ldi     r16, low(ealtim)
        ldi     r17, high(ealtim)                       ;EEPROM-Adresse Alarm-Zeit
        rcall   eeread                                  ;EEPROM lesen
        cpi     r18, 21                                 ;Alarm-Zeit>20?
        brcs    reava2                                  ;nein -> weiter
        ldi     r18, 10                                 ;sonst auf 10 setzen
reava2: sts     altime, r18                             ;Alarm-Zeit speichern
;
        ldi     r16, low(elang)
        ldi     r17, high(elang)                        ;EEPROM-Adresse Men-Sprache
        rcall   eeread                                  ;EEPROM lesen
        sts     lang, r18                               ;Men-Sprache speichern
;
; LCD Adress-Tabelle in RAM kopieren
;
        ldi     zl, low(2 * lcdtab)
        ldi     zh, high(2 * lcdtab)                    ;Zeiger auf LCD-Adress-Tabelle im Flash
        ldi     xl, low(adrtab)
        ldi     xh, high(adrtab)                        ;Zeiger auf LCD-Adress-Tabelle im RAM
        ldi     r17, 8                                  ;8 Werte kopieren
copadr: lpm     r16, z+                                 ;Wert aus Flash holen
        st      x+, r16                                 ;in RAM schreiben
        dec     r17                                     ;alle Zeichen kopiert?
        brne    copadr                                  ;nein -> Schleife
;
; LCD initialisieren und Sonderzeichen programmieren
;
        clr     r17                                     ;LCD in Befehlsmodus setzen
        ldi     r18, 0x30
        rcall   lcd8bt                                  ;erste Initialisierung im 8-Bit Mode ausgeben
        rcall   w004ms                                  ;Warteschleife 4,1ms
        ldi     r18, 0x30
        rcall   lcd8bt                                  ;zweite Initialisierung im 8-Bit Mode ausgeben
        rcall   w100us                                  ;Warteschleife 100s
        ldi     r18, 0x30
        rcall   lcd8bt                                  ;dritte Initialisierung im 8-Bit Mode ausgeben
        rcall   w040us                                  ;Warteschleife 40s
        ldi     r18, 0x20                               ;4-Bit Mode einstellen
        rcall   lcd8bt                                  ;und nochmals im 8-Bit Mode ausgeben
        rcall   w040us                                  ;Warteschleife 40s
        ldi     r18, 0x28                               ;4-Bit Mode, 2-Line, 5x7 Font
        rcall   lcdw40                                  ;ausgeben
        ldi     r18, 0x0c                               ;Display einschalten, keinen Cursor anzeigen
        rcall   lcdw40                                  ;ausgeben
        ldi     r18, 0x06                               ;Increment, kein Scolling
        rcall   lcdw40                                  ;ausgeben
;
        ldi     r18, 0x40                               ;LCD-Adresse der Sonderzeichen laden
        clr     r17                                     ;LCD in Befehlsmodus setzen
        rcall   lcdw40                                  ;Adresse setzen
        ldi     r19, 5 * 8                              ;jeweils 8 Bytes fr 5 Sonderzeichen ausgeben
        ldi     r17, 1                                  ;LCD in Datenmodus setzen
        ldi     zl, low(2 * xchars)
        ldi     zh, high(2 * xchars)                    ;Zeiger auf Bitmaps fr Sonderzeichen (Pfeile)
lcdini: lpm     r18, z+                                 ;Datenbyte aus Flash holen
        rcall   lcdw40                                  ;an LCD ausgeben
        dec     r19                                     ;alle Zeilen ausgegeben?
        brne    lcdini                                  ;nein -> Schleife
;
; 10s Wartezeit anzeigen, anschlieend werden alle Sensorbusse aktiviert (Reset aller Sensoren)
;
sreset: ldi     r18, 0x01                               ;Kommando "Clear LCD"
        clr     r17                                     ;LCD in Befehlsmodus setzen
        rcall   lcdw40                                  ;LCD lschen
        rcall   w016ms                                  ;16ms warten
;
        lds     r18, adrtab + 1                         ;LCD-Adresse von Zeilenanfang 2
        clr     r17                                     ;LCD in Befehlsmodus setzen
        rcall   lcdw40                                  ;ausgeben
        ldi     r19, 17                                 ;Text 17 (Sensor-Reset)
        rcall   lcdtxt                                  ;ausgeben
        lds     r18, adrtab + 2                         ;LCD-Adresse von Zeilenanfang 3
        clr     r17                                     ;LCD in Befehlsmodus setzen
        rcall   lcdw40                                  ;ausgeben
;
        ldi     r20, 20                                 ;20 Minuszeichen ausgeben (Fortschritt anzeigen)
sres10: ldi     r18, '-'                                ;Minus-Zeichen laden
        ldi     r17, 1                                  ;LCD in Datenmodus setzen
        rcall   lcdw40                                  ;ausgeben
        ldi     r19, 10                                 ;10 * 50ms = 500ms warten
sres20: rcall   w050ms                                  ;50ms warten
        dec     r19                                     ;alle Wartezyklen abgearbeitet?
        brne    sres20                                  ;nein -> Schleife
        dec     r20                                     ;alle Minuszeichen ausgegeben?
        brne    sres10                                  ;nein -> Schleife
;
        in      r16, ddrc                               ;PortC Richtungsregister lesen
        andi    r16, 0xf0                               ;PortC3 - C0 auf Eingang umschalten
        out     ddrc, r16                               ;PortC Richtungsregister neu setzen
        in      r16, ddrd                               ;PortD Richtungsregister lesen
        andi    r16, 0x0f                               ;PortD7 - D4 auf Eingang umschalten
        out     ddrd, r16                               ;PortD Richtungsregister neu setzen
        rcall   w050ms                                  ;50ms warten
;
; Temperatur- und Luftfeuchtigkeitssensortabelle lschen
;
initab: ldi     xl, low(tsentb)
        ldi     xh, high(tsentb)                        ;Zeiger auf Sensortabelle setzen
        ldi     yl, low(tsent2)
        ldi     yh, high(tsent2)                        ;Zeiger auf Sensortabelle setzen (Update-Funktion)
        ldi     r16, 0xff                               ;Startwert 0xff voreinstellen
init10: st      x+, r16                                 ;in Puffer schreiben
        st      y+, r16                                 ;in Puffer schreiben
        cpi     xl, low(tsentb + 24)                    ;Tabellenende erreicht?
        brne    init10                                  ;nein -> Schleife
;
        ldi     xl, low(hsentb)
        ldi     xh, high(hsentb)                        ;Zeiger auf Sensortabelle setzen
        ldi     yl, low(hsent2)
        ldi     yh, high(hsent2)                        ;Zeiger auf Sensortabelle setzen (Update-Funktion)
        ldi     r16, 0xff                               ;Startwert 0xff voreinstellen
init20: st      x+, r16                                 ;in Puffer schreiben
        st      y+, r16                                 ;in Puffer schreiben
        cpi     xl, low(hsentb + 3)                     ;Tabellenende erreicht?
        brne    init20                                  ;nein -> Schleife
;
        lds     r16, modflg                             ;Modus-Flag holen
        cpi     r16, 2                                  ;Sensordaten aus EEPROM laden?
        brcc    search                                  ;nein -> Sensor-Suche starten
;
; Temperatursensordaten aus EEPROM auslesen
;
        ldi     r19, 24 * 8                             ;Daten von 24 Sensoren lesen, jeweils 8 Bytes
        ldi     r16, low(etseid)
        ldi     r17, high(etseid)                       ;EEPROM-Adresse der Sensor-IDs
        ldi     xl, low(tsenid)
        ldi     xh, high(tsenid)                        ;Zeiger auf Sensor-IDs im RAM
lotid1: rcall   eeread                                  ;EEPROM lesen
        st      x+, r18                                 ;gelesenes Byte speichern
        inc     r16                                     ;nchste EEPROM-Adresse Low
        brne    lotid2                                  ;bertrag? nein -> weiter
        inc     r17                                     ;nchste EEPROM-Adresse High
lotid2: dec     r19                                     ;alle Sensordaten gelesen?
        brne    lotid1                                  ;nein -> Schleife
;
        ldi     r19, 24                                 ;Daten von 24 Sensoren lesen
        ldi     r16, low(etsetb)
        ldi     r17, high(etsetb)                       ;EEPROM-Adresse der Sensor-Tabelle
        ldi     xl, low(tsentb)
        ldi     xh, high(tsentb)                        ;Zeiger auf Sensor-Tabelle im RAM
lottb1: rcall   eeread                                  ;EEPROM lesen
        st      x+, r18                                 ;gelesenes Byte speichern
        inc     r16                                     ;nchste EEPROM-Adresse Low
        brne    lottb2                                  ;bertrag? nein -> weiter
        inc     r17                                     ;nchste EEPROM-Adresse High
lottb2: dec     r19                                     ;alle Sensoren gelesen?
        brne    lottb1                                  ;nein -> Schleife
;
; Luftfeuchtigkeitssensordaten aus EEPROM auslesen
;
        ldi     r19, 3 * 8                              ;Daten von 3 Sensoren lesen, jweils 8 Bytes
        ldi     r16, low(ehseid)
        ldi     r17, high(ehseid)                       ;EEPROM-Adresse der Sensor-IDs
        ldi     xl, low(hsenid)
        ldi     xh, high(hsenid)                        ;Zeiger auf Sensor-IDs im RAM
lohid1: rcall   eeread                                  ;EEPROM lesen
        st      x+, r18                                 ;gelesenes Byte speichern
        inc     r16                                     ;nchste EEPROM-Adresse Low
        brne    lohid2                                  ;bertrag? nein -> weiter
        inc     r17                                     ;nchste EEPROM-Adresse High
lohid2: dec     r19                                     ;alle Sensoren gelesen?
        brne    lohid1                                  ;nein -> Schleife
;
        ldi     r19, 3                                  ;Daten von 3 Sensoren lesen
        ldi     r16, low(ehsetb)
        ldi     r17, high(ehsetb)                       ;EEPROM-Adresse der Sensor-Tabelle
        ldi     xl, low(hsentb)
        ldi     xh, high(hsentb)                        ;Zeiger auf Sensor-Tabelle im RAM
lohtb1: rcall   eeread                                  ;EEPROM lesen
        st      x+, r18                                 ;gelesenes Byte speichern
        inc     r16                                     ;nchste EEPROM-Adresse Low
        brne    lohtb2                                  ;bertrag? nein -> weiter
        inc     r17                                     ;nchste EEPROM-Adresse High
lohtb2: dec     r19                                     ;alle Sensoren gelesen?
        brne    lohtb1                                  ;nein -> Schleife
;
; Anzahl der Luftdrucksensoren aus EPROM auslesen
;
        ldi     r16, low(eprsen)
        ldi     r17, high(eprsen)                       ;EEPROM-Adresse Anzahl der Luftdrucksensoren
        rcall   eeread                                  ;EEPROM lesen
        andi    r18, 0x01                               ;Wertebereich begrenzen (0 oder 1)
        sts     prssen, r18                             ;Anzahl der Luftdrucksensoren speichern
;
; Beim Update-Modus zustzlich eine Sensorsuche starten
;
        lds     r16, modflg                             ;Modus-Flag holen
        cpi     r16, 1                                  ;Update-Modus?
        breq    search                                  ;ja -> Suche starten
        rjmp    inmain                                  ;sonst Ende
;
; Drucksensor suchen - wenn der Messwert bei offenem ADC-Eingang <0,5V ist oder bei aktiviertem Pull-up-Widerstand >4,6V ist, dann ist kein Sensor vorhanden (wird nur beim Sensormodul3 ausgefhrt)
;
search:
#if     sm3                                             ;--- Sensormodul3 ---
;
        rcall   w004ms                                  ;4ms warten bis der ADC-Wert stabil ist
        in      r16, adcl
        in      r17, adch                               ;ADC-Wert lesen
        ldi     r18, low(102)
        ldi     r19, high(102)                          ;Vergleichswert laden (entspricht ungefhr 0,5V)
        cp      r16, r18
        cpc     r17, r19                                ;ist der ADC-Wert zu niedrig (<0,5V)?
        brcs    seasen                                  ;ja -> kein Drucksensor, weiter zur 1-Wire Sensorsuche
;
        sbi     portc, 4                                ;sonst Pull-up am Sensoreingang aktivieren
        rcall   w004ms                                  ;4ms warten bis der ADC-Wert stabil ist
        in      r16, adcl
        in      r17, adch                               ;ADC-Wert lesen
        cbi     portc, 4                                ;Pull-up am Sensoreingang wieder ausschalten
        ldi     r18, low(942)
        ldi     r19, high(942)                          ;Vergleichswert laden (entspricht ungefhr 4,6V)
        cp      r16, r18
        cpc     r17, r19                                ;ist der ADC-Wert zu hoch (>4,6V)?
        brcc    seasen                                  ;ja -> weiter zur 1-Wire Sensorsuche
;
        ldi     r17, 1                                  ;Anzahl der Drucksensoren
        lds     r16, modflg                             ;Modus-Flag holen
        cpi     r16, 1                                  ;Update-Modus aktiv?
        breq    seaprs                                  ;ja -> weiter
        sts     prssen, r17                             ;sonst Anzahl der Drucksensoren speichern
        rjmp    seasen                                  ;weiter zu 1-Wire-Suche
seaprs: sts     prsse2, r17                             ;Anzahl der Drucksensoren (Update) speichern
#endif
;
; 1-Wire Sensoren suchen und ROM-IDs in SRAM schreiben, bis zu 24 Temperatursensoren und bis zu 3 Luftfeuchtigkeitssensoren werden erfasst, es wird der Algorithmus nach Application Note 187 von
; Dallas/Maxim verwendet; bei der Update-Funktion ist die Suche identisch, die Daten werden aber in einen anderen Speicherbereich geschrieben
;
seasen: ldi     r22, 0                                  ;Temperatursensornummer = 0 setzen
        ldi     r23, 0                                  ;Luftfeuchtigkeitssensornummer = 0 setzen
        ldi     r20, 0                                  ;Busnummer = 0 setzen
        ldi     zl, low(tsenid)
        ldi     zh, high(tsenid)                        ;Zeiger auf Sensor-IDs der Temperatursensoren
        ldi     yl, low(hsenid)
        ldi     yh, high(hsenid)                        ;Zeiger auf Sensor-IDs der Luftfeuchtigkeitssensoren
;
        lds     r16, modflg                             ;Modus-Flag holen
        cpi     r16, 1                                  ;Update-Modus?
        brne    seanew                                  ;nein -> normale Suche
        ldi     zl, low(tseni2)                         ;sonst
        ldi     zh, high(tseni2)                        ;Zeiger auf Sensor-IDs der Temperatursensoren (Update)
        ldi     yl, low(hseni2)
        ldi     yh, high(hseni2)                        ;Zeiger auf Sensor-IDs der Luftfeuchtigkeitssensoren (Update)
;
seanew: clr     r21                                     ;LastDiscrepancy lschen
        clr     r9                                      ;LastDeviceFlag lschen
        clr     r8                                      ;ReturnValue lschen
searom: rcall   owres                                   ;Reset senden
        brcs    sea700                                  ;Presence? nein -> Suchende
        tst     r9                                      ;LastDeviceFlag gesetzt?
        brne    sea700                                  ;ja -> Suchende
        clr     r11
        inc     r11                                     ;Bit-Number = 1 setzen
        clr     r10                                     ;Last-Zero = 0 setzen
        ldi     r19, 0xf0                               ;Search-ROM-Kommando (0xf0)
        rcall   owwrit                                  ;senden
sea100: rcall   owbrea                                  ;einzelnes Bit lesen (Bit A)
        rol     r18                                     ;Bit A sichern
        andi    r18, 0x01                               ;restliche Bits lschen
        rcall   owbrea                                  ;einzelnes Bit lesen (Bit B)
        rol     r19                                     ;Bit B sichern
        andi    r19, 0x01                               ;restliche Bits lschen
        mov     r17, r19                                ;Bit B kopieren
        and     r17, r18                                ;beide Bits = 1?
        brne    sea700                                  ;ja -> weiter
        mov     r17, r19                                ;sonst Bit B kopieren
        or      r17, r18                                ;beide Bits = 0?
        breq    sea400                                  ;ja -> weiter
        mov     r16, r18                                ;sonst Bit A speichern
sea200: bst     r16, 0                                  ;Bit sichern
        rcall   bitsto                                  ;in Puffer einschieben
        bld     r16, 0                                  ;gesichertes Bit holen
        ror     r16                                     ;und in Carry schieben
        rcall   owbwri                                  ;Bit in Carry senden
        inc     r11                                     ;Bit-Number erhhen
        mov     r16, r11                                ;Bit-Index kopieren
        cpi     r16, 65                                 ;Endwert erreicht (>64)?
        brcs    sea100                                  ;nein -> Schleife
        mov     r21, r10                                ;LastDiscrepancy = Last-Zero
        tst     r21                                     ;LastDiscrepancy = 0?
        brne    sea300                                  ;nein -> weiter
        inc     r9                                      ;sonst LastDeviceFlag setzen
sea300: rcall   crc                                     ;CRC der ROM-ID prfen
        tst     r17                                     ;CRC-Wert = 0 (ok)?
        brne    sea700                                  ;nein -> nchster Sensor
        inc     r8                                      ;Return-Value setzen
        rjmp    sea800                                  ;Ende
;
sea400: cp      r11, r21                                ;Bit-Number = Last-Discrepancy?
        brne    sea500                                  ;nein -> weiter
        ldi     r16, 1                                  ;sonst "1"-Bit setzen
sea450: cpi     r16, 0                                  ;Search-Direction = 0?
        brne    sea200                                  ;nein -> Bit speichern/senden
        mov     r10, r11                                ;Last-Zero = Bit-Number
        rjmp    sea200                                  ;Bit speichern/senden
;
sea500: brcs    sea600                                  ;Bit-Number<Last-Discrepancy? ja -> weiter
        clr     r16                                     ;sonst "0"-Bit setzen
        rjmp    sea450                                  ;Bit speichern und senden
;
sea600: lds     r16, owbuff + 7                         ;entprechendes Bit von letzter
        andi    r16, 0x01                               ;Suche holen und maskieren
        rjmp    sea450                                  ;Bit speichern und senden
;
sea700: clr     r21                                     ;LastDiscrepancy = 0 setzen
        clr     r9                                      ;LastDeviceFlag = 0 setzen
        clr     r8                                      ;ReturnValue = 0 setzen
;
sea800: tst     r8                                      ;wurde ein Sensor gefunden?
        breq    sea900                                  ;nein -> nchsten Bus testen
;
        lds     r16, owbuff + 7                         ;sonst Family-Code holen
        cpi     r16, 0                                  ;gltiger Family-Code (>0)?
        breq    sea900                                  ;nein -> nchster Bus
        cpi     r16, 0x26                               ;Family-Code = 0x26 (DS2438)?
        breq    sea840                                  ;ja -> bearbeiten
        cpi     r16, 0x10                               ;Family-Code = 0x10 (DS18S20)?
        breq    sea810                                  ;ja -> bearbeiten
        cpi     r16, 0x28                               ;Family-Code = 0x28 (DS18B20)?
        breq    sea810                                  ;ja -> bearbeiten
        cpi     r16, 0x22                               ;Family-Code = 0x22 (DS1822)?
        brne    sea880                                  ;nein -> nchster Sensor
;
sea810: cpi     r22, 24                                 ;sind bereits 24 Temperatursensoren ermittelt worden?
        brcc    sea880                                  ;ja -> Daten ignorieren und nchsten Sensor bearbeiten
        ldi     xl, low(owbuff)
        ldi     xh, high(owbuff)                        ;Zeiger auf 1-Wire-Puffer
        ldi     r19, 8                                  ;8 Bytes kopieren
sea820: ld      r18, x+                                 ;Bytes aus 1-Wire-Puffer holen
        st      z+, r18                                 ;und im ROM-ID-Feld der Temperatursensoren speichern
        dec     r19                                     ;alle Bytes bearbeitet?
        brne    sea820                                  ;nein -> Schleife
        ldi     xl, low(tsentb)
        ldi     xh, high(tsentb)                        ;Zeiger auf Temperatursensortabelle
;
        lds     r16, modflg                             ;Modus-Flag holen
        cpi     r16, 1                                  ;Update-Modus?
        brne    sea830                                  ;nein -> normale Suche
        ldi     xl, low(tsent2)                         ;sonst
        ldi     xh, high(tsent2)                        ;Zeiger auf Temperatursensortabelle (Update)
;
sea830: add     xl, r22                                 ;Sensornummer addieren
        adc     xh, r19                                 ;Tabellenplatz berechnen
        st      x, r20                                  ;Busnummer speichern
        inc     r22                                     ;Sensornummer erhhen
        rjmp    sea880                                  ;sonst nchster Sensor
;
sea840: cpi     r23, 3                                  ;sind bereits 3 Luftfeuchtigkeitssensoren ermittelt worden?
        brcc    sea880                                  ;ja -> Daten ignorieren und nchsten Sensor bearbeiten
        ldi     xl, low(owbuff)
        ldi     xh, high(owbuff)                        ;Zeiger auf 1-Wire-Puffer
        ldi     r19, 8                                  ;8 Bytes kopieren
sea850: ld      r18, x+                                 ;Bytes aus 1-Wire-Puffer holen
        st      y+, r18                                 ;und im ROM-ID-Feld der Luftfeuchtigkeitssensoren speichern
        dec     r19                                     ;alle Bytes bearbeitet?
        brne    sea850                                  ;nein -> Schleife
        ldi     xl, low(hsentb)
        ldi     xh, high(hsentb)                        ;Zeiger auf Luftfeuchtigkeitssensortabelle
;
        lds     r16, modflg                             ;Modus-Flag holen
        cpi     r16, 1                                  ;Update-Modus?
        brne    sea860                                  ;nein -> normale Suche
        ldi     xl, low(hsent2)                         ;sonst
        ldi     xh, high(hsent2)                        ;Zeiger auf Luftfeuchtigkeitssensortabelle (Update)
;
sea860: add     xl, r23                                 ;Sensornummer addieren
        adc     xh, r19                                 ;Tabellenplatz berechnen
        st      x, r20                                  ;Busnummer speichern
        inc     r23                                     ;Sensornummer erhhen
;
sea880: tst     r9                                      ;weitere Sensoren am Bus?
        brne    sea900                                  ;nein -> nchsten Bus testen
        rjmp    searom                                  ;sonst weitere Sensoren lesen
;
sea900: inc     r20                                     ;Busnummer erhhen
        cpi     r20, 8                                  ;Maximum 8 erreicht?
        breq    inmain                                  ;ja -> Ende
        rjmp    seanew                                  ;neue Suche beginnen
;
; Initialisierung der Hauptprogrammschleife
;
inmain: rcall   getsen                                  ;Anzahl der Sensoren ermitteln
        rcall   getmap                                  ;Sensorbelegung ermitteln
        rcall   dstatu                                  ;Status auf LCD ausgeben
;
        ldi     r16, 80                                 ;80 * 50ms (4s) warten
        clr     r17
        sts     intc16, r16
        sts     intc16 + 1, r17                         ;Interruptzhler auf 80 * 50ms (4s) setzen
        sei                                             ;Interrupts freigeben
inm020: lds     r16, intc16                             ;Interrupt-Zhler L holen
        lds     r17, intc16 + 1                         ;Interrupt-Zhler H holen
        or      r16, r17                                ;Zhler abgelaufen?
        brne    inm020                                  ;nein -> warten
;
        ldi     r16, cr                                 ;CR laden und
        rcall   rssend                                  ;zur Synchronisierung mit dem Empfnger senden
        ldi     xl, low(senmap)
        ldi     xh, high(senmap)                        ;Zeiger auf Sensorbelegung
        ldi     r17, 8                                  ;8 Zeichen senden
inm030: ld      r16, x+                                 ;Zeichen holen
        rcall   rssend                                  ;Zeichen senden
        dec     r17                                     ;alle Zeichen gesendet?
        brne    inm030                                  ;nein -> Schleife
        ldi     r16, cr                                 ;CR-Code laden
        rcall   rssend                                  ;und senden
;
        cli                                             ;Interrupts sperren
        ldi     r16, 21
        clr     r17
        sts     intc16, r16                             ;Interruptzhler auf
        sts     intc16 + 1, r17                         ;20 * 50ms (1s) setzen
        sei                                             ;Interrupts wieder freigeben
inm040: lds     r16, intc16                             ;Interrupt-Zhler L holen
        lds     r17, intc16 + 1                         ;Interrupt-Zhler H holen
        or      r16, r17                                ;Zhler abgelaufen?
        brne    inm040                                  ;nein -> warten
;
        ldi     zl, low(alstat)
        ldi     zh, high(alstat)                        ;Zeiger auf Alarmstatus
        ldi     r20, 4                                  ;4 Alarme bearbeiten
        cli                                             ;Interrupts sperren
inm050: ld      r19, z                                  ;Alarmstatus holen
        cbr     r19, 0b00001000                         ;Melde-Flag lschen
        st      z+, r19                                 ;Alarmstatus wieder speichern
        dec     r20                                     ;alle Alarme bearbeitet?
        brne    inm050                                  ;nein -> Schleife
;
        cbr     flags, 0b00000001                       ;Interrupt-Flag lschen
        sei                                             ;Interrupts wieder freigeben
;
        lds     r16, modflg                             ;Modus-Flag holen
        cpi     r16, 1                                  ;Update-Modus?
        brne    main                                    ;nein -> normaler Betrieb
        rjmp    update                                  ;sonst Update bearbeiten
;
;
; =====================================================================================================================================================================================================
; Hauptprogrammschleife
; =====================================================================================================================================================================================================
;
; Aktuelle Alarmmeldungen ausgeben
;
main:   ldi     zl, low(alstat)
        ldi     zh, high(alstat)                        ;Zeiger auf Alarmstatus
        ldi     r20, 0                                  ;aktuelle Alarmnummer
ma0010: ld      r19, z                                  ;Alarmstatus holen
        bst     r19, 3                                  ;Melde-Flag gesetzt?
        brtc    ma0030                                  ;nein -> weiter
;
        bst     flags, 2                                ;Automatik-Anzeige aktiv?
        brts    ma0020                                  ;nein -> kein Anzeigewechsel
;
        sts     lastal, r20                             ;Alarmnummer speichern als
        ldi     r16, low(almtim)                        ;letzten Alarm
        ldi     r17, high(almtim)
        sts     intc16, r16                             ;Interrupt-Zhler auf
        sts     intc16 + 1, r17                         ;Alarm-Anzeige-Zeit setzen
;
ma0020: cli                                             ;Interrupts sperren
        ld      r19, z                                  ;Alarmstatus nochmals holen
        cbr     r19, 0b00001000                         ;Melde-Flag lschen
        st      z, r19                                  ;Alarmstatus wieder speichern
        sei                                             ;Interrupts wieder aktivieren
        swap    r19                                     ;Alarmstatus an Bitposition 0
        andi    r19, 0x01                               ;restliche Bits entfernen
        rcall   rsaout                                  ;Alarmmeldung senden
        sbr     flags, 0b00000010                       ;nderungs-Flag setzen
;
ma0030: adiw    zl, 1                                   ;Zeiger auf nchsten Alarm
        inc     r20                                     ;Alarmnummer erhhen
        cpi     r20, 4                                  ;alle Alarme bearbeitet?
        brne    ma0010                                  ;nein -> Schleife
;
; Interrupt-Zhler prfen fr automatische Rckkehr aus Einstellungs-Men in die Automatik-Anzeige
;
        lds     r16, semode                             ;Einstellungs-Modus holen
        tst     r16                                     ;Einstellungs-Modus aktiv?
        breq    ma0040                                  ;nein -> berspringen
        lds     r16, intc16                             ;Interrupt-Zhler L holen
        lds     r17, intc16 + 1                         ;Interrupt-Zhler H holen
        or      r16, r17                                ;Zhler = 0?
        brne    ma0040                                  ;nein -> bersrpingen
        sts     semode, r16                             ;Einstellungs-Modus=0 setzen
        sts     dimode, r16                             ;Anzeige-Modus=0 setzen
        cbr     flags, 0b00000100                       ;Anzeige-Automatik aktivieren
        sbr     flags, 0b00000010                       ;nderungs-Flag setzen
;
; Zyklus-Steuerung nach folgendem Muster:
;
; Zeitpunkt        0ms: Messung starten
; Zeitpunkt      900ms: Messung beenden
; Zeitpunkt     1000ms: Sensornummer 1 setzen / Sensor 1 lesen
; Zeitpunkt + <intval>: Sensornummer 2 setzen / Sensor 2 lesen
; Zeitpunkt + <intval>: Sensornummer n setzen / Sensor n lesen (bis n = 24)
; Zeitpunkt + <intval>: Sensornummer 25 setzen / Luftdrucksensor lesen (wenn vorhanden)
; Zeitpunkt + <intval>: Sensornummer 26 setzen / Temperatur von Luftfeuchtigkeitssensor 1 lesen (wenn vorhanden)
; Zeitpunkt + <intval>: Sensornummer 27 setzen / Temperatur von Luftfeuchtigkeitssensor 2 lesen (wenn vorhanden)
; Zeitpunkt + <intval>: Sensornummer 28 setzen / Temperatur von Luftfeuchtigkeitssensor 3 lesen (wenn vorhanden)
; Zeitpunkt + <intval>: Sensornummer 29 setzen / Feuchtigkeit von Luftfeuchtigkeitssensor 1 lesen (wenn vorhanden)
; Zeitpunkt + <intval>: Sensornummer 30 setzen / Feuchtigkeit von Luftfeuchtigkeitssensor 2 lesen (wenn vorhanden)
; Zeitpunkt + <intval>: Sensornummer 31 setzen / Feuchtigkeit von Luftfeuchtigkeitssensor 3 lesen (wenn vorhanden)
; Zeitpunkt + <intval>: Sensornummer 0 setzen und Sprung zu Zeitpunkt 0
;
ma0040: bst     flags, 0                                ;Interrupt-Flag gesetzt?
        brts    ma0050                                  ;ja -> weiter
        rjmp    ma2000                                  ;sonst berspringen
;
ma0050: cbr     flags, 0b00000001                       ;Interrupt-Flag wieder lschen
        bst     flags, 6                                ;langer Tastendruck aktiv?
        brts    ma0070                                  ;ja -> bearbeiten
;
        clr     r16                                     ;sonst Zhler fr erweiterten Tastendruck lschen
ma0060: sts     extkey, r16                             ;und speichern
        rjmp    ma0080
;
ma0070: lds     r16, extkey                             ;Zhler fr erweiterten Tastendruck holen
        inc     r16                                     ;erhhen
        cpi     r16, keytix - keytio                    ;Zhlerstand erreicht (entspricht 10s)?
        brcs    ma0060                                  ;nein -> Zhler speichern und Ende
;
        ldi     r16, low(emodfl)                        ;sonst
        ldi     r17, high(emodfl)                       ;Adresse des Modus-Flag laden
        ldi     r18, 1                                  ;Modus 1 (Update)
        rcall   eewrit                                  ;in EEPROM speichern
        rcall   w050ms                                  ;50ms warten
        rjmp    reset                                   ;System neu starten
;
ma0080: lds     r17, intcnt                             ;Interruptzhler holen
        inc     r17                                     ;erhhen
        sts     intcnt, r17                             ;und wieder speichern
        lds     r16, sencnt                             ;Sensorzhler holen
        tst     r16                                     ;Sensorzhler = 0?
        brne    ma0180                                  ;nein -> weiter testen
        cpi     r17, 1                                  ;Mess-Start-Zeitpunkt?
        breq    ma0200                                  ;ja -> bearbeiten
        cpi     r17, 19                                 ;Mess-Stop-Zeitpunkt?
        brne    ma0090                                  ;nein -> weiter
        rjmp    ma0400                                  ;sonst bearbeiten
ma0090: cpi     r17, 21                                 ;Sensor-Abfrage beginnen?
        brne    ma0190                                  ;nein -> Ende
;
ma0100: clr     r17                                     ;Interruptzhler lschen
        sts     intcnt, r17                             ;und wieder speichern
        inc     r16                                     ;Sensorzhler erhhen
        sts     sencnt, r16                             ;und wieder speichern
        lds     r18, temsen                             ;Sensoranzahl holen
        inc     r18                                     ;erhhen
        cp      r16, r18                                ;letzter Temperatursensor erreicht?
        brcc    ma0110                                  ;ja -> weiter mit Druck/Feuchte-Sensoren
        rjmp    ma1000                                  ;sonst Sensor lesen
;
ma0110: cpi     r16, 25                                 ;Sensornummer < 25?
        brcc    ma0120                                  ;nein -> weiter
        ldi     r16, 25                                 ;sonst Sensornummer = 25 setzen (Luftdruck)
        sts     sencnt, r16                             ;und wieder speichern
ma0120: cpi     r16, 25                                 ;Sensornummer = 25?
        brne    ma0140                                  ;nein -> weiter
        lds     r18, prssen                             ;Anzahl der Luftdrucksensoren holen
        tst     r18                                     ;Luftdrucksensor vorhanden?
        breq    ma0130                                  ;nein -> weiter
        rjmp    ma1600                                  ;sonst Luftdrucksensor lesen
;
ma0130: inc     r16                                     ;Sensornummer auf Temperatur des ersten Feuchtigkeitssensor setzen (26)
        sts     sencnt, r16                             ;und speichern
ma0140: lds     r18, humsen                             ;Anzahl der Luftfeuchtigkeitssensoren holen
        subi    r18, -26                                ;Basiswert 26 addieren
        cp      r16, r18                                ;letzter Luftfeuchtigkeitssensor erreicht?
        brcc    ma0150                                  ;ja -> weiter mit Luftfeuchtigkeitsberechnung
        rjmp    ma1700                                  ;sonst Temperatur des Luftfeuchtigkeitssensors lesen
;
ma0150: cpi     r16, 29                                 ;Sensornummer < 29?
        brcc    ma0160                                  ;nein -> weiter
        ldi     r16, 29                                 ;sonst Sensornummer = 29 (Temperatur der Feuchtesensoren)
        sts     sencnt, r16                             ;und wieder speichern
ma0160: lds     r18, humsen                             ;Anzahl der Luftfeuchtigkeitssensoren holen
        subi    r18, -29                                ;Basiswert 29 addieren
        cp      r16, r18                                ;letzter Luftfeuchtigkeitssensor erreicht?
        brcc    ma0170                                  ;ja -> Ende, neuen Zyklus beginnen
        rjmp    ma1a00                                  ;sonst Luftfeuchtigkeit ermitteln
;
ma0170: clr     r16                                     ;sonst Sensorzhler lschen
        sts     sencnt, r16                             ;und wieder speichern
        sbr     flags, 0b00000010                       ;nderungs-Flag setzen
        rjmp    ma0080                                  ;neuer Zyklus
;
ma0180: lds     r18, intval                             ;Sende-Intervall holen
        cp      r17, r18                                ;Endwert erreicht?
        brcc    ma0100                                  ;ja -> bearbeiten
ma0190: rjmp    ma2000                                  ;sonst Ende
;
; Neue Messung starten
;
ma0200: clr     r20                                     ;Busnummer = 0 setzen
ma0210: ldi     xl, low(tsentb)
        ldi     xh, high(tsentb)                        ;Zeiger auf Temperatursensortabelle
ma0220: ld      r19, x+                                 ;verwendeten Bus des Sensors holen
        cp      r19, r20                                ;Sensor am aktuellen Bus?
        breq    ma0240                                  ;ja -> Messung auf Bus starten
        cpi     xl, low(tsentb + 24)                    ;Tabellenende erreicht?
        brne    ma0220                                  ;nein -> nchster Tabellenplatz
;
        ldi     xl, low(hsentb)
        ldi     xh, high(hsentb)                        ;Zeiger auf Luftfeuchtigkeitssensortabelle
ma0230: ld      r19, x+                                 ;verwendeten Bus des Sensors holen
        cp      r19, r20                                ;Sensor am aktuellen Bus?
        breq    ma0240                                  ;ja -> Messung auf Bus starten
        cpi     xl, low(hsentb + 3)                     ;Tabellenende erreicht?
        brne    ma0230                                  ;nein -> nchster Tabellenplatz
        rjmp    ma0250                                  ;sonst nchster Bus
;
ma0240: cli                                             ;Interrupts sperren
        rcall   owres                                   ;1-Wire-Reset senden
        brcs    ma0250                                  ;Presence? nein -> Fehler
        ldi     r19, 0xcc                               ;ROM-Kommando "Skip ROM" senden
        rcall   owwrit                                  ;(alle Sensoren ansprechen)
        ldi     r19, 0x44                               ;Funktion-Kommando "Convert T"
        rcall   owwrit                                  ;senden
        rcall   owstpu                                  ;fr sichere Messung Strong-Pull-up aktivieren
;
ma0250: sei                                             ;Interrupts wieder freigeben
        inc     r20                                     ;nchsten Bus bearbeiten
        cpi     r20, 8                                  ;letzter Bus erreicht?
        brne    ma0210                                  ;nein -> Schleife
;
; Alarmmeldungen erneut senden
;
        ldi     zl, low(alstat)
        ldi     zh, high(alstat)                        ;Zeiger auf Alarmstatus
        ldi     r20, 0                                  ;aktuelle Alarmnummer
ma0300: ld      r19, z+                                 ;Alarmstatus holen
        swap    r19                                     ;Alarmstatus an Bitposition 0
        andi    r19, 0x01                               ;restliche Bits entfernen
        rcall   rsaout                                  ;Alarmmeldung senden
        inc     r20                                     ;Alarmnummer erhhen
        cpi     r20, 4                                  ;alle Alarme bearbeitet?
        brne    ma0300                                  ;nein -> Schleife
        rjmp    ma2000
;
; Messung stoppen
;
ma0400: clr     r20                                     ;Busnummer=0 setzen
ma0410: ldi     xl, low(tsentb)
        ldi     xh, high(tsentb)                        ;Zeiger auf Temperatursensortabelle
ma0420: ld      r19, x+                                 ;verwendeten Bus des Sensors holen
        cp      r19, r20                                ;Sensor am aktuellen Bus?
        breq    ma0430                                  ;ja -> Messung auf Bus stoppen
        cpi     xl, low(tsentb + 24)                    ;Tabellenende erreicht?
        brne    ma0420                                  ;nein -> nchster Tabellenplatz
;
        ldi     xl, low(hsentb)
        ldi     xh, high(hsentb)                        ;Zeiger auf Luftfeuchtigkeitssensortabelle
ma0425: ld      r19, x+                                 ;verwendeten Bus des Sensors holen
        cp      r19, r20                                ;Sensor am aktuellen Bus?
        breq    ma0430                                  ;ja -> Messung auf Bus stoppen
        cpi     xl, low(hsentb + 3)                     ;Tabellenende erreicht?
        brne    ma0425                                  ;nein -> nchster Tabellenplatz
        rjmp    ma0440                                  ;sonst nchster Bus
;
ma0430: cli                                             ;Interrupts sperren
        rcall   owres                                   ;1-Wire-Reset, Strong-Pull-up wieder ausschalten
        sei                                             ;Interrupts wieder freigeben
ma0440: inc     r20                                     ;nchsten Bus bearbeiten
        cpi     r20, 8                                  ;letzter Bus erreicht?
        brne    ma0410                                  ;nein -> Schleife
        rjmp    ma2000
;
; Temperatursensoren bearbeiten
; Temperaturen: Sensor-Daten holen, Scratchpad aus Sensor auslesen
;
ma1000: dec     r16                                     ;Sensornummer korrigieren
;
ma1100: mov     r17, r16                                ;Sensor-Nummer kopieren
        mov     r22, r16                                ;Sensor-Nummer sichern
        cpi     r17, 8                                  ;Sensor-Nummer>8?
        brcc    ma1110                                  ;ja -> als Buchstabe ausgeben
        ldi     r17, 49
        add     r17, r16                                ;in ASCII (1-8) wandeln
        rjmp    ma1120                                  ;ausgeben
ma1110: ldi     r17, 89
        add     r17, r16                                ;in ASCII (a-h) wandeln
ma1120: sts     rsbuff, r17                             ;in LCD-Puffer schreiben
        ldi     r17, ':'                                ;Doppelpunkt
        sts     rsbuff + 1, r17                         ;in LCD-Puffer schreiben
        ldi     r17, ' '                                ;Leerzeichen
        sts     rsbuff + 2, r17                         ;in LCD-Puffer schreiben
;
        ldi     xl, low(tsentb)
        ldi     xh, high(tsentb)                        ;Zeiger auf Sensor-Tabelle
        clr     r17                                     ;H-Byte=0 fr Tabellenplatzberechnung
        add     xl, r16                                 ;Tabellenplatz berechnen
        adc     xh, r17                                 ;bertrag addieren
        ld      r20, x                                  ;zugehrige Busnummer holen
        cpi     r20, 0xff                               ;Sensor vorhanden?
        brne    ma1130                                  ;ja -> weiter
        rjmp    ma1500                                  ;sonst Fehler
;
ma1130: ldi     xl, low(tsenid)
        ldi     xh, high(tsenid)                        ;Zeiger auf Sensor-ROM-IDs
        add     r16, r16
        add     r16, r16
        add     r16, r16                                ;Sensor-Nummer * 8 als Offset
        add     xl, r16                                 ;Tabellenplatz berechnen
        adc     xh, r17                                 ;bertrag addieren
        adiw    xl, 7                                   ;Zeiger auf Platz vom Family-Code
        ld      r7, x+                                  ;Code holen, Zeiger auf Ende
;
        cli                                             ;Interrupts sperren
        rcall   owres                                   ;1-Wire-Reset senden
        sei                                             ;Interrupts wieder freigeben
        brcc    ma1140                                  ;Presence? ja -> weiter
        rjmp    ma1500                                  ;sonst Fehler
;
ma1140: cli                                             ;Interrupts wieder sperren
        rcall   owmaro                                  ;ROM-Kommando "Match ROM" und ROM-ID senden
        ldi     r19, 0xbe                               ;Funktion "Read Scratchpad"
        rcall   owwrit                                  ;senden
        ldi     xl, low(owbuff + 8)
        ldi     xh, high(owbuff + 8)                    ;Zeiger auf 1-Wire-Pufferende
        ldi     r21, 8                                  ;8 Bytes Scratchpad einlesen
ma1160: rcall   owread                                  ;Byte lesen
        st      -x, r19                                 ;Byte in Puffer speichern
        dec     r21                                     ;alle Bytes gelesen?
        brne    ma1160                                  ;nein -> Schleife
        rcall   owread                                  ;CRC-Byte lesen
;
        sei                                             ;Interrupts wieder freigeben
        mov     r21, r19                                ;gelesenen CRC-Wert sichern
        rcall   crc                                     ;CRC vom Scratchpad ermitteln
        cp      r17, r21                                ;war CRC-Wert ok?
        brne    ma1170                                  ;nein -> Fehler
        lds     r16, owbuff                             ;zustzlich Count_Per_C holen
        cpi     r16, 0x10                               ;und prfen, Count_Per_C = 10h?
        breq    ma1180                                  ;ja -> weiter
ma1170: rjmp    ma1500                                  ;sonst Fehler
;
ma1180: ldi     r16, 0x10                               ;Family-Code 0x10 (DS18S20)
        cp      r7, r16                                 ;Family-Code ok?
        brne    ma1250                                  ;nein -> andere Family-Codes bearbeiten
;
; Temperaturen: DS1820/DS18S20/DS1920: Gelesene Daten aufbereiten, Erhhung der Mess-Genauigkeit durch Algorithmus nach Datenblatt; die Formel wurde etwas modifiziert, so dass mit ganzzahligen Werten
; gerechnet werden kann und keine Divisionsroutine ntig ist: Temp = (160 * Temp_Read - 40 + 10 * (Count_Per_C - Count_Remain)) / 16. Das Ergebnis liegt dann in Zehntel Grad vor.
;
ma1200: lds     r17, owbuff + 6                         ;Temperaturwert H holen
        lds     r16, owbuff + 7                         ;Temperaturwert L holen
        ror     r17
        ror     r16                                     ;LSB abschneiden (Temp_Read)
        ldi     r17, 160                                ;Wert fr hhere Genauigkeit mit 160
        mulsu   r16, r17                                ;multiplizieren
        movw    xl, r0                                  ;Ergebnis kopieren
        sbiw    xl, 40                                  ;40 subtrahieren
        ldi     r16, 16                                 ;Count_Per_C laden (immer 16)
        lds     r17, owbuff + 1                         ;Count_Remain holen
        sub     r16, r17                                ;Count_Per_C - Count_Remain
        add     r16, r16                                ;Wert * 2
        mov     r17, r16                                ;kopieren
        add     r16, r16                                ;Wert nochmals * 2 (* 4)
        add     r16, r16                                ;Wert nochmals * 2 (* 8)
        add     r16, r17                                ;* 2 Zwischenergebnis addieren (* 10)
        clr     r17                                     ;H-Byte=0
        add     xl, r16                                 ;zur vorherigen Rechnung addieren
        adc     xh, r17                                 ;bertrag addieren
        asr     xh
        ror     xl                                      ;Wert / 2
        asr     xh
        ror     xl                                      ;Wert / 2 (/ 4)
        asr     xh
        ror     xl                                      ;Wert / 2 (/ 8)
        asr     xh
        ror     xl                                      ;Wert / 2 (/ 16)
        brcc    ma1210                                  ;runden? nein -> weiter
        adiw    xl, 1                                   ;sonst Ergebnis aufrunden
ma1210: mov     r20, xh                                 ;Vorzeichen sichern
        bst     xh, 7                                   ;Ergebnis negativ?
        brtc    ma1300                                  ;nein -> berspringen
        clr     r16                                     ;Wert 0x10000 zur Wandlung ins
        clr     r17                                     ;Zweierkomplement laden
        sub     r16, xl                                 ;L subtrahieren
        sbc     r17, xh                                 ;H mit bertrag subtrahieren
        movw    xl, r16                                 ;ins Ergebnis-Register kopieren
        rjmp    ma1300                                  ;in ASCII wandeln
;
; Temperaturen: DS18B20/DS1822: Gelesene Daten aufbereiten, Umrechnung des 12-Bit Binrwertes in einen Dezimalwert in Zehntel Grad nach folgender Formel: Temp = Temp_Register * 5 / 8
;
ma1250: lds     r16, owbuff + 3                         ;Configuration Register holen
        cpi     r16, 0x7f                               ;Sensor auf 12-Bit eingestellt?
        breq    ma1260                                  ;ja -> weiter
        rjmp    ma1500                                  ;sonst Fehler
ma1260: lds     xh, owbuff + 6                          ;Temperaturwert H holen
        lds     xl, owbuff + 7                          ;Temperaturwert L holen
        mov     r20, xh                                 ;Vorzeichen sichern
        bst     xh, 7                                   ;Temperaturwert negativ?
        brtc    ma1270                                  ;nein -> weiter
        clr     r16
        clr     r17                                     ;Wert 0x10000 zur Wandlung ins Zweierkomplement laden
        sub     r16, xl                                 ;L subtrahieren
        sbc     r17, xh                                 ;H mit bertrag subtrahieren
        movw    xl, r16                                 ;ins Ergebnis-Register kopieren
ma1270: movw    r16, xl                                 ;Temperaturwert kopieren
        add     xl, xl
        adc     xh, xh                                  ;Temperaturwert verdoppeln
        add     xl, xl
        adc     xh, xh                                  ;nochmals verdoppeln (* 4)
        add     xl, r16
        adc     xh, r17                                 ;und Originalwert addieren (* 5)
        asr     xh
        ror     xl                                      ;Wert / 2
        asr     xh
        ror     xl                                      ;Wert / 2 (/ 4)
        asr     xh
        ror     xl                                      ;Wert / 2 (/ 8)
        brcc    ma1300                                  ;runden erforderlich? nein -> weiter
        adiw    xl, 1                                   ;sonst Ergebnis aufrunden
;
; Temperaturen: Ergebnis in Dezimal-ASCII umwandeln und aufbereiten
;
ma1300: rcall   dezasc                                  ;in Dezimal-ASCII umwandeln
        cpi     r19, '0'                                ;ASCII Stelle 1 =0?
        brne    ma1350                                  ;nein -> weiter
        ldi     r19, ' '                                ;sonst durch Leerzeichen ersetzen
        cpi     r18, '0'                                ;hchste Ziffer=0?
        brne    ma1350                                  ;nein -> berspringen
        ldi     r18, ' '                                ;sonst durch Leerzeichen ersetzen
ma1350: bst     r20, 7                                  ;ist Temperaturwert negativ?
        brtc    ma1370                                  ;nein -> berspringen
        cpi     r18, ' '                                ;Stelle 2 = Leerzeichen?
        breq    ma1360                                  ;ja -> durch Minus ersetzen
        ldi     r19, '-'                                ;sonst Stelle 1 = Minuszeichen
        rjmp    ma1370
ma1360: ldi     r18, '-'                                ;Stelle 2 = Minuszeichen
;
; Temperaturen: Ergebnis in RS-232 Puffer und Wertetabelle legen
;
ma1370: sts     rsbuff + 2, r19                         ;ASCII Stelle 1 in RS232-Puffer
        sts     rsbuff + 3, r18                         ;ASCII Stelle 2 in RS232-Puffer
        sts     rsbuff + 4, r17                         ;ASCII Stelle 3 in RS232-Puffer
        ldi     r20, '.'                                ;Dezimalpunkt
        sts     rsbuff + 5, r20                         ;in RS232-Puffer schreiben
        sts     rsbuff + 6, r16                         ;ASCII Stelle 4 in RS232-Puffer
;
        mov     r21, r22                                ;gesicherte Sensornummer holen
        add     r21, r21
        add     r21, r21
        add     r21, r22                                ;Adress-Offset von Wertetabelle berechnen (Sensornummer * 5)
        clr     r23                                     ;H-Byte=0
        ldi     zl, low(temval)
        ldi     zh, high(temval)                        ;Zeiger auf Wertetabelle
        add     zl, r21
        adc     zh, r23                                 ;Tabellenplatz berechnen
        st      z+, r19                                 ;ASCII-Stelle 1 in Tabelle
        st      z+, r18                                 ;ASCII-Stelle 2 in Tabelle
        st      z+, r17                                 ;ASCII-Stelle 3 in Tabelle
        ldi     r20, ','                                ;Komma
        st      z+, r20                                 ;in Tabelle schreiben
        st      z+, r16                                 ;ASCII-Stelle 4 in Tabelle
        sbr     flags, 0b00000010                       ;nderungs-Flag setzen
;
; Temperaturen: Flags bearbeiten, Wert auf 85,0 Grad prfen
;
        ldi     xl, low(senflg)
        ldi     xh, high(senflg)                        ;Zeiger auf Sensorflags
        add     xl, r22                                 ;gesicherte Sensornummer
        adc     xh, r23                                 ;addieren, bertrag
        ld      r17, x                                  ;Sensorflags holen
;
        lds     r16, rsbuff + 2                         ;ASCII Stelle 1 holen
        cpi     r16, ' '                                ;Wert=0?
        brne    ma1390                                  ;nein -> Flag lschen
        lds     r16, rsbuff + 3                         ;ASCII Stelle 2 holen
        cpi     r16, '8'                                ;Wert=8?
        brne    ma1390                                  ;nein -> Flag lschen
        lds     r16, rsbuff + 4                         ;ASCII Stelle 3 holen
        cpi     r16, '5'                                ;Wert=5?
        brne    ma1390                                  ;nein -> Flag lschen
        lds     r16, rsbuff + 6                         ;ASCII Stelle 4 holen
        cpi     r16, '0'                                ;Wert=0?
        brne    ma1390                                  ;nein -> Flag lschen
;
        bst     r17, 0                                  ;85-Grad-Flag gesetzt?
        brts    ma1400                                  ;ja -> Daten ausgeben
        sbr     r17, 0b00000001                         ;sonst 85-Grad-Flag setzen
        st      x, r17                                  ;Sensorflags wieder speichern
        rjmp    ma2000                                  ;Datenausgabe berspringen
;
ma1390: cbr     r17, 0b00000001                         ;85-Grad-Flag lschen
        st      x, r17                                  ;Sensorflags wieder speichern
;
; Temperaturen: Daten ber RS-232 senden
;
ma1400: ldi     xl, low(rsbuff)
        ldi     xh, high(rsbuff)                        ;Zeiger auf RS232-Puffer
        ldi     r17, 7                                  ;7 Zeichen senden
ma1410: ld      r16, x+                                 ;Zeichen holen
        rcall   rssend                                  ;Zeichen senden
        dec     r17                                     ;alle Zeichen gesendet?
        brne    ma1410                                  ;nein -> Schleife
        ldi     r16, cr                                 ;CR-Code laden
        rcall   rssend                                  ;und senden
        rjmp    ma2000
;
; Temperaturen: Fehler aufgetreten - Wert in Tabelle lschen
;
ma1500: mov     r21, r22                                ;gesicherte Sensornummer holen
        add     r21, r21
        add     r21, r21                                ;Adress-Offset von Wertetabelle
        add     r21, r22                                ;berechnen (Sensornummer * 5)
        clr     r23                                     ;H-Byte=0
        ldi     zl, low(temval)
        ldi     zh, high(temval)                        ;Zeiger auf Wertetabelle
        add     zl, r21                                 ;Tabellenplatz
        adc     zh, r23                                 ;berechnen
        ldi     r20, ' '                                ;Leerzeichen
        st      z+, r20                                 ;ASCII-Stelle 1 in Tabelle
        st      z+, r20                                 ;ASCII-Stelle 2 in Tabelle
        st      z+, r20                                 ;ASCII-Stelle 3 in Tabelle
        st      z+, r20                                 ;Trennzeichen in Tabelle
        st      z+, r20                                 ;ASCII-Stelle 4 in Tabelle
        sbr     flags, 0b00000010                       ;nderungs-Flag setzen
        rjmp    ma2000
;
; Luftdruck-Sensor bearbeiten
; Luftdruck: ADC auslesen und Wert im Puffer speichern
;
ma1600: ldi     r17, 'w'
        sts     rsbuff, r17                             ;in LCD-Puffer schreiben
        ldi     r17, ':'                                ;Doppelpunkt
        sts     rsbuff + 1, r17                         ;in LCD-Puffer schreiben
        ldi     r17, ' '                                ;Leerzeichen
        sts     rsbuff + 2, r17                         ;in LCD-Puffer schreiben
;
        ldi     zl, low(prbuff)                         ;zunchst die letzten 7 Werte im Puffer verschieben
        ldi     zh, high(prbuff)                        ;Zeiger auf Puffer fr Drucksensordaten
        ldi     r18, 7                                  ;7 Eintrge verschieben
        clr     r19                                     ;Flag fr Erkennung des leeren Puffers lschen
ma1610: ldd     r16, z + 2                              ;Eintrag vom Puffer-Platz n + 1 holen (Low-Byte)
        ldd     r17, z + 3                              ;Eintrag vom Puffer-Platz n + 1 holen (High-Byte)
        st      z+, r16                                 ;auf Puffer-Platz n ablegen (Low)
        st      z+, r17                                 ;auf Puffer-Platz n ablegen (High)
        or      r19, r16                                ;Low-Wert mit Flag verknpfen
        or      r19, r17                                ;High-Wert mit Flag verknpfen
        dec     r18                                     ;alle Eintrge verschoben?
        brne    ma1610                                  ;nein -> Schleife
;
        in      r16, adcl                               ;neuen Luftdruckwert vom Sensor lesen (Low)
        in      r17, adch                               ;neuen Luftdruckwert vom Sensor lesen (High)
        st      z+, r16                                 ;Eintrag auf Puffer-Platz 7 speichern (Low)
        st      z+, r17                                 ;Eintrag auf Puffer-Platz 7 speichern (High)
        tst     r19                                     ;war Puffer leer?
        brne    ma1630                                  ;nein -> weiter
;
        ldi     zl, low(prbuff)                         ;sonst Puffer komplett mit dem ersten Sensorwert fllen
        ldi     zh, high(prbuff)                        ;Zeiger auf Puffer fr Drucksensordaten
        ldi     r18, 7                                  ;7 Eintrge schreiben
ma1620: st      z+, r16                                 ;auf Puffer-Platz n ablegen (Low)
        st      z+, r17                                 ;auf Puffer-Platz n ablegen (High)
        dec     r18                                     ;alle Eintrge verschoben?
        brne    ma1620                                  ;nein -> Schleife
;

ma1630: ldi     zl, low(prbuff)                         ;Mittelwert der 8 Sensorwerte im Puffer ermitteln
        ldi     zh, high(prbuff)                        ;Zeiger auf Puffer fr Drucksensordaten
        clr     r16
        clr     r17                                     ;Anfangswert fr Mittelwertberechnung = 0 setzen
        ldi     r20, 8                                  ;8 Sensorwerte addieren
ma1640: ld      r18, z+                                 ;Pufferplatz lesen (Low)
        ld      r19, z+                                 ;Pufferplatz lesen (High)
        add     r16, r18                                ;Sensorwert addieren (Low)
        adc     r17, r19                                ;Sensorwert addieren (High)
        dec     r20                                     ;alle Sensorwerte addiert?
        brne    ma1640                                  ;nein -> Schleife
        lsr     r17                                     ;Summenwert durch 8 dividieren (3 mal nach rechtsverschieben)
        ror     r16                                     ;Summenwert nach rechts verschieben (/ 2)
        lsr     r17
        ror     r16                                     ;Summenwert nach rechts verschieben (/ 2)
        lsr     r17
        ror     r16                                     ;Summenwert nach rechts verschieben (/ 2)
        brcc    ma1650                                  ;Runden erforderlich? nein -> weiter
        ldi     r18, 1                                  ;sonst 1 addieren
        add     r16, r18                                ;Low-Byte addieren
        adc     r17, r20                                ;High-Byte (bertrag) addieren
;
; Luftdruck: Umrechnen des Sensormittelwertes nach der Transfer-Funktion des Datenblattes des MPXA6115A:
;
; Vout = Vs * (0.009 * P - 0.095)       Vout = Spannung am Sensorausgang, Vs = Betriebsspannung, P = Druck in kPa
;
; Wir haben den Messwert Vout und wollen P ermitteln:
;
; P = (Vout / Vs + 0.095) / 0.009
;
; Modifikation der Formel: Die Spannung am Sensorausgang steht im gleichen Verhltnis zu 5V wie der ADC-Wert zu 1024, somit kann nochmals vereinfacht werden:
;
; P = (ADC-Wert / 1024 + 0.095) / 0.009
;
; Um mit der ganzzahligen Arithmetik des ATmega auszukommen, wird mit den tausendfachen Werten gerechnet. Wenn jetzt der ADC-Wert nochmals verzehnfacht wird, dann erhalten wir das fertige Ergebnis in
; hPa:
;
; P = (ADC-Wert * 10000 / 1024 + 950) / 9
;
ma1650: movw    r4, r16                                 ;Mittelwert-Ergebnis kopieren
        ldi     r20, low(10000)                         ;und
        ldi     r21, high(10000)                        ;zunchst mit 10000 multiplizieren
        rcall   umul16                                  ;Multiplikation r21:r20 * r5:r4 = r19:r16
;
        mov     r16, r17                                ;jetzt den Wert durch 1024 dividieren, zunchst eine Division durch 256
        mov     r17, r18                                ;durch kopieren von Middle- auf Low-Byte und High- auf Middle-Byte
        lsr     r17                                     ;Wert durch Rechtsverschiebung halbieren (High), ergibt / 512
        ror     r16                                     ;Wert durch Rechtsverschiebung halbieren (Low)
        lsr     r17                                     ;Wert durch Rechtsverschiebung nochmals halbieren (High), ergibt / 1024
        ror     r16                                     ;Wert durch Rechtsverschiebung nochmals halbieren (Low)
;
        ldi     r18, low(950)                           ;Festwert 950 laden (Low)
        ldi     r19, high(950)                          ;Festwert 950 laden (High)
        add     r16, r18                                ;Festwert addieren (Low)
        adc     r17, r19                                ;Festwert addieren (High)
;
        clr     r18                                     ;hchstwertiges Byte fr Division lschen
        ldi     r19, 9                                  ;Division durch 9
        mov     r4, r19                                 ;an Divisor-Position kopieren
        clr     r5                                      ;Divisor High-Byte lschen
        rcall   udiv24                                  ;Division r18:r16 / r5:r4 = r17:r16
;
; Luftdruck: Ergebnis in Dezimal-ASCII umwandeln
;
        movw    xl, r16                                 ;Endergebnis kopieren
        rcall   dezasc                                  ;in Dezimal-ASCII umwandeln
        cpi     r19, '1'                                ;ist die Tausender-Stelle = 1?
        breq    ma1660                                  ;ja -> weiter
        ldi     r19, ' '                                ;sonst durch Leerzeichen ersetzen
;
; Luftdruck: Ergebnis in RS-232 Puffer und Wertetabelle legen
;
ma1660: sts     rsbuff + 3, r19                         ;ASCII Tausender-Stelle in RS232-Puffer legen
        sts     rsbuff + 4, r18                         ;ASCII Hunderter-Stelle in RS232-Puffer legen
        sts     rsbuff + 5, r17                         ;ASCII Zehner-Stelle in RS232-Puffer legen
        sts     rsbuff + 6, r16                         ;ASCII Einer-Stelle in RS232-Puffer legen
;
        ldi     zl, low(senval + 1)
        ldi     zh, high(senval + 1)                    ;Zeiger auf Luftdruckwert
        st      z+, r19                                 ;Tausender speichern
        st      z+, r18                                 ;Hunderter speichern
        st      z+, r17                                 ;Zehner speichern
        st      z+, r16                                 ;Einer speichern
        sbr     flags, 0b00000010                       ;nderungs-Flag setzen
;
; Luftdruck: Daten ber RS-232 senden
;
        ldi     xl, low(rsbuff)
        ldi     xh, high(rsbuff)                        ;Zeiger auf RS232-Puffer
        ldi     r17, 7                                  ;7 Zeichen senden
ma1670: ld      r16, x+                                 ;Zeichen holen
        rcall   rssend                                  ;Zeichen senden
        dec     r17                                     ;alle Zeichen gesendet?
        brne    ma1670                                  ;nein -> Schleife
        ldi     r16, cr                                 ;CR-Code laden
        rcall   rssend                                  ;und senden
        rjmp    ma2000
;
; Luftfeuchtigkeitssensoren bearbeiten - Temperatur
; Luftfeuchtigkeit (Temperatur): ADC starten und Temperatur sowie Betriebsspannung des Luftfeuchtigkeitssensors lesen
;
ma1700: subi    r16, 26                                 ;Sensornummer korrigieren (26-28 -> 0-2)
        mov     r17, r16                                ;Sensor-Nummer kopieren
        mov     r22, r16                                ;Sensor-Nummer sichern
        subi    r17, -'q'                               ;Sensor-Kennung generieren (q, r, s)
        sts     rsbuff, r17                             ;in LCD-Puffer schreiben
        ldi     r17, ':'                                ;Doppelpunkt
        sts     rsbuff + 1, r17                         ;in LCD-Puffer schreiben
        ldi     r17, ' '                                ;Leerzeichen
        sts     rsbuff + 2, r17                         ;in LCD-Puffer schreiben
;
        ldi     xl, low(hsentb)
        ldi     xh, high(hsentb)                        ;Zeiger auf Sensor-Tabelle
        clr     r17                                     ;H-Byte = 0 fr Tabellenplatzberechnung
        add     xl, r16                                 ;Tabellenplatz berechnen
        adc     xh, r17                                 ;bertrag addieren
        ld      r20, x                                  ;zugehrige Busnummer holen
        cpi     r20, 0xff                               ;Sensor vorhanden?
        brne    ma1710                                  ;ja -> weiter
        rjmp    ma1990                                  ;sonst Fehler
;
ma1710: ldi     xl, low(hsenid)
        ldi     xh, high(hsenid)                        ;Zeiger auf Sensor-ROM-IDs
        add     r16, r16
        add     r16, r16
        add     r16, r16                                ;Sensor-Nummer * 8 als Offset
        add     xl, r16                                 ;Tabellenplatz berechnen
        adc     xh, r17                                 ;bertrag addieren
        adiw    xl, 8                                   ;Zeiger auf Tabellenende
        movw    zl, xl                                  ;berechneten Zeiger auf Sensor-ROM-ID sichern
;
        cli                                             ;Interrupts sperren
        rcall   owres                                   ;1-Wire-Reset senden
        sei                                             ;Interrupts wieder freigeben
        brcc    ma1720                                  ;Presence? ja -> weiter
        rjmp    ma1990                                  ;sonst Fehler
;
ma1720: cli                                             ;Interrupts sperren
        rcall   owmaro                                  ;ROM-Kommando "Match ROM" und ROM-ID senden
        ldi     r19, 0x4e                               ;Funktion "Write Scratchpad"
        rcall   owwrit                                  ;senden
        ldi     r19, 0x00                               ;Parameter "Page 0"
        rcall   owwrit                                  ;senden
        ldi     r19, 0x0f                               ;Konfigurationsbyte mit Bit3 = 1 (AD auf VDD setzen, Betriebsspannung messen)
        rcall   owwrit                                  ;senden
;
        sei                                             ;Interrupts wieder freigeben
        rcall   w001ms                                  ;1ms warten
        cli                                             ;Interrupts sperren
        rcall   owres                                   ;1-Wire-Reset senden
        sei                                             ;Interrupts wieder freigeben
        brcc    ma1730                                  ;Presence? ja -> weiter
        rjmp    ma1990                                  ;sonst Fehler
;
ma1730: movw    xl, zl                                  ;gesicherten Zeiger auf Sensor-ROM-ID holen
        cli                                             ;Interrupts sperren
        rcall   owmaro                                  ;ROM-Kommando "Match ROM" und ROM-ID senden
        ldi     r19, 0xb4                               ;Funktion "Convert V"
        rcall   owwrit                                  ;senden
        sei                                             ;Interrupts wieder freigeben
        rcall   w010ms                                  ;10ms warten bis AD-Conversion abgeschlossen ist
        cli                                             ;Interrupts sperren
        rcall   owres                                   ;1-Wire-Reset senden
        sei                                             ;Interrupts wieder freigeben
        brcc    ma1740                                  ;Presence? ja -> weiter
        rjmp    ma1990                                  ;sonst Fehler
;
ma1740: movw    xl, zl                                  ;gesicherten Zeiger auf Sensor-ROM-ID holen
        cli                                             ;Interrupts sperren
        rcall   owmaro                                  ;ROM-Kommando "Match ROM" und ROM-ID senden
        ldi     r19, 0xb8                               ;Funktion "Recall Memory"
        rcall   owwrit                                  ;senden
        ldi     r19, 0x00                               ;Parameter "Page 0"
        rcall   owwrit                                  ;senden
;
        sei                                             ;Interrupts wieder freigeben
        rcall   w001ms                                  ;1ms warten
        cli                                             ;Interrupts sperren
        rcall   owres                                   ;1-Wire-Reset senden
        sei                                             ;Interrupts wieder freigeben
        brcc    ma1760                                  ;Presence? ja -> weiter
        rjmp    ma1990                                  ;sonst Fehler
;
ma1760: movw    xl, zl                                  ;gesicherten Zeiger auf Sensor-ROM-ID holen
        cli                                             ;Interrupts sperren
        rcall   owmaro                                  ;ROM-Kommando "Match ROM" und ROM-ID senden
        ldi     r19, 0xbe                               ;Funktion "Read Scratchpad"
        rcall   owwrit                                  ;senden
        ldi     r19, 0x00                               ;Parameter "Page 0"
        rcall   owwrit                                  ;senden
;
        ldi     xl, low(owbuff + 8)
        ldi     xh, high(owbuff + 8)                    ;Zeiger auf 1-Wire-Pufferende
        ldi     r21, 8                                  ;8 Bytes Scratchpad einlesen
ma1780: rcall   owread                                  ;Byte lesen
        st      -x, r19                                 ;Byte in Puffer speichern
        dec     r21                                     ;alle Bytes gelesen?
        brne    ma1780                                  ;nein -> Schleife
        rcall   owread                                  ;CRC-Byte lesen
;
        sei                                             ;Interrupts wieder freigeben
        mov     r21, r19                                ;gelesenen CRC-Wert sichern
        rcall   crc                                     ;CRC vom Scratchpad ermitteln
        cp      r17, r21                                ;war CRC-Wert ok?
        brne    ma1790                                  ;nein -> Fehler
        lds     r16, owbuff + 7                         ;zustzlich Status/Konfigurations-Byte holen
        andi    r16, 0x70                               ;alle Konversionen abgeschlossen?
        breq    ma1800                                  ;ja -> weiter
ma1790: rjmp    ma1990                                  ;sonst Fehler
;
; Luftfeuchtigkeit (Temperatur): Gelesenen Temperaturwert im Puffer speichern und Mittelwert berechnen
;
ma1800: ldi     zl, low(htbuff)                         ;zunchst die letzten 7 Werte im Puffer verschieben
        ldi     zh, high(htbuff)                        ;Zeiger auf Puffer fr Temperaturdaten der Luftfeuchtigkeitssensoren
        mov     r18, r22                                ;korrigierte Sensornummer kopieren
        swap    r18                                     ;Sensornummer * 16 (entspricht dem ersten Puffer-Platz des Sensors)
        clr     r19                                     ;High-Byte = 0 und spteres Flag fr Erkennung des leeren Puffers
        add     zl, r18
        adc     zh, r19                                 ;Puffer-Position berechen
        movw    xl, zl                                  ;Puffer-Position sichern
;
        ldi     r18, 7                                  ;7 Eintrge verschieben
ma1810: ldd     r16, z + 2                              ;Eintrag vom Puffer-Platz n + 1 holen (Low-Byte)
        ldd     r17, z + 3                              ;Eintrag vom Puffer-Platz n + 1 holen (High-Byte)
        st      z+, r16                                 ;auf Puffer-Platz n ablegen (Low)
        st      z+, r17                                 ;auf Puffer-Platz n ablegen (High)
        or      r19, r16                                ;Low-Wert mit Flag verknpfen
        or      r19, r17                                ;High-Wert mit Flag verknpfen
        dec     r18                                     ;alle Eintrge verschoben?
        brne    ma1810                                  ;nein -> Schleife
;
        lds     r16, owbuff + 6                         ;Nachkommastellen des Temperaturwertes (Low-Byte) holen
        lds     r17, owbuff + 5                         ;ganzzahligen Temperaturwert (High-Byte) holen
        asr     r17
        ror     r16
        asr     r17
        ror     r16
        asr     r17
        ror     r16                                     ;Sensorwert auf richtige Bitposition verschieben, 3 Stellen nach rechts
;
        st      z+, r16                                 ;neuen Wert auf Puffer-Platz 7 speichern (Low)
        st      z+, r17                                 ;neuen Wert auf Puffer-Platz 7 speichern (High)
        tst     r19                                     ;war Puffer leer?
        brne    ma1830                                  ;nein -> fertig
;
        movw    zl, xl                                  ;sonst gesicherte Puffer-Position holen und Puffer komplett mit dem ersten Sensorwert fllen
        ldi     r18, 7                                  ;7 Eintrge schreiben
ma1820: st      z+, r16                                 ;auf Puffer-Platz n ablegen (Low)
        st      z+, r17                                 ;auf Puffer-Platz n ablegen (High)
        dec     r18                                     ;alle Eintrge verschoben?
        brne    ma1820                                  ;nein -> Schleife
;
ma1830: movw    zl, xl                                  ;gesicherte Puffer-Position holen
        clr     xl
        clr     xh                                      ;Anfangswert fr Mittelwertberechnung = 0 setzen
        ldi     r18, 8                                  ;8 Sensorwerte addieren
ma1840: ld      r16, z+                                 ;Pufferplatz lesen (Low)
        ld      r17, z+                                 ;Pufferplatz lesen (High)
        add     xl, r16                                 ;Sensorwert addieren (Low)
        adc     xh, r17                                 ;Sensorwert addieren (High)
        dec     r18                                     ;alle Sensorwerte addiert?
        brne    ma1840                                  ;nein -> Schleife
;
        ldi     zl, low(htmval)
        ldi     zh, high(htmval)                        ;Zeiger auf Temperaturwerte fr Temp-Kompensation der Feuchtesensoren
        clr     r16                                     ;H-Byte = 0
        add     zl, r22                                 ;korrigierte Sensornummer addieren
        adc     zh, r16                                 ;bertrag addieren
        mov     r17, xh                                 ;ganzzahligen Temperaturwert kopieren
        sbrc    xl, 7                                   ;Runden erforderlich (Nachkommastelle > 0,5)? nein -> weiter
        inc     r17                                     ;sonst Temperatur aufrunden
        st      z, r17                                  ;ganzzahligen Wert speichern (Temp-Kompensation der Feuchtesensoren)
;
        asr     xh                                      ;Summenwert durch 8 dividieren (3 mal nach rechts verschieben)
        ror     xl                                      ;Summenwert nach rechts verschieben (/ 2)
        asr     xh
        ror     xl                                      ;Summenwert nach rechts verschieben (/ 2)
        asr     xh
        ror     xl                                      ;Summenwert nach rechts verschieben (/ 2)
        brcc    ma1850                                  ;Runden erforderlich? nein -> weiter
        adiw    xl, 1                                   ;sonst 1 addieren
;
; Luftfeuchtigkeit (Temperatur): Temperaturmittelwert aufbereiten, Umrechnung des 13-Bit Binrwertes in einen Dezimalwert in Zehntel Grad nach folgender Formel: Temp = Temperaturwert * 5 / 16
;
ma1850: mov     r20, xh                                 ;Vorzeichen sichern
        bst     xh, 7                                   ;Temperaturwert negativ?
        brtc    ma1860                                  ;nein -> weiter
        clr     r16                                     ;sonst
        clr     r17                                     ;Wert 0x10000 zur Wandlung ins Zweierkomplement laden
        sub     r16, xl                                 ;Low-Wert subtrahieren
        sbc     r17, xh                                 ;High-Wert mit bertrag subtrahieren
        movw    xl, r16                                 ;ins Ergebnis-Register kopieren
ma1860: movw    r16, xl                                 ;Temperaturwert kopieren
        add     xl, xl
        adc     xh, xh                                  ;Temperaturwert verdoppeln
        add     xl, xl
        adc     xh, xh                                  ;nochmals verdoppeln (* 4)
        add     xl, r16
        adc     xh, r17                                 ;und Originalwert addieren (* 5)
        asr     xh
        ror     xl                                      ;Wert / 2
        asr     xh
        ror     xl                                      ;Wert / 2 (/ 4)
        asr     xh
        ror     xl                                      ;Wert / 2 (/ 8)
        asr     xh
        ror     xl                                      ;Wert / 2 (/ 16)
        brcc    ma1870                                  ;Runden erforderlich? nein -> weiter
        adiw    xl, 1                                   ;sonst Ergebnis aufrunden
;
; Luftfeuchtigkeit (Temperatur): Ergebnis in Dezimal-ASCII umwandeln und aufbereiten
;
ma1870: mov     r21, xl                                 ;Hex-Wert Low kopieren
        or      r21, xh                                 ;und mit Hex-Wert High verknpfen
        rcall   dezasc                                  ;in Dezimal-ASCII umwandeln
        cpi     r19, '0'                                ;ASCII Stelle 1 = 0?
        brne    ma1880                                  ;nein -> weiter
        ldi     r19, ' '                                ;sonst durch Leerzeichen ersetzen
        cpi     r18, '0'                                ;hchste Ziffer = 0?
        brne    ma1880                                  ;nein -> berspringen
        ldi     r18, ' '                                ;sonst durch Leerzeichen ersetzen
ma1880: bst     r20, 7                                  ;ist Temperaturwert negativ?
        brtc    ma1900                                  ;nein -> berspringen
        tst     r21                                     ;ist der Temperaturwert = 0?
        breq    ma1900                                  ;ja -> berspringen (Minuszeichen vor 0 vermeiden)
        cpi     r18, ' '                                ;Stelle 2 = Leerzeichen?
        breq    ma1890                                  ;ja -> durch Minus ersetzen
        ldi     r19, '-'                                ;sonst Stelle 1 = Minuszeichen
        rjmp    ma1900
ma1890: ldi     r18, '-'                                ;Stelle 2 = Minuszeichen
;
; Luftfeuchtigkeit (Temperatur): Ergebnis in RS-232 Puffer und Wertetabelle legen
;
ma1900: sts     rsbuff + 2, r19                         ;ASCII Stelle 1 in RS232-Puffer
        sts     rsbuff + 3, r18                         ;ASCII Stelle 2 in RS232-Puffer
        sts     rsbuff + 4, r17                         ;ASCII Stelle 3 in RS232-Puffer
        ldi     r20, '.'                                ;Dezimalpunkt
        sts     rsbuff + 5, r20                         ;in RS232-Puffer schreiben
        sts     rsbuff + 6, r16                         ;ASCII Stelle 4 in RS232-Puffer
;
        mov     r21, r22                                ;korrigierte Sensornummer kopieren
        add     r21, r21
        add     r21, r21
        add     r21, r22                                ;Adress-Offset von Wertetabelle berechnen (Sensornummer * 5)
        clr     r23                                     ;H-Byte = 0
        ldi     zl, low(senval + 5)
        ldi     zh, high(senval + 5)                    ;Zeiger auf Sensorwerte (Temperaturen)
        add     zl, r21
        adc     zh, r23                                 ;Tabellenplatz berechnen
        st      z+, r19                                 ;ASCII-Stelle 1 in Tabelle
        st      z+, r18                                 ;ASCII-Stelle 2 in Tabelle
        st      z+, r17                                 ;ASCII-Stelle 3 in Tabelle
        ldi     r20, ','                                ;Komma
        st      z+, r20                                 ;in Tabelle schreiben
        st      z+, r16                                 ;ASCII-Stelle 4 in Tabelle
        sbr     flags, 0b00000010                       ;nderungs-Flag setzen
;
; Luftfeuchtigkeit (Temperatur): Daten ber RS-232 senden
;
        ldi     xl, low(rsbuff)
        ldi     xh, high(rsbuff)                        ;Zeiger auf RS232-Puffer
        ldi     r17, 7                                  ;7 Zeichen senden
ma1910: ld      r16, x+                                 ;Zeichen holen
        rcall   rssend                                  ;Zeichen senden
        dec     r17                                     ;alle Zeichen gesendet?
        brne    ma1910                                  ;nein -> Schleife
        ldi     r16, cr                                 ;CR-Code laden
        rcall   rssend                                  ;und senden
;
; Luftfeuchtigkeit: Betriebsspannungswert des Sensors fr sptere Feuchtigkeitsberechnung in Puffer speichern
;
        ldi     zl, low(h1buff)                         ;zunchst die letzten 7 Werte im Puffer 1 verschieben
        ldi     zh, high(h1buff)                        ;Zeiger auf Puffer fr Feuchtigkeitssensordaten 1 (Betriebsspannung)
        mov     r18, r22                                ;korrigierte Sensornummer kopieren
        swap    r18                                     ;Sensornummer * 16 (entspricht dem ersten Puffer-Platz des Sensors)
        clr     r19                                     ;High-Byte = 0 und spteres Flag fr Erkennung des leeren Puffers
        add     zl, r18
        adc     zh, r19                                 ;Puffer-Position berechen
        movw    xl, zl                                  ;Puffer-Position sichern
;
        ldi     r18, 7                                  ;7 Eintrge verschieben
ma1920: ldd     r16, z + 2                              ;Eintrag vom Puffer-Platz n + 1 holen (Low-Byte)
        ldd     r17, z + 3                              ;Eintrag vom Puffer-Platz n + 1 holen (High-Byte)
        st      z+, r16                                 ;auf Puffer-Platz n ablegen (Low)
        st      z+, r17                                 ;auf Puffer-Platz n ablegen (High)
        or      r19, r16                                ;Low-Wert mit Flag verknpfen
        or      r19, r17                                ;High-Wert mit Flag verknpfen
        dec     r18                                     ;alle Eintrge verschoben?
        brne    ma1920                                  ;nein -> Schleife
;
        lds     r16, owbuff + 4                         ;ADC-Wert der Feuchtigkeitssensor-Betriebsspannung (Low) holen
        lds     r17, owbuff + 3                         ;ADC-Wert der Feuchtigkeitssensor-Betriebsspannung (High) holen
        st      z+, r16                                 ;neuen Wert auf Puffer-Platz 7 speichern (Low)
        st      z+, r17                                 ;neuen Wert auf Puffer-Platz 7 speichern (High)
        tst     r19                                     ;war Puffer leer?
        brne    ma1940                                  ;nein -> fertig
;
        movw    zl, xl                                  ;sonst gesicherte Puffer-Position holen und Puffer komplett mit dem
        ldi     r18, 7                                  ;ersten Sensorwert fllen, 7 Eintrge schreiben
ma1930: st      z+, r16                                 ;auf Puffer-Platz n ablegen (Low)
        st      z+, r17                                 ;auf Puffer-Platz n ablegen (High)
        dec     r18                                     ;alle Eintrge verschoben?
        brne    ma1930                                  ;nein -> Schleife
;
ma1940: rjmp    ma2000
;
; Luftfeuchtigkeit (Temperatur): Fehler aufgetreten - Wert in Tabelle lschen
;
ma1990: mov     r21, r22                                ;korrigierte Sensornummer holen
        add     r21, r21
        add     r21, r21
        add     r21, r22                                ;Adress-Offset von Wertetabelle berechnen (Sensornummer * 5)
        clr     r23                                     ;H-Byte = 0
        ldi     zl, low(senval + 5)
        ldi     zh, high(senval + 5)                    ;Zeiger auf Sensorwerte (Temperaturen)
        add     zl, r21                                 ;Tabellenplatz
        adc     zh, r23                                 ;berechnen
        ldi     r20, ' '                                ;Leerzeichen
        st      z+, r20                                 ;ASCII-Stelle 1 in Tabelle
        st      z+, r20                                 ;ASCII-Stelle 2 in Tabelle
        st      z+, r20                                 ;ASCII-Stelle 3 in Tabelle
        st      z+, r20                                 ;Trennzeichen in Tabelle
        st      z+, r20                                 ;ASCII-Stelle 4 in Tabelle
        sbr     flags, 0b00000010                       ;nderungs-Flag setzen
        rjmp    ma2000
;
; Luftfeuchtigkeitssensoren bearbeiten - Luftfeuchtigkeit
; Luftfeuchtigkeit: ADC auf VAD umschalten und Ausgangsspannung des Luftfeuchtigkeitssensors lesen
;
ma1a00: subi    r16, 29                                 ;Sensornummer korrigieren (29-31 -> 0-2)
        mov     r17, r16                                ;Sensor-Nummer kopieren
        mov     r22, r16                                ;Sensor-Nummer sichern
        subi    r17, -'x'                               ;Sensor-Kennung generieren (x, y, z)
        sts     rsbuff, r17                             ;in LCD-Puffer schreiben
        ldi     r17, ':'                                ;Doppelpunkt
        sts     rsbuff + 1, r17                         ;in LCD-Puffer schreiben
        ldi     r17, ' '                                ;Leerzeichen
        sts     rsbuff + 2, r17                         ;in LCD-Puffer schreiben (Position bleibt immer leer)
        sts     rsbuff + 3, r17                         ;in LCD-Puffer schreiben (Position bleibt immer leer)
;
        ldi     xl, low(hsentb)
        ldi     xh, high(hsentb)                        ;Zeiger auf Sensor-Tabelle
        clr     r17                                     ;H-Byte = 0 fr Tabellenplatzberechnung
        add     xl, r16                                 ;Tabellenplatz berechnen
        adc     xh, r17                                 ;bertrag addieren
        ld      r20, x                                  ;zugehrige Busnummer holen
        cpi     r20, 0xff                               ;Sensor vorhanden?
        brne    ma1a10                                  ;ja -> weiter
        rjmp    ma1d00                                  ;sonst Fehler
;
ma1a10: ldi     xl, low(hsenid)
        ldi     xh, high(hsenid)                        ;Zeiger auf Sensor-ROM-IDs
        add     r16, r16
        add     r16, r16
        add     r16, r16                                ;Sensor-Nummer * 8 als Offset
        add     xl, r16                                 ;Tabellenplatz berechnen
        adc     xh, r17                                 ;bertrag addieren
        adiw    xl, 8                                   ;Zeiger auf Tabellenende
        movw    zl, xl                                  ;berechneten Zeiger auf Sensor-ROM-ID sichern
;
        cli                                             ;Interrupts sperren
        rcall   owres                                   ;1-Wire-Reset senden
        sei                                             ;Interrupts wieder freigeben
        brcc    ma1a20                                  ;Presence? ja -> weiter
        rjmp    ma1d00                                  ;sonst Fehler
;
ma1a20: cli                                             ;Interrupts sperren
        rcall   owmaro                                  ;ROM-Kommando "Match ROM" und ROM-ID senden
        ldi     r19, 0x4e                               ;Funktion "Write Scratchpad"
        rcall   owwrit                                  ;senden
        ldi     r19, 0x00                               ;Parameter "Page 0"
        rcall   owwrit                                  ;senden
        ldi     r19, 0x07                               ;Konfigurationsbyte mit Bit3 = 0 (AD auf VAD setzen, Spannung vom Sensor messen)
        rcall   owwrit                                  ;senden
;
        sei                                             ;Interrupts wieder freigeben
        cli                                             ;Interrupts sperren
        rcall   owres                                   ;1-Wire-Reset senden
        sei                                             ;Interrupts wieder freigeben
        brcc    ma1a40                                  ;Presence? ja -> weiter
        rjmp    ma1d00                                  ;sonst Fehler
;
ma1a40: movw    xl, zl                                  ;gesicherten Zeiger auf Sensor-ROM-ID holen
        cli                                             ;Interrupts sperren
        rcall   owmaro                                  ;ROM-Kommando "Match ROM" und ROM-ID senden
        sei                                             ;Interrupts wieder freigeben
        rcall   w001ms                                  ;1ms auf Stabilisierung der Spannung warten
        cli                                             ;Interrupts sperren
        ldi     r19, 0xb4                               ;Funktion "Convert V"
        rcall   owwrit                                  ;senden
;
        sei                                             ;Interrupts wieder freigeben
        rcall   w010ms                                  ;10ms warten bis AD-Conversion abgeschlossen ist
        cli                                             ;Interrupts sperren
        rcall   owres                                   ;1-Wire-Reset senden
        sei                                             ;Interrupts wieder freigeben
        brcc    ma1a60                                  ;Presence? ja -> weiter
        rjmp    ma1d00                                  ;sonst Fehler
;
ma1a60: movw    xl, zl                                  ;gesicherten Zeiger auf Sensor-ROM-ID holen
        cli                                             ;Interrupts sperren
        rcall   owmaro                                  ;ROM-Kommando "Match ROM" und ROM-ID senden
        ldi     r19, 0xb8                               ;Funktion "Recall Memory"
        rcall   owwrit                                  ;senden
        ldi     r19, 0x00                               ;Parameter "Page 0"
        rcall   owwrit                                  ;senden
;
        sei                                             ;Interrupts wieder freigeben
        cli                                             ;Interrupts sperren
        rcall   owres                                   ;1-Wire-Reset senden
        sei                                             ;Interrupts wieder freigeben
        brcc    ma1a80                                  ;Presence? ja -> weiter
        rjmp    ma1d00                                  ;sonst Fehler
;
ma1a80: movw    xl, zl                                  ;gesicherten Zeiger auf Sensor-ROM-ID holen
        cli                                             ;Interrupts sperren
        rcall   owmaro                                  ;ROM-Kommando "Match ROM" und ROM-ID senden
        ldi     r19, 0xbe                               ;Funktion "Read Scratchpad"
        rcall   owwrit                                  ;senden
        ldi     r19, 0x00                               ;Parameter "Page 0"
        rcall   owwrit                                  ;senden
;
        ldi     xl, low(owbuff + 8)
        ldi     xh, high(owbuff + 8)                    ;Zeiger auf 1-Wire-Pufferende
        ldi     r21, 8                                  ;8 Bytes Scratchpad einlesen
ma1b00: rcall   owread                                  ;Byte lesen
        st      -x, r19                                 ;Byte in Puffer speichern
        dec     r21                                     ;alle Bytes gelesen?
        brne    ma1b00                                  ;nein -> Schleife
        rcall   owread                                  ;CRC-Byte lesen
;
        sei                                             ;Interrupts wieder freigeben
        mov     r21, r19                                ;gelesenen CRC-Wert sichern
        rcall   crc                                     ;CRC vom Scratchpad ermitteln
        cp      r17, r21                                ;war CRC-Wert ok?
        brne    ma1b10                                  ;nein -> Fehler
        lds     r16, owbuff + 7                         ;zustzlich Status/Konfigurations-Byte holen
        andi    r16, 0x70                               ;alle Konversionen abgeschlossen?
        breq    ma1b20                                  ;ja -> weiter
ma1b10: rjmp    ma1d00                                  ;sonst Fehler
;
; Luftfeuchtigkeit: Ausgangsspannungswert des Sensors fr Feuchtigkeitsberechnung in Puffer speichern
;
ma1b20: ldi     zl, low(h2buff)                         ;zunchst die letzten 7 Werte im Puffer 2 verschieben
        ldi     zh, high(h2buff)                        ;Zeiger auf Puffer fr Feuchtigkeitssensordaten 2 (Ausgangsspannung)
        mov     r18, r22                                ;korrigierte Sensornummer kopieren
        swap    r18                                     ;Sensornummer * 16 (entspricht dem ersten Puffer-Platz des Sensors)
        clr     r19                                     ;High-Byte = 0 und spteres Flag fr Erkennung des leeren Puffers
        add     zl, r18
        adc     zh, r19                                 ;Puffer-Position berechen
        movw    xl, zl                                  ;Puffer-Position sichern
;
        ldi     r18, 7                                  ;7 Eintrge verschieben
ma1b30: ldd     r16, z + 2                              ;Eintrag vom Puffer-Platz n + 1 holen (Low-Byte)
        ldd     r17, z + 3                              ;Eintrag vom Puffer-Platz n + 1 holen (High-Byte)
        st      z+, r16                                 ;auf Puffer-Platz n ablegen (Low)
        st      z+, r17                                 ;auf Puffer-Platz n ablegen (High)
        or      r19, r16                                ;Low-Wert mit Flag verknpfen
        or      r19, r17                                ;High-Wert mit Flag verknpfen
        dec     r18                                     ;alle Eintrge verschoben?
        brne    ma1b30                                  ;nein -> Schleife
;
        lds     r16, owbuff + 4                         ;ADC-Wert der Feuchtigkeitssensor-Ausgangsspannung (Low) holen
        lds     r17, owbuff + 3                         ;ADC-Wert der Feuchtigkeitssensor-Ausgangsspannung (High) holen
        st      z+, r16                                 ;neuen Wert auf Puffer-Platz 7 speichern (Low)
        st      z+, r17                                 ;neuen Wert auf Puffer-Platz 7 speichern (High)
        tst     r19                                     ;war Puffer leer?
        brne    ma1b50                                  ;nein -> fertig
;
        movw    zl, xl                                  ;sonst gesicherte Puffer-Position holen und Puffer komplett mit dem
        ldi     r18, 7                                  ;ersten Sensorwert fllen, 7 Eintrge schreiben
ma1b40: st      z+, r16                                 ;auf Puffer-Platz n ablegen (Low)
        st      z+, r17                                 ;auf Puffer-Platz n ablegen (High)
        dec     r18                                     ;alle Eintrge verschoben?
        brne    ma1b40                                  ;nein -> Schleife
;
; Luftfeuchtigkeit: Umrechnen der Sensormittelwerte von Ausgangsspannung und Betriebsspannung nach der Formel des Datenblattes des HIH-4000:
;
; Vout = Vs * (0.0062 * RH + 0.16)                      Vout = Sensorausgangsspannung, Vs = Betriebsspannung, RH = Relative Feuchte (%)
;
; Wir haben die Messwerte Vout und Vs und wollen daraus RH ermitteln:
;
; RH = (Vout / Vs - 0.16) / 0.0062                      Modifikation der Formel zu Vereinfachung der Berechnung:
;
; RH = (Vout / Vs) / 0.0062 - 0.16 / 0.0062             Weitere Vereinfachung:
;
; RH = (Vout / Vs) / 0.0062 - 25,8
;
; Nochmals Vereinfachung: Division durch 0.0062 wird durch eine Multiplikation mit 1 / 0.0062 = 161,3 ersetzt:
;
; RH = Vout / Vs * 161,3 - 25,8
;
; Rechnen mit den zehnfachen Werten, um mit ganzen Zahlen auszukommen und ein Ergebnis in Zehntel % zu erhalten:
;
; RH = Vout / Vs * 1613 - 258
; RH = Vout * 1613 / Vs - 258
; 
ma1b50: ldi     zl, low(h2buff)                         ;zunchst Mittelwert der 8 Sensorwerte im Puffer ermitteln (Vout)
        ldi     zh, high(h2buff)                        ;Zeiger auf Puffer fr Feuchtigkeitssensordaten 2 (Ausgangsspannung)
        mov     r18, r22                                ;korrigierte Sensornummer kopieren
        swap    r18                                     ;Sensornummer * 16 (entspricht dem ersten Puffer-Platz des Sensors)
        clr     r19                                     ;High-Byte = 0
        add     zl, r18
        adc     zh, r19                                 ;Puffer-Position berechen
;
        clr     r4
        clr     r5                                      ;Anfangswert fr Mittelwertberechnung = 0 setzen
        ldi     r18, 8                                  ;8 Sensorwerte addieren
ma1b60: ld      r16, z+                                 ;Pufferplatz lesen (Low)
        ld      r17, z+                                 ;Pufferplatz lesen (High)
        add     r4, r16                                 ;Sensorwert addieren (Low)
        adc     r5, r17                                 ;Sensorwert addieren (High)
        dec     r18                                     ;alle Sensorwerte addiert?
        brne    ma1b60                                  ;nein -> Schleife
        lsr     r5                                      ;Summenwert durch 8 dividieren (3 mal nach rechts verschieben)
        ror     r4                                      ;Summenwert nach rechts verschieben (/ 2)
        lsr     r5
        ror     r4                                      ;Summenwert nach rechts verschieben (/ 2)
        lsr     r5
        ror     r4                                      ;Summenwert nach rechts verschieben (/ 2)
        brcc    ma1b70                                  ;Runden erforderlich? nein -> weiter
        ldi     r16, 1                                  ;sonst 1 addieren
        add     r4, r16                                 ;Low-Byte addieren
        adc     r5, r18                                 ;High-Byte (bertrag) addieren
;
ma1b70: ldi     r16, low(500)                           ;prfen, ob die Ausgangsspannung kleiner als 5V ist
        ldi     r17, high(500)                          ;Vergleichswert 500 laden (entspricht 5V)
        cp      r4, r16
        cpc     r5, r17                                 ;ist die Ausgangsspannung kleiner als 5V?                         
        brcs    ma1b80                                  ;ja -> weiter
        rjmp    ma1d00                                  ;sonst Fehler
;
ma1b80: ldi     r16, low(50)                            ;prfen, ob die Ausgangsspannung grer als 0,5V ist
        ldi     r17, high(50)                           ;Vergleichswert 50 laden (entspricht 0,5V)
        cp      r4, r16
        cpc     r5, r17                                 ;ist die Ausgangsspannung grer als 0,5V?                         
        brcc    ma1b90                                  ;ja -> weiter
        rjmp    ma1d00                                  ;sonst Fehler
;
ma1b90: ldi     r20, low(1613)                          ;Konstante laden (Low)
        ldi     r21, high(1613)                         ;und mit Mittelwert der Sensor-Ausgangsspannung multiplizieren
        rcall   umul16                                  ;Multiplikation r21:r20 * r5:r4 = r19:r16
        movw    r2, r16                                 ;Zwischenergebnis Byte 0 und 1 sichern
        mov     r6, r18                                 ;Zwischenergebnis Byte 2 sichern, Byte 3 wird verworfen (ist immer 0)
;
        ldi     zl, low(h1buff)                         ;Mittelwert der 8 Sensorwerte im Puffer ermitteln (Vs)
        ldi     zh, high(h1buff)                        ;Zeiger auf Puffer fr Feuchtigkeitssensordaten 1 (Betriebsspannung)
        mov     r18, r22                                ;korrigierte Sensornummer kopieren
        swap    r18                                     ;Sensornummer * 16 (entspricht dem ersten Puffer-Platz des Sensors)
        clr     r19                                     ;High-Byte = 0 und spteres Flag fr Erkennung des leeren Puffers
        add     zl, r18
        adc     zh, r19                                 ;Puffer-Position berechen
;
        clr     r4
        clr     r5                                      ;Anfangswert fr Mittelwertberechnung = 0 setzen
        ldi     r18, 8                                  ;8 Sensorwerte addieren
ma1c00: ld      r16, z+                                 ;Pufferplatz lesen (Low)
        ld      r17, z+                                 ;Pufferplatz lesen (High)
        add     r4, r16                                 ;Sensorwert addieren (Low)
        adc     r5, r17                                 ;Sensorwert addieren (High)
        dec     r18                                     ;alle Sensorwerte addiert?
        brne    ma1c00                                  ;nein -> Schleife
        lsr     r5                                      ;Summenwert durch 8 dividieren (3 mal nach rechts verschieben)
        ror     r4                                      ;Summenwert nach rechts verschieben (/ 2)
        lsr     r5
        ror     r4                                      ;Summenwert nach rechts verschieben (/ 2)
        lsr     r5
        ror     r4                                      ;Summenwert nach rechts verschieben (/ 2)
        brcc    ma1c10                                  ;Runden erforderlich? nein -> weiter
        ldi     r16, 1                                  ;sonst 1 addieren
        add     r4, r16                                 ;Low-Byte addieren
        adc     r5, r18                                 ;High-Byte (bertrag) addieren
;
ma1c10: ldi     r16, low(600)                           ;prfen, ob die Betriebsspannung kleiner als 6V ist
        ldi     r17, high(600)                          ;Vergleichswert 600 laden (entspricht 6V)
        cp      r4, r16
        cpc     r5, r17                                 ;ist die Betriebsspannung kleiner als 6V?                         
        brcs    ma1c20                                  ;ja -> weiter
        rjmp    ma1d00                                  ;sonst Fehler
;
ma1c20: ldi     r16, low(200)                           ;prfen, ob die Betriebsspannung grer als 2V ist
        ldi     r17, high(200)                          ;Vergleichswert 200 laden (entspricht 2V)
        cp      r4, r16
        cpc     r5, r17                                 ;ist die Betriebsspannung grer als 2V?                         
        brcc    ma1c30                                  ;ja -> weiter
        rjmp    ma1d00                                  ;sonst Fehler
;
ma1c30: mov     r18, r6                                 ;Zwischenergebnis Byte 2 holen
        movw    r16, r2                                 ;Zwischenergebnis Byte 0 und 1 holen
        rcall   udiv24                                  ;Division r18:r16 / r5:r4 = r18:r16
        ldi     r18, low(258)                           ;zum Schluss noch
        ldi     r19, high(258)                          ;Konstante 258 subtrahieren
        sub     r16, r18                                ;Low-Byte subtrahieren
        sbc     r17, r19                                ;High-Byte subtrahieren
        brcc    ma1c40                                  ;Ergebnis negativ? nein -> weiter
        rjmp    ma1d00                                  ;sonst Fehler
;
; Luftfeuchtigkeit: Temperaturkompensation des ermittelten Wertes nach Datenblatt des HIH-4000:
;
; RHT = RH / (1.0546 - 0.00216 * T)                     RHT = Temperatur-kompensierte Luftfeuchte, RH = Luftfeuchte, T = Temperatur
;
; Rechnen mit den 10000-fachen Werten, um mit ganzen Zahlen auszukommen, die Abweichung durch das Runden ist minimal:
;
; RHT = RH * 10000 / (10546 - 22 * T)
;
ma1c40: movw    r4, r16                                 ;nicht kompensierten Luftfeuchtigkeitswert kopieren
        ldi     r20, low(10000)
        ldi     r21, high(10000)                        ;Konstante 10000 laden
        rcall   umul16                                  ;Multiplikation r21:r20 * r5:r4 = r19:r16
        push    r16                                     ;Zwischenergebnis Byte 0 sichern
        push    r17                                     ;Zwischenergebnis Byte 1 sichern
        push    r18                                     ;Zwischenergebnis Byte 2 sichern, Byte 3 wird verworfen (ist immer 0)
;
        ldi     zl, low(htmval)
        ldi     zh, high(htmval)                        ;Zeiger auf Temperaturwerte fr Temp-Kompensation der Feuchtesensoren
        clr     r16                                     ;H-Byte = 0
        add     zl, r22                                 ;korrigierte Sensornummer addieren
        adc     zh, r16                                 ;bertrag addieren
        ld      r16, z                                  ;gespeicherten Temperaturwert holen
        ldi     r17, 22                                 ;Konstante 22 laden
        mulsu   r16, r17                                ;Multiplikation mit der Temperatur im Zweierkomplement
        ldi     r16, low(10546)
        ldi     r17, high(10546)                        ;Konstante 10546 laden
        sub     r16, r0
        sbc     r17, r1                                 ;vorheriges Multiplikations-Ergebnis subtrahieren
        movw    r4, r16                                 ;Zwischenergebnis kopieren
;
        pop     r18                                     ;Zwischenergebnis Byte 2 holen
        pop     r17                                     ;Zwischenergebnis Byte 1 holen
        pop     r16                                     ;Zwischenergebnis Byte 0 holen
        rcall   udiv24                                  ;Division r18:r16 / r5:r4 = r18:r16
        tst     r18                                     ;ist das hchstwertige Ergebnis-Byte = 0?
        brne    ma1d00                                  ;nein -> Fehler
;
        ldi     r18, low(1001)                          ;prfen, ob das Ergebnis zwischen 0% und 100% liegt
        ldi     r19, high(1001)                         ;Vergleichswert 1001 laden (entspricht 100,1%)
        cp      r16, r18
        cpc     r17, r19                                ;ist das Ergebnis kleiner als 100,1%?
        brcc    ma1d00                                  ;nein -> Fehler
;
        ldi     r18, low(5)
        ldi     r19, high(5)                            ;zunchst Luftfeuchtigkeit durch 5 dividieren
        movw    r4, r18                                 ;in Divisor-Register kopieren
        clr     r18                                     ;hchstwertiges Byte lschen
        rcall   udiv24                                  ;Division r18:r16 / r5:r4 = r18:r16
        asr     r17                                     ;Ergebnis halbieren (High-Byte)
        ror     r16                                     ;Ergebnis halbieren (Low-Byte), ingesamt ergibt sich Division / 10
        brcc    ma1c50                                  ;Runden erforderlich? nein -> weiter
        subi    r16, -1                                 ;sonst 1 addieren (oder 0xffff subtrahieren)
        sbci    r17, -1                                 ;bertrag addieren
;
ma1c50: movw    xl, r16                                 ;Endergebnis kopieren
        rcall   dezasc                                  ;in Dezimal-ASCII umwandeln
        ldi     r19, ' '                                ;Leerzeichen an Tausender-Stelle (ist immer leer)
        cpi     r18, '0'                                ;ist die Hunderter-Stelle = 0?
        brne    ma1c60                                  ;nein -> berspringen
        mov     r18, r19                                ;sonst Hunderter durch Leerzeichen ersetzen
        cpi     r17, '0'                                ;ist auerdem die Zehner-Stelle = 0?
        brne    ma1c60                                  ;nein -> berspringen
        mov     r17, r19                                ;sonst Zehner durch Leerzeichen ersetzen
;
ma1c60: sts     rsbuff + 4, r18                         ;Hunderter-Stelle in RS232-Puffer schreiben
        sts     rsbuff + 5, r17                         ;Zehner-Stelle in RS232-Puffer schreiben
        sts     rsbuff + 6, r16                         ;Einer-Stelle in RS232-Puffer schreiben
;
        mov     r21, r22                                ;korrigierte Sensornummer holen
        add     r21, r21
        add     r21, r21
        add     r21, r22                                ;Adress-Offset von Wertetabelle berechnen (Sensornummer * 5)
        clr     r23                                     ;H-Byte = 0
        ldi     zl, low(senval + 20)
        ldi     zh, high(senval + 20)                   ;Zeiger auf Sensorwerte (Luftfeuchtigkeits-Werte)
        add     zl, r21                                 ;Tabellenplatz
        adc     zh, r23                                 ;berechnen
;
        st      z+, r19                                 ;erste Stelle in Tabelle schreiben (ist immer leer)
        st      z+, r19                                 ;Zausender-Stelle in Tabelle schreiben (ist immer leer)
        st      z+, r18                                 ;Hunderter-Stelle in Tabelle schreiben
        st      z+, r17                                 ;Zehner-Stelle in Tabelle schreiben
        st      z+, r16                                 ;Einer-Stelle in Tabelle schreiben
        sbr     flags, 0b00000010                       ;nderungs-Flag setzen
;
; Luftfeuchtigkeit: Daten ber RS-232 senden
;
        ldi     xl, low(rsbuff)
        ldi     xh, high(rsbuff)                        ;Zeiger auf RS232-Puffer
        ldi     r17, 7                                  ;7 Zeichen senden
ma1c70: ld      r16, x+                                 ;Zeichen holen
        rcall   rssend                                  ;Zeichen senden
        dec     r17                                     ;alle Zeichen gesendet?
        brne    ma1c70                                  ;nein -> Schleife
        ldi     r16, cr                                 ;CR-Code laden
        rcall   rssend                                  ;und senden
        rjmp    ma2000
;
; Luftfeuchtigkeit: Fehler aufgetreten - Wert in Tabelle lschen
;
ma1d00: mov     r21, r22                                ;korrigierte Sensornummer holen
        add     r21, r21
        add     r21, r21
        add     r21, r22                                ;Adress-Offset von Wertetabelle berechnen (Sensornummer * 5)
        clr     r23                                     ;H-Byte = 0
        ldi     zl, low(senval + 20)
        ldi     zh, high(senval + 20)                   ;Zeiger auf Sensorwerte (Luftfeuchtigkeits-Werte)
        add     zl, r21                                 ;Tabellenplatz
        adc     zh, r23                                 ;berechnen
        ldi     r20, ' '                                ;Leerzeichen
        st      z+, r20                                 ;ASCII-Stelle 1 in Tabelle
        st      z+, r20                                 ;ASCII-Stelle 2 in Tabelle
        st      z+, r20                                 ;ASCII-Stelle 3 in Tabelle
        st      z+, r20                                 ;Trennzeichen in Tabelle
        st      z+, r20                                 ;ASCII-Stelle 4 in Tabelle
        sbr     flags, 0b00000010                       ;nderungs-Flag setzen
        rjmp    ma2000
;
; Taster: je nach Mode (Normal oder Einstellungen) verzweigen
;
ma2000: cli                                             ;Interrupts sperren
        lds     r16, semode                             ;Einstellungs-Modus holen
        tst     r16                                     ;Einstellungs-Modus aktiv?
        breq    ma2010                                  ;nein -> weiter
        rjmp    ma2200                                  ;sonst Einstellungen bearbeiten
;
; Taster: Normal-Modus, kurzen Tastendruck abfragen und bearbeiten
;
ma2010: bst     flags, 4                                ;kurzer Tastendruck?
        brtc    ma2020                                  ;nein -> Ende
        bst     flags, 5                                ;kurzer Tastendruck quittiert?
        brts    ma2020                                  ;ja -> Ende
;
        sbr     flags, 0b00100010                       ;Quittieren + nderungs-Flag setzen
        bst     flags, 2                                ;Anzeige-Automatik gesperrt?
        brts    ma2030                                  ;ja -> weiter
        sbr     flags, 0b00000100                       ;Anzeige-Automatik sperren
        clr     r16                                     ;Anzeige-Modus = 0 setzen (Sensoren 1-8)
        sts     dimode, r16                             ;Anzeige-Modus speichern
ma2020: rjmp    ma2100
;
ma2030: lds     r16, dimode                             ;Anzeige-Modus holen
        inc     r16                                     ;erhhen
        sts     dimode, r16                             ;und wieder speichern
        cpi     r16, 1                                  ;Anzeige-Modus = 1? (Sensoren 9-16)
        brne    ma2040                                  ;nein -> weiter
        lds     r17, temsen                             ;sonst Anzahl der Temperatursensoren holen
        cpi     r17, 9                                  ;mehr als 8 Sensoren?
        brcc    ma2040                                  ;ja -> weiter
        inc     r16                                     ;sonst Anzeige-Modus 1 berspringen
        sts     dimode, r16                             ;und speichern
;
ma2040: cpi     r16, 2                                  ;Anzeige-Modus = 2? (Sensoren 17-24)
        brne    ma2050                                  ;nein -> weiter
        lds     r17, temsen                             ;sonst Anzahl der Temperatursensoren holen
        cpi     r17, 17                                 ;mehr als 16 Sensoren?
        brcc    ma2050                                  ;ja -> weiter
        inc     r16                                     ;sonst Anzeige-Modus 2 berspringen
        sts     dimode, r16                             ;und speichern
;
ma2050: cpi     r16, 3                                  ;Anzeige-Modus = 3? (Sensoren 25-28)
        brne    ma2060                                  ;ja -> weiter
        lds     r17, prssen                             ;sonst Anzahl der Luftdrucksensoren holen
        lds     r18, humsen                             ;Anzahl der Luftfeuchtigkeitssensoren holen
        add     r17, r18                                ;Anzahlen addieren
        tst     r17                                     ;sind Luftdruck- und Luftfeuchtigkeitssensoren vorhanden?
        brne    ma2060                                  ;ja -> weiter
        inc     r16                                     ;sonst Anzeige-Modus 3 berspringen
        sts     dimode, r16                             ;und speichern
;
ma2060: cpi     r16, 6                                  ;Modus-Bereich Sensor-ROM-Codes erreicht?
        brcs    ma2100                                  ;nein -> Ende
;
        lds     r17, temsen                             ;Anzahl der Temperatursensoren holen
        lds     r18, humsen                             ;Anzahl der Luftfeuchtigkeitssensoren holen
        add     r17, r18                                ;Gesamtanzahl ermitteln
        lsr     r17                                     ;Sensoranzahl durch 2 teilen
        ror     r18                                     ;bertrag auffangen
        lsr     r17                                     ;Sensoranzahl nochmals durch 2 teilen
        ror     r18                                     ;bertrag auffangen
        subi    r17, -6                                 ;hchsten Anzeige-Modus ermitteln
        andi    r18, 0xc0                               ;bertrag-Bits filtern, weitere Seite erforderlich?
        breq    ma2070                                  ;nein -> berspringen
        inc     r17                                     ;sonst hchsten Anzeige-Modus korrigieren
;
ma2070: cp      r16, r17                                ;hchster Anzeige-Modus erreicht?
        brcs    ma2100                                  ;nein -> Ende
;
        clr     r16                                     ;sonst Anzeige-Modus = 0 setzen
        sts     dimode, r16                             ;und speichern
        cbr     flags, 0b00000100                       ;Anzeige-Automatik freigeben
;
; Taster: Normal-Modus, langen Tastendruck abfragen und bearbeiten
;
ma2100: bst     flags, 6                                ;langer Tastendruck?
        brtc    ma2110                                  ;nein -> Ende
        bst     flags, 7                                ;langer Tastendruck quittiert?
        brts    ma2110                                  ;ja -> Ende
;
        sbr     flags, 0b10000000                       ;langen Tastendruck quittieren
        ldi     r16, 1                                  ;Einstellungs-Modus 1 setzen
        sts     semode, r16                             ;und speichern
        rcall   setinc                                  ;Interrupt-Zhler setzen
ma2110: sei                                             ;Interrupts wieder freigeben
        rjmp    ma3000
;
; Taster: Einstellungs-Modus, langen Tastendruck abfragen und bearbeiten
;
ma2200: bst     flags, 6                                ;langer Tastendruck?
        brtc    ma2210                                  ;nein -> Ende
        bst     flags, 7                                ;langer Tastendruck quittiert?
        brtc    ma2220                                  ;nein -> bearbeiten
;
ma2210: rjmp    ma2500
;
ma2220: sbr     flags, 0b10000000                       ;langen Tastendruck quittieren
        lds     r16, flags2                             ;Flags 2 holen
        cbr     r16, 0b00000001                         ;Merker lschen
        sts     flags2, r16                             ;Flags 2 wieder speichern
        rcall   setinc                                  ;Interrupt-Zhler setzen
        sei                                             ;Interrupts wieder freigeben
;
        lds     r16, semode                             ;Einstellungs-Modus holen
        cpi     r16, 1                                  ;Modus-Einstellung?
        brne    ma2300                                  ;nein -> weiter testen
;
        lds     r18, modflg                             ;sonst Modus-Flag holen
        tst     r18                                     ;Direkt-Modus aktiv?
        brne    ma2230                                  ;ja -> weiter
        ser     r18                                     ;sonst in Direkt-Modus wechseln
        rjmp    ma2240                                  ;Modus speichern
;
ma2230: rcall   savsen                                  ;Sensordaten im EEPROM speichern
        clr     r18                                     ;in Speicher-Modus wechseln
ma2240: sts     modflg, r18                             ;Modus-Flag wieder speichern
        ldi     r16, low(emodfl)
        ldi     r17, high(emodfl)                       ;EERPOM-Adresse Modus-Flag
        rcall   eewrit                                  ;Modus-Flag speichern
        rjmp    ma3000
;
ma2300: cpi     r16, 2                                  ;Sende-Intervall-Einstellung?
        brne    ma2400                                  ;nein -> weiter testen
;
        lds     r18, intval                             ;Sende-Intervall holen
        cpi     r18, 5                                  ;Wert >= 5?
        brcc    ma2310                                  ;ja -> weiter
        ldi     r18, 5                                  ;sonst Wert = 5 setzen
        rjmp    ma2350                                  ;speichern
ma2310: cpi     r18, 10                                 ;Wert >= 10?
        brcc    ma2320                                  ;ja -> weiter
        ldi     r18, 10                                 ;sonst Wert = 10 setzen
        rjmp    ma2350                                  ;speichern
ma2320: cpi     r18, 15                                 ;Wert >= 15?
        brcc    ma2330                                  ;ja -> weiter
        ldi     r18, 15                                 ;sonst Wert = 15 setzen
        rjmp    ma2350                                  ;speichern
ma2330: cpi     r18, 20                                 ;Wert >= 20?
        brcc    ma2340                                  ;ja -> weiter
        ldi     r18, 20                                 ;sonst Wert = 20 setzen
        rjmp    ma2350                                  ;speichern
ma2340: ldi     r18, 2                                  ;Wert = 2 setzen
ma2350: sts     intval, r18                             ;neues Sende-Intervall speichern
        ldi     r16, low(eintva)
        ldi     r17, high(eintva)                       ;EERPOM-Adresse Sende-Intervall
        rcall   eewrit                                  ;Sende-Intervall speichern
        rjmp    ma3000
;
ma2400: cpi     r16, 3                                  ;Alarm-Zeit-Einstellung?
        brne    ma2450                                  ;nein -> weiter
;
        lds     r18, altime                             ;Alarm-Zeit holen
        cpi     r18, 10                                 ;Wert > 10?
        brcc    ma2410                                  ;ja -> weiter
        ldi     r18, 10                                 ;sonst Wert = 10 setzen
        rjmp    ma2440                                  ;speichern
ma2410: cpi     r18, 15                                 ;Wert > 15?
        brcc    ma2420                                  ;ja -> weiter
        ldi     r18, 15                                 ;sonst Wert = 15 setzen
        rjmp    ma2440                                  ;speichern
ma2420: cpi     r18, 20                                 ;Wert > 20?
        brcc    ma2430                                  ;ja -> weiter
        ldi     r18, 20                                 ;sonst Wert = 20 setzen
        rjmp    ma2440                                  ;speichern
ma2430: ldi     r18, 5                                  ;Wert = 5 setzen
ma2440: sts     altime, r18                             ;neue Alarm-Zeit speichern
        ldi     r16, low(ealtim)
        ldi     r17, high(ealtim)                       ;EERPOM-Adresse Alarm-Zeit
        rcall   eewrit                                  ;Alarm-Zeit speichern
        rjmp    ma3000
;
ma2450: lds     r18, lang                               ;Sprach-Einstellung holen
        tst     r18                                     ;Sprache Deutsch?
        brne    ma2460                                  ;ja -> weiter
        ser     r18                                     ;sonst zu Deutsch wechseln
        rjmp    ma2470                                  ;Sprache speichern
;
ma2460: clr     r18                                     ;zu Englisch wechseln
ma2470: sts     lang, r18                               ;Sprache wieder speichern
        ldi     r16, low(elang)
        ldi     r17, high(elang)                        ;EERPOM-Adresse Sprache
        rcall   eewrit                                  ;Sprache speichern
        rjmp    ma3000
;
; Taster: Einstellungs-Modus, kurzen Tastendruck abfragen und bearbeiten; Aktion wird erst nach Loslassen des Tasters ausgelst
;
ma2500: bst     flags, 4                                ;kurzer Tastendruck?
        brtc    ma2510                                  ;nein -> weiter
        bst     flags, 5                                ;kurzer Tastendruck quittiert?
        brts    ma2520                                  ;ja -> Ende
;
        sbr     flags, 0b00100000                       ;Tastendruck quittieren
        lds     r16, flags2                             ;Flags 2 holen
        sbr     r16, 0b00000001                         ;Merker setzen
        sts     flags2, r16                             ;Flags 2 wieder speichern
        rcall   setinc                                  ;Interrupt-Zhler setzen
        rjmp    ma2520                                  ;Ende
;
ma2510: lds     r16, flags2                             ;Flags 2 holen
        bst     r16, 0                                  ;Merker gesetzt?
        brts    ma2530                                  ;ja -> weiter
ma2520: sei                                             ;Interrupts wieder freigeben
        rjmp    ma3000                                  ;Ende
;
ma2530: cbr     r16, 0b00000001                         ;Merker lschen
        sts     flags2, r16                             ;Flags 2 wieder speichern
        sei                                             ;Interrupts wieder freigeben
        lds     r16, semode                             ;Einstellungs-Modus holen
        inc     r16                                     ;Modus erhhen
        sts     semode, r16                             ;und wieder speichern
        cpi     r16, 5                                  ;Bereich berschritten?
        brcs    ma3000                                  ;nein -> Ende
        clr     r16                                     ;Sonst Modus = 0 setzen
        sts     semode, r16                             ;und speichern
        sts     dimode, r16                             ;Anzeige-Modus = 0 setzen
        cli                                             ;Interrupts sperren
        sts     intc16, r16
        sts     intc16 + 1, r16                         ;Interrupt-Zhler lschen
        sei                                             ;Interrupts wieder freigeben
        cbr     flags, 0b00000100                       ;Anzeige-Automatik aktivieren
        sbr     flags, 0b11110010                       ;Tasten quittieren, nderungs-Flag setzen
;
; Automatische Steuerung der Sensoranzeige
;
ma3000: lds     r16, semode                             ;Einstellungsmodus holen
        tst     r16                                     ;Einstellungsmodus aktiv?
        breq    ma3010                                  ;nein -> weiter
        rjmp    ma3200                                  ;sonst Einstellungen bearbeiten
;
ma3010: bst     flags, 1                                ;nderungs-Flag gesetzt?
        brts    ma3020                                  ;ja -> weiter
        rjmp    ma4000                                  ;sonst Ende
;
ma3020: cbr     flags, 0b00000010                       ;nderungs-Flag wieder lschen
        bst     flags, 2                                ;Anzeige-Automatik gesperrt?
        brts    ma3100                                  ;ja -> manuelle Anzeige
;
        lds     r16, intc16                             ;Interrupt-Zhler L holen
        lds     r17, intc16 + 1                         ;Interrupt-Zhler H holen
        or      r16, r17                                ;Interrupt-Zhler = 0?
        breq    ma3030                                  ;ja -> weiter
        rcall   dalarm                                  ;sonst auf Alarmanzeige umschalten
        rjmp    ma4000
;
ma3030: lds     r22, sencnt                             ;aktuelle Sensornummer holen
        tst     r22                                     ;Mess-Phase aktiv (Sensor = 0)?
        brne    ma3040                                  ;nein -> weiter
;
        ldi     r22, 27                                 ;sonst Sensornummer = 27 setzen
        lds     r21, humsen                             ;Anzahl der Luftfeuchtigkeitssensoren holen
        add     r22, r21                                ;addieren
        cpi     r22, 27                                 ;hat sich die Sensornummer gendert?
        brne    ma3080                                  ;ja -> Sensordaten mit richtiger Markierung ausgeben
        ldi     r22, 23                                 ;sonst Sensornummer = 23 setzen
        lds     r21, prssen                             ;Anzahl der Luftdrucksensoren holen
        add     r22, r21                                ;addieren
        cpi     r22, 23                                 ;hat sich die Sensornummer gendert?
        brne    ma3080                                  ;ja -> Sensordaten mit richtiger Markierung ausgeben
        lds     r22, temsen                             ;sonst Anzahl der Temperatursensoren holen
        tst     r22                                     ;sind Sensoren vorhanden?
        brne    ma3040                                  ;ja -> weiter
        ser     r22                                     ;sonst ungltigen Wert setzen
        rjmp    ma3050                                  ;Sensoren 1-8 ausgeben
;
ma3040: dec     r22                                     ;korrigieren
        cpi     r22, 8                                  ;Sensornummer > 8?
        brcc    ma3060                                  ;ja -> berspringen
ma3050: rcall   dsens1                                  ;sonst Sensordaten 1-8 ausgeben
        rjmp    ma4000
;
ma3060: cpi     r22, 16                                 ;Sensornummer > 16?
        brcc    ma3070                                  ;ja -> berspringen
        rcall   dsens2                                  ;sonst Sensordaten 9-16 ausgeben
        rjmp    ma4000
;
ma3070: cpi     r22, 24                                 ;Sensornummer > 24?
        brcc    ma3080                                  ;ja -> berspringen
        rcall   dsens3                                  ;sonst Sensordaten 17-24 ausgeben
        rjmp    ma4000
;
ma3080: rcall   dsens4                                  ;Sensordaten 25-31 ausgeben
        rjmp    ma4000
;
; Manuelle Anzeige
;
ma3100: lds     r16, dimode                             ;Anzeige-Modus holen
        tst     r16                                     ;Anzeige-Modus = 0? (Sensoren 1-8)
        brne    ma3110                                  ;nein -> weiter testen
        rcall   dsens1                                  ;sonst Sensoren 1-8 anzeigen
        rjmp    ma4000
;
ma3110: cpi     r16, 1                                  ;Anzeige-Modus = 1? (Sensoren 9-16)
        brne    ma3120                                  ;nein -> weiter testen
        rcall   dsens2                                  ;sonst Sensoren 9-16 anzeigen
        rjmp    ma4000
;
ma3120: cpi     r16, 2                                  ;Anzeige-Modus = 2? (Sensoren 17-24)
        brne    ma3130                                  ;nein -> weiter testen
        rcall   dsens3                                  ;sonst Sensoren 17-24 anzeigen
        rjmp    ma4000
;
ma3130: cpi     r16, 3                                  ;Anzeige-Modus = 3? (Sensoren 25-28)
        brne    ma3140                                  ;nein -> weiter testen
        rcall   dsens4                                  ;sonst Sensoren 25-28 anzeigen
        rjmp    ma4000
;
ma3140: cpi     r16, 4                                  ;Anzeige-Modus = 4? (Alarme)
        brne    ma3150                                  ;nein -> weiter testen
        rcall   dalarm                                  ;sonst Alarme anzeigen
        rjmp    ma4000
;
ma3150: cpi     r16, 5                                  ;Anzeige-Modus = 5? (Status)
        brne    ma3160                                  ;nein -> ROM-Codes anzeigen
        rcall   dstatu                                  ;sonst Status anzeigen
        rjmp    ma4000
;
ma3160: rcall   drcode                                  ;ROM-Codes anzeigen
        rjmp    ma4000
;
; Einstellungs-Anzeige: Modus
;
ma3200: cbr     flags, 0b00000010                       ;nderungs-Flag lschen
        cpi     r16, 1                                  ;Modus-Einstellung?
        brne    ma3300                                  ;nein -> weiter testen
;
        lds     r18, adrtab                             ;LCD-Adresse Zeile 1
        clr     r17                                     ;LCD in Befehlsmodus setzen
        rcall   lcdw40                                  ;Adresse setzen
        ldi     r20, 20                                 ;Zeichenzhler setzen
        ldi     r19, 11                                 ;Text 11 (Modus:)
        rcall   lcdtxt                                  ;ausgeben
        rcall   lcdlin                                  ;Zeilenende lschen
;
        lds     r18, adrtab + 1                         ;LCD-Adresse Zeile 2
        clr     r17                                     ;LCD in Befehlsmodus setzen
        rcall   lcdw40                                  ;Adresse setzen
        ldi     r20, 20                                 ;Zeichenzhler setzen
        ldi     r19, 6                                  ;Text 6 (Pfeil rechts)
        rcall   lcdtxt                                  ;ausgeben
;
        ldi     r19, 4                                  ;Text 4 (Direkt)
        lds     r16, modflg                             ;Modus-Flag holen
        tst     r16                                     ;Modus Direkt?
        brne    ma3210                                  ;nein -> weiter
        ldi     r19, 5                                  ;sonst Text 5 (Speicher)
ma3210: rcall   lcdtxt                                  ;ausgeben
        ldi     r19, 7                                  ;Text 7 (Pfeil links)
        rcall   lcdtxt                                  ;ausgeben
        rcall   lcdlin                                  ;Zeilenende lschen
        rjmp    ma3700                                  ;Menzeilen ausgeben
;
; Einstellungs-Anzeige: Sende-Intervall
;
ma3300: cpi     r16, 2                                  ;Intervall-Eistellung?
        brne    ma3400                                  ;nein -> weiter testen
;
        lds     r18, adrtab                             ;LCD-Adresse Zeile 1
        clr     r17                                     ;LCD in Befehlsmodus setzen
        rcall   lcdw40                                  ;Adresse setzen
        ldi     r20, 20                                 ;Zeichenzhler setzen
        ldi     r19, 12                                 ;Text 12 (Sende-Intervall:)
        rcall   lcdtxt                                  ;ausgeben
        rcall   lcdlin                                  ;Zeilenende lschen
;
        lds     r18, adrtab + 1                         ;LCD-Adresse Zeile 2
        clr     r17                                     ;LCD in Befehlsmodus setzen
        rcall   lcdw40                                  ;Adresse setzen
        ldi     r20, 20                                 ;Zeichenzhler setzen
        ldi     r19, 6                                  ;Text 6 (Pfeil rechts)
        rcall   lcdtxt                                  ;ausgeben
;
        lds     r16, intval                             ;Sende-Intervall holen
        ldi     r18, 50                                 ;Multiplikator
        mul     r16, r18                                ;Ergebnis in Millisekunden
        rcall   lcddez                                  ;Ergebnis auf LCD ausgeben
        ldi     r19, 16                                 ;Text 16 (ms)
        rcall   lcdtxt                                  ;ausgeben
        ldi     r19, 7                                  ;Text 7 (Pfeil links)
        rcall   lcdtxt                                  ;ausgeben
        rcall   lcdlin                                  ;Zeilenende lschen
        rjmp    ma3700                                  ;Menzeilen ausgeben
;
; Einstellungs-Anzeige: Alarm-Zeit
;
ma3400: cpi     r16, 3                                  ;Alarmzeit-Eistellung?
        brne    ma3500                                  ;nein -> weiter testen
;
        lds     r18, adrtab                             ;LCD-Adresse Zeile 1
        clr     r17                                     ;LCD in Befehlsmodus setzen
        rcall   lcdw40                                  ;Adresse setzen
        ldi     r20, 20                                 ;Zeichenzhler setzen
        ldi     r19, 13                                 ;Text 13 (Alarm-Zeit:)
        rcall   lcdtxt                                  ;ausgeben
        rcall   lcdlin                                  ;Zeilenende lschen
;
        lds     r18, adrtab + 1                         ;LCD-Adresse Zeile 2
        clr     r17                                     ;LCD in Befehlsmodus setzen
        rcall   lcdw40                                  ;Adresse setzen
        ldi     r20, 20                                 ;Zeichenzhler setzen
        ldi     r19, 6                                  ;Text 6 (Pfeil rechts)
        rcall   lcdtxt                                  ;ausgeben
;
        lds     r16, altime                             ;Alarm-Zeit holen
        ldi     r18, 50                                 ;Multiplikator
        mul     r16, r18                                ;Ergebnis in Millisekunden
        rcall   lcddez                                  ;Ergebnis auf LCD ausgeben
        ldi     r19, 16                                 ;Text 16 (ms)
        rcall   lcdtxt                                  ;ausgeben
        ldi     r19, 7                                  ;Text 7 (Pfeil links)
        rcall   lcdtxt                                  ;ausgeben
        rcall   lcdlin                                  ;Zeilenende lschen
        rjmp    ma3700                                  ;Menzeilen ausgeben
;
; Einstellungs-Anzeige: Sprache
;
ma3500: cpi     r16, 4                                  ;Sprach-Einstellung?
        brne    ma4000                                  ;nein -> Ende
;
        lds     r18, adrtab                             ;LCD-Adresse Zeile 1
        clr     r17                                     ;LCD in Befehlsmodus setzen
        rcall   lcdw40                                  ;Adresse setzen
        ldi     r20, 20                                 ;Zeichenzhler setzen
        ldi     r19, 14                                 ;Text 14 (Sprache:)
        rcall   lcdtxt                                  ;ausgeben
        rcall   lcdlin                                  ;Zeilenende lschen
;
        lds     r18, adrtab + 1                         ;LCD-Adresse Zeile 2
        clr     r17                                     ;LCD in Befehlsmodus setzen
        rcall   lcdw40                                  ;Adresse setzen
        ldi     r20, 20                                 ;Zeichenzhler setzen
        ldi     r19, 6                                  ;Text 6 (Pfeil rechts)
        rcall   lcdtxt                                  ;ausgeben
;
        ldi     r19, 15                                 ;Text 15 (Deutsch)
        rcall   lcdtxt                                  ;ausgeben
        ldi     r19, 7                                  ;Text 7 (Pfeil links)
        rcall   lcdtxt                                  ;ausgeben
        rcall   lcdlin                                  ;Zeilenende lschen
;
; Einstellungs-Anzeige: Menzeilen
;
ma3700: lds     r18, adrtab + 2                         ;LCD-Adresse Zeile 3
        clr     r17                                     ;LCD in Befehlsmodus setzen
        rcall   lcdw40                                  ;Adresse setzen
        ldi     r20, 20                                 ;Zeichenzhler setzen
        ldi     r19, 8                                  ;Text 8 (Lang -> ndern)
        rcall   lcdtxt                                  ;ausgeben
        rcall   lcdlin                                  ;Zeilenende lschen
;
        lds     r18, adrtab + 3                         ;LCD-Adresse Zeile 4
        clr     r17                                     ;LCD in Befehlsmodus setzen
        rcall   lcdw40                                  ;Adresse setzen
        ldi     r20, 20                                 ;Zeichenzhler setzen
        ldi     r19, 9                                  ;Text 9 (Kurz -> Weiter)
        lds     r16, semode                             ;Einstellungs-Modus holen
        cpi     r16, 4                                  ;Sprach-Einstellung aktiv?
        brne    ma3710                                  ;nein -> weiter
        ldi     r19, 10                                 ;sonst Text 10 (Kurz -> Ende)
ma3710: rcall   lcdtxt                                  ;ausgeben
        rcall   lcdlin                                  ;Zeilenende lschen
;
; Sonstiges
;
ma4000: rjmp    main
;
; Update-Funktion: Modus-Flag wieder auf "Speicher" setzen
;
update: ldi     r16, low(emodfl)
        ldi     r17, high(emodfl)                       ;EEPROM-Adresse Modus-Flag
        clr     r18                                     ;Modus wieder auf "Speicher" zurcksetzen
        rcall   eewrit                                  ;in EEPROM schreiben
;
; Update-Funktion: Prfen, ob die bei der Update-Suche gefundenen Temperatursensoren bereits vorhanden sind und im positiven Fall aus der Update-Liste lschen
;
        ldi     zl, low(tsent2)
        ldi     zh, high(tsent2)                        ;Zeiger auf Sensortabelle der Update-Suche
        ldi     xl, low(tseni2)
        ldi     xh, high(tseni2)                        ;Zeiger auf ROM-IDs der Temperatursensoren (Update-Suche)
        ldi     r21, 24                                 ;24 Sensoren mit den gespeicherten Sensordaten vergleichen
;
upd010: ld      r16, z                                  ;Eintrag aus Sensortabelle der Update-Suche holen
        cpi     r16, 8                                  ;gltige Busnummer (0-7)?
        brcc    upd060                                  ;nein -> Ende
        movw    r0, xl                                  ;sonst ROM-ID-Adresse des gerade bearbeiteten Sensors sichern
        ldi     yl, low(tsenid)
        ldi     yh, high(tsenid)                        ;Zeiger auf ROM-IDs der gespeicherten Temperatursensoren
        ldi     r20, 24                                 ;mit 24 Sensoren vergleichen
;
upd020: movw    xl, r0                                  ;ROM-ID-Adresse des gerade bearbeiteten Sensors zurckholen
        ldi     r19, 8                                  ;8 Bytes vergleichen
        clr     r18                                     ;Fehlerzhler fr Vergleich lschen
;
upd030: ld      r16, x+                                 ;ein Byte der ROM-ID holen (Update-Suche)
        ld      r17, y+                                 ;ein Byte der gespeicherten ROM-ID holen
        cp      r16, r17                                ;Bytes identisch?
        breq    upd040                                  ;ja -> weiter
        inc     r18                                     ;sonst Fehlerzhler erhhen
upd040: dec     r19                                     ;alle Bytes vergleichen?
        brne    upd030                                  ;nein -> Schleife
        tst     r18                                     ;Fehlerzhler = 0?
        brne    upd050                                  ;nein -> weiter
        ldi     r16, 0xff                               ;sonst Wert 0xff
        st      z, r16                                  ;in Sensortabelle schreiben (Sensor ist bereits installiert)
;
upd050: dec     r20                                     ;wurde mit allen 24 gespeicherten Sensoren vergleichen?
        brne    upd020                                  ;nein -> Schleife
        adiw    zl, 1                                   ;sonst Zeiger auf nchsten Tabelleneintrag
        dec     r21                                     ;wurden alle Sensoren der Update-Suche bearbeitet?
        brne    upd010                                  ;nein -> Schleife
;
; Update-Funktion: neue Temperatursensoren zhlen
;
upd060: ldi     zl, low(tsent2)
        ldi     zh, high(tsent2)                        ;Zeiger auf Sensortabelle der Update-Suche
        ldi     r21, 24                                 ;24 Sensoren bearbeiten
        clr     r20                                     ;Zhler fr neue Sensoren lschen
upd070: ld      r16, z+                                 ;Eintrag aus Sensortabelle der Update-Suche holen
        cpi     r16, 8                                  ;gltige Busnummer (0-7)?
        brcc    upd080                                  ;nein -> weiter
        inc     r20                                     ;sonst neuen Sensor hinzuzhlen
upd080: dec     r21                                     ;alle Sensoren bearbeitet?
        brne    upd070                                  ;nein -> Schleife
        sts     temse2, r20                             ;ermittelte Anzahl der neuen Sensoren speichern
;
; Update-Funktion: Prfen, ob die bei der Update-Suche gefundenen Luftfeuchtigkeitssensoren bereits vorhanden sind und im positiven Fall aus der Update-Liste lschen
;
        ldi     zl, low(hsent2)
        ldi     zh, high(hsent2)                        ;Zeiger auf Sensortabelle der Update-Suche
        ldi     xl, low(hseni2)
        ldi     xh, high(hseni2)                        ;Zeiger auf ROM-IDs der Luftfeuchtigkeitssensoren (Update-Suche)
        ldi     r21, 3                                  ;3 Sensoren mit den gespeicherten Sensordaten vergleichen
;
upd110: ld      r16, z                                  ;Eintrag aus Sensortabelle der Update-Suche holen
        cpi     r16, 8                                  ;gltige Busnummer (0-7)?
        brcc    upd160                                  ;nein -> Ende
        movw    r0, xl                                  ;sonst ROM-ID-Adresse des gerade bearbeiteten Sensors sichern
        ldi     yl, low(hsenid)
        ldi     yh, high(hsenid)                        ;Zeiger auf ROM-IDs der gespeicherten Luftfeuchtigkeitssensoren
        ldi     r20, 3                                  ;mit 3 Sensoren vergleichen
;
upd120: movw    xl, r0                                  ;ROM-ID-Adresse des gerade bearbeiteten Sensors zurckholen
        ldi     r19, 8                                  ;8 Bytes vergleichen
        clr     r18                                     ;Fehlerzhler fr Vergleich lschen
;
upd130: ld      r16, x+                                 ;ein Byte der ROM-ID holen (Update-Suche)
        ld      r17, y+                                 ;ein Byte der gespeicherten ROM-ID holen
        cp      r16, r17                                ;Bytes identisch?
        breq    upd140                                  ;ja -> weiter
        inc     r18                                     ;sonst Fehlerzhler erhhen
upd140: dec     r19                                     ;alle Bytes vergleichen?
        brne    upd130                                  ;nein -> Schleife
        tst     r18                                     ;Fehlerzhler = 0?
        brne    upd150                                  ;nein -> weiter
        ldi     r16, 0xff                               ;sonst Wert 0xff
        st      z, r16                                  ;in Sensortabelle schreiben (Sensor ist bereits installiert)
;
upd150: dec     r20                                     ;wurde mit allen 3 gespeicherten Sensoren vergleichen?
        brne    upd120                                  ;nein -> Schleife
        adiw    zl, 1                                   ;sonst Zeiger auf nchsten Tabelleneintrag
        dec     r21                                     ;wurden alle Sensoren der Update-Suche bearbeitet?
        brne    upd110                                  ;nein -> Schleife
;
; Update-Funktion: neue Luftfeuchtigkeitssensoren zhlen
;
upd160: ldi     zl, low(hsent2)
        ldi     zh, high(hsent2)                        ;Zeiger auf Sensortabelle der Update-Suche
        ldi     r21, 3                                  ;3 Sensoren bearbeiten
        clr     r20                                     ;Zhler fr neue Sensoren lschen
upd170: ld      r16, z+                                 ;Eintrag aus Sensortabelle der Update-Suche holen
        cpi     r16, 8                                  ;gltige Busnummer (0-7)?
        brcc    upd180                                  ;nein -> weiter
        inc     r20                                     ;sonst neuen Sensor hinzuzhlen
upd180: dec     r21                                     ;alle Sensoren bearbeitet?
        brne    upd170                                  ;nein -> Schleife
        sts     humse2, r20                             ;ermittelte Anzahl der neuen Sensoren speichern
;
; Update-Funktion: neue Luftdrucksensoren ermitteln
;
        lds     r16, prsse2                             ;ermittelte Anzahl Luftdrucksensoren holen
        cpi     r16, 1                                  ;Anzahl = 1?
        brne    upd200                                  ;nein -> Ende
        lds     r17, prssen                             ;sonst gespeicherte Anzahl holen
        cpi     r17, 1                                  ;Sensor bereits vorhanden?
        brne    upd200                                  ;nein -> Ende
        clr     r16                                     ;sonst
        sts     prsse2, r16                             ;neue Anzahl = 0 setzen
;
; Update-Funktion: Anzahl der neuen Sensoren auf dem LCD ausgeben und auf langen Tastendruck warten
;
upd200: ldi     r18, 0x01                               ;Kommando "Clear LCD"
        clr     r17                                     ;LCD in Befehlsmodus setzen
        rcall   lcdw40                                  ;LCD lschen
        rcall   w016ms                                  ;16ms warten
;
        ldi     r17, 1                                  ;LCD in Datenmodus setzen
        ldi     r18, 'T'                                ;Zeichen "T" (Temperatursensoren)
        rcall   lcdw40                                  ;ausgeben
;
        lds     xl, temse2                              ;Anzahl der neuen Temperatursensoren holen
        clr     xh                                      ;High-Byte lschen
        rcall   dezasc                                  ;in Dezimal-ASII umwandeln
        mov     r19, r16                                ;Einer-Stelle sichern
        cpi     r17, '0'                                ;Zehner-Stelle = 0?
        breq    upd210                                  ;ja -> nur Einer ausgeben
        mov     r18, r17                                ;sonst Zehner-Stelle holen
        rcall   lcdw40                                  ;ausgeben
upd210: mov     r18, r19                                ;Einer-Stelle holen
        rcall   lcdw40                                  ;ausgeben
;
        ldi     r18, ' '                                ;Leerzeichen
        rcall   lcdw40                                  ;ausgeben
        ldi     r18, 'P'                                ;Zeichen "P" (Luftdrucksensoren)
        rcall   lcdw40                                  ;ausgeben
        lds     r18, prsse2                             ;Anzahl der neuen Luftdrucksensoren holen
        ori     r18, 0x30                               ;in ASCII wandeln
        rcall   lcdw40                                  ;ausgeben
;
        ldi     r18, ' '                                ;Leerzeichen
        rcall   lcdw40                                  ;ausgeben
        ldi     r18, 'H'                                ;Zeichen "H" (Luftfeuchtigkeitssensoren)
        rcall   lcdw40                                  ;ausgeben
        lds     r18, humse2                             ;Anzahl der neuen Luftfeuchtigkeitssensoren holen
        ori     r18, 0x30                               ;in ASCII wandeln
        rcall   lcdw40                                  ;ausgeben
;
upd300: bst     flags, 6                                ;langer Tastendruck aktiv?
        brtc    upd300                                  ;nein -> Schleife
;
; Update-Funktion: Hinzufgen der neuen Temperatursensoren
;
        ldi     zl, low(tsentb + 23)
        ldi     zh, high(tsentb + 23)                   ;Zeiger auf letzten Platz der Sensortabelle
        ldi     r21, 24                                 ;24 Eintrge prfen
upd310: ld      r16, z                                  ;Tabelleneintrag holen
        cpi     r16, 8                                  ;gltige Busnummer (0-7)?
        brcs    upd320                                  ;ja -> letzten belegten Platz gefunden
        sbiw    zl, 1                                   ;Zeiger auf vorherigen Platz
        dec     r21                                     ;alle Eintrge geprft?
        brne    upd310                                  ;nein -> Schleife
;
upd320: adiw    zl, 1                                   ;Tabellenplatz korrigieren (auf nchsten freien Platz)
        movw    r2, zl                                  ;Zeiger sichern
        ldi     zl, low(tsent2)
        ldi     zh, high(tsent2)                        ;Zeiger auf Sensortabelle der Update-Suche
        movw    r4, zl                                  ;Zeiger sichern
        ldi     xl, low(tsenid)
        ldi     xh, high(tsenid)                        ;Zeiger auf ROM-IDs der gespeicherten Temperatursensoren
        ldi     r16, 8                                  ;eine ROM-ID besteht aus 8 Bytes
        mul     r16, r21                                ;mit Eintragsnummer multipliziert ergibt Position in ROM-ID-Liste
        add     xl, r0                                  ;Position zum Zeiger addieren
        adc     xh, r1                                  ;bertrag
        ldi     yl, low(tseni2)
        ldi     yh, high(tseni2)                        ;Zeiger auf ROM-IDs der Temperatursensoren (Update-Suche)
        ldi     r20, 24                                 ;maximale 24 neue Sensoren hinzufgen
;
upd330: cpi     r21, 24                                 ;noch Platz in der Sensortabelle?
        breq    upd400                                  ;nein -> Ende
        movw    zl, r4                                  ;Zeiger auf Sensortabelle der Update-Suche holen
        ld      r18, z+                                 ;Eintrag holen
        movw    r4, zl                                  ;Zeiger wieder speichern
        movw    zl, r2                                  ;Zeiger auf Sensortabelle der gespeicherten Daten holen
        cpi     r18, 8                                  ;gltige Busnummer (0-7)?
        brcc    upd340                                  ;nein -> Daten verwerfen
        st      z+, r18                                 ;sonst Busnummer speichern
        movw    r2, zl                                  ;und Zeiger wieder speichern
        inc     r21                                     ;nchsten Eintrag der Sensortabelle bearbeiten
;
upd340: ldi     r17, 8                                  ;ROM-ID (8 Bytes) kopieren
upd350: ld      r16, y+                                 ;ein Byte aus ROM-ID-Liste der Update-Suche holen
        cpi     r18, 8                                  ;gltige Busnummer (0-7)?
        brcc    upd360                                  ;nein -> Daten verwerfen
        st      x+, r16                                 ;sonst zur ROM-ID-Liste der gespeicherten Daten hinzufgen
upd360: dec     r17                                     ;alle Bytes kopiert?
        brne    upd350                                  ;nein -> Schleife
;
        dec     r20                                     ;wurden alle neuen Sensoren bearbeitet?
        brne    upd330                                  ;nein -> Schleife
;
; Update-Funktion: Hinzufgen der neuen Luftfeuchtigkeitssensoren
;
upd400: ldi     zl, low(hsentb + 2)
        ldi     zh, high(hsentb + 2)                    ;Zeiger auf letzten Platz der Sensortabelle
        ldi     r21, 3                                  ;3 Eintrge prfen
upd410: ld      r16, z                                  ;Tabelleneintrag holen
        cpi     r16, 8                                  ;gltige Busnummer (0-7)?
        brcs    upd420                                  ;ja -> letzten belegten Platz gefunden
        sbiw    zl, 1                                   ;Zeiger auf vorherigen Platz
        dec     r21                                     ;alle Eintrge geprft?
        brne    upd410                                  ;nein -> Schleife
;
upd420: adiw    zl, 1                                   ;Tabellenplatz korrigieren (auf nchsten freien Platz)
        movw    r2, zl                                  ;Zeiger sichern
        ldi     zl, low(hsent2)
        ldi     zh, high(hsent2)                        ;Zeiger auf Sensortabelle der Update-Suche
        movw    r4, zl                                  ;Zeiger sichern
        ldi     xl, low(hsenid)
        ldi     xh, high(hsenid)                        ;Zeiger auf ROM-IDs der gespeicherten Luftfeuchtigkeitssensoren
        ldi     r16, 8                                  ;eine ROM-ID besteht aus 8 Bytes
        mul     r16, r21                                ;mit Eintragsnummer multipliziert ergibt Position in ROM-ID-Liste
        add     xl, r0                                  ;Position zum Zeiger addieren
        adc     xh, r1                                  ;bertrag
        ldi     yl, low(hseni2)
        ldi     yh, high(hseni2)                        ;Zeiger auf ROM-IDs der Luftfeuchtigkeitssensoren (Update-Suche)
        ldi     r20, 3                                  ;maximale 3 neue Sensoren hinzufgen
;
upd430: cpi     r21, 3                                  ;noch Platz in der Sensortabelle?
        breq    upd500                                  ;nein -> Ende
        movw    zl, r4                                  ;Zeiger auf Sensortabelle der Update-Suche holen
        ld      r18, z+                                 ;Eintrag holen
        movw    r4, zl                                  ;Zeiger wieder speichern
        movw    zl, r2                                  ;Zeiger auf Sensortabelle der gespeicherten Daten holen
        cpi     r18, 8                                  ;gltige Busnummer (0-7)?
        brcc    upd440                                  ;nein -> Daten verwerfen
        st      z+, r18                                 ;sonst Busnummer speichern
        movw    r2, zl                                  ;und Zeiger wieder speichern
        inc     r21                                     ;nchsten Eintrag der Sensortabelle bearbeiten
;
upd440: ldi     r17, 8                                  ;8 Bytes der ROM-ID kopieren
upd450: ld      r16, y+                                 ;ein Byte aus ROM-ID-Liste der Update-Suche holen
        cpi     r18, 8                                  ;gltige Busnummer (0-7)?
        brcc    upd460                                  ;nein -> Daten verwerfen
        st      x+, r16                                 ;sonst zur ROM-ID-Liste der gespeicherten Daten hinzufgen
upd460: dec     r17                                     ;alle Bytes kopiert?
        brne    upd450                                  ;nein -> Schleife
;
        dec     r20                                     ;wurden alle neuen Sensoren bearbeitet?
        brne    upd430                                  ;nein -> Schleife
;
; Update-Funktion: neue Anzahl der Luftdrucksensoren speichern und "ok" ausgeben
;
upd500: lds     r16, prssen                             ;Anzahl der gespeicherten Drucksensoren holen
        lds     r17, prsse2                             ;Anzahl der Drucksensoren holen (Update-Suche)
        or      r16, r17                                ;hat sich die Anzahl gendert?
        sts     prssen, r16                             ;neuen Wert speichern
;
        rcall   savsen                                  ;neue Sensordaten im EEPROM speichern
        ldi     r17, 1                                  ;LCD in Datenmodus setzen
        ldi     r18, ' '                                ;Leerzeichen
        rcall   lcdw40                                  ;ausgeben
        ldi     r18, 0x7e                               ;Pfeil nach rechts
        rcall   lcdw40                                  ;ausgeben
        ldi     r18, ' '                                ;Leerzeichen
        rcall   lcdw40                                  ;ausgeben
        ldi     r18, 'o'                                ;Zeichen "o"
        rcall   lcdw40                                  ;ausgeben
        ldi     r18, 'k'                                ;Zeichen "k"
        rcall   lcdw40                                  ;ausgeben
;
upend:  rjmp    upend                                   ;Endlosschleife
;
; =====================================================================================================================================================================================================
; Unterprogramm-Routinen
; =====================================================================================================================================================================================================
;
; ----------------
; Anzeige-Routinen
; ----------------
;
; Sensoranzeige: Temperatursensoren 1-8
; Register: r16, r17, r18, r19, r20, r21, x, z
;           r22 <- aktuell bearbeiteter Sensor (fr Markierung, korrigiert, 0 entspricht Sensor 1)
;
dsens1: ldi     xl, low(temval)
        ldi     xh, high(temval)                        ;Zeiger auf Wertetabelle
        ldi     zl, low(temtxt * 2)
        ldi     zh, high(temtxt * 2)                    ;Zeiger auf Sensortexte
        clr     r19                                     ;Sensorzhler = 0 setzen
        clr     r20                                     ;Wertezhler = 0 setzen
        rjmp    dse010
;
; Sensoranzeige: Temperatursensoren 9-16
; Register: r16, r17, r18, r19, r20, r21, x, z
;           r22 <- aktuell bearbeiteter Sensor (fr Markierung, korrigiert, 8 entspricht Sensor 9)
;
dsens2: ldi     xl, low(temval + 8 * 5)
        ldi     xh, high(temval + 8 * 5)                ;Zeiger auf Wertetabelle
        ldi     zl, low(temtxt * 2 + 8 * 4)
        ldi     zh, high(temtxt * 2 + 8 * 4)            ;Zeiger auf Sensortexte
        ldi     r19, 8                                  ;Sensorzhler = 8 setzen
        clr     r20                                     ;Wertezhler = 0 setzen
        rjmp    dse010
;
; Sensoranzeige: Temperatursensoren 17-24
; Register: r16, r17, r18, r19, r20, r21, x, z
;           r22 <- aktuell bearbeiteter Sensor (fr Markierung, korrigiert, 16 entspricht Sensor 17)
;
dsens3: ldi     xl, low(temval + 16 * 5)
        ldi     xh, high(temval + 16 * 5)               ;Zeiger auf Wertetabelle
        ldi     zl, low(temtxt * 2 + 16 * 4)
        ldi     zh, high(temtxt * 2 + 16 * 4)           ;Zeiger auf Sensortexte
        ldi     r19, 16                                 ;Sensorzhler = 16 setzen
        clr     r20                                     ;Wertezhler = 0 setzen
;
dse010: movw    r2, xl                                  ;Zeiger sichern
        ldi     xl, low(adrtab)
        ldi     xh, high(adrtab)                        ;Zeiger auf LCD-Adresstabelle
        clr     r21                                     ;H-Byte = 0
        add     xl, r20
        adc     xh, r21                                 ;Tabellenplatz ermitteln
        ld      r18, x                                  ;LCD-Adresse holen
        movw    xl, r2                                  ;Zeiger wiederherstellen
        clr     r17                                     ;LCD in Befehlsmode setzen
        rcall   lcdw40                                  ;ausgeben
;
        ldi     r17, 1                                  ;LCD in Datenmodus setzen
        ldi     r21, 4                                  ;4 Zeichen ausgeben
dse020: lpm     r18, z+                                 ;Zeichen holen
        rcall   lcdw40                                  ;ausgeben
        dec     r21                                     ;alle Zeichen bearbeitet?
        brne    dse020                                  ;nein -> Schleife
;
        ldi     r21, 5                                  ;5 Zeichen ausgeben
dse030: ld      r18, x+                                 ;Zeichen holen
        rcall   lcdw40                                  ;ausgeben
        dec     r21                                     ;alle Zeichen bearbeitet?
        brne    dse030                                  ;nein -> Schleife
;
        ldi     r18, ' '                                ;Leerzeichen laden
        bst     flags, 2                                ;Anzeige-Automatik gesperrt?
        brts    dse040                                  ;ja -> Leerzeichen ausgeben
        cp      r19, r22                                ;sonst: aktueller Sensor?
        brne    dse040                                  ;nein -> berspringen
        ldi     r18, 0                                  ;sonst Pfeil links laden
dse040: rcall   lcdw40                                  ;ausgeben
;
        inc     r19                                     ;Sensorzhler erhhen
        inc     r20                                     ;Wertezhler erhhen
        cpi     r20, 8                                  ;alle Daten ausgegeben?
        brne    dse010                                  ;nein -> Schleife
        rjmp    dsymbl                                  ;Manuell-Symbol ausgeben
;
; Sensoranzeige: Luftdrucksensor 1 und Luftfeuchtigkeitssensoren 1-3 (Sensor 25, 26-28 und 29-31)
; Register: r16, r17, r18, r19, r20, r21, x, z
;           r22 <- aktuell bearbeiteter Sensor (fr Markierung, korrigiert, 24 entspricht Sensor 25)
;
dsens4: ldi     xl, low(senval)                         ;Luftdruckwert auf der ersten Zeile anzeigen
        ldi     xh, high(senval)                        ;Zeiger auf Wertetabelle
        lds     r18, adrtab                             ;LCD-Adresse von Zeilenanfang 1 holen
        clr     r17                                     ;LCD in Befehlsmode setzen
        rcall   lcdw40                                  ;ausgeben
        ldi     r19, 24                                 ;Text "Luftdruck"
        rcall   lcdtxt                                  ;ausgeben
;
        ldi     r21, 5                                  ;5 Zeichen ausgeben (Drucksensorwert)
dse080: ld      r18, x+                                 ;Zeichen holen
        rcall   lcdw40                                  ;ausgeben
        dec     r21                                     ;alle Zeichen bearbeitet?
        brne    dse080                                  ;nein -> Schleife
        ldi     r19, 25                                 ;Text "hPa"
        rcall   lcdtxt                                  ;ausgeben
;
        ldi     r18, ' '                                ;Leerzeichen laden
        bst     flags, 2                                ;Anzeige-Automatik gesperrt?
        brts    dse095                                  ;ja -> Leerzeichen ausgeben
        cpi     r22, 24                                 ;sonst: aktueller Sensor?
        brne    dse095                                  ;nein -> berspringen
        ldi     r18, 0                                  ;sonst Pfeil links laden
dse095: rcall   lcdw40                                  ;ausgeben
;
        ldi     zl, low(humtxt * 2)                     ;Temperaturen auf den Zeilenanfngen 2-4 ausgeben
        ldi     zh, high(humtxt * 2)                    ;Zeiger auf Feuchtesensortexte
        ldi     r19, 25                                 ;Sensorzhler = 25 setzen
        ldi     r20, 1                                  ;Wertezhler = 1 setzen
;
dse100: movw    r2, xl                                  ;Zeiger sichern
        ldi     xl, low(adrtab)
        ldi     xh, high(adrtab)                        ;Zeiger auf LCD-Adresstabelle
        clr     r21                                     ;H-Byte = 0
        add     xl, r20
        adc     xh, r21                                 ;Tabellenplatz ermitteln
        ld      r18, x                                  ;LCD-Adresse holen
        movw    xl, r2                                  ;Zeiger wiederherstellen
        clr     r17                                     ;LCD in Befehlsmode setzen
        rcall   lcdw40                                  ;ausgeben
;
        ldi     r17, 1                                  ;LCD in Datenmodus setzen
        ldi     r21, 4                                  ;4 Zeichen ausgeben (Sensorname)
dse110: lpm     r18, z+                                 ;Zeichen holen
        rcall   lcdw40                                  ;ausgeben
        dec     r21                                     ;alle Zeichen bearbeitet?
        brne    dse110                                  ;nein -> Schleife
;
        ldi     r21, 5                                  ;5 Zeichen ausgeben (Sensorwert Temperatur)
dse120: ld      r18, x+                                 ;Zeichen holen
        rcall   lcdw40                                  ;ausgeben
        dec     r21                                     ;alle Zeichen bearbeitet?
        brne    dse120                                  ;nein -> Schleife
;
        ldi     r18, ' '                                ;Leerzeichen laden
        bst     flags, 2                                ;Anzeige-Automatik gesperrt?
        brts    dse122                                  ;ja -> Leerzeichen ausgeben
        cp      r19, r22                                ;sonst: aktueller Sensor?
        brne    dse122                                  ;nein -> berspringen
        ldi     r18, 0                                  ;sonst Pfeil links laden
dse122: rcall   lcdw40                                  ;ausgeben
;
        adiw    zl, 6                                   ;Zeiger auf nchsten Sensortext setzen
        inc     r19                                     ;Sensorzhler erhhen
        inc     r20                                     ;Wertezhler erhhen
        cpi     r20, 4                                  ;alle Daten ausgegeben?
        brne    dse100                                  ;nein -> Schleife
;
        ldi     zl, low(humtxt * 2 + 4)                 ;Luftfeuchtigkeit auf den Zeilenmitten 2-4 ausgeben
        ldi     zh, high(humtxt * 2 + 4)                ;Zeiger auf Feuchtesensortexte
        ldi     r19, 28                                 ;Sensorzhler = 28 setzen
        ldi     r20, 5                                  ;Wertezhler = 5 setzen
;
dse200: movw    r2, xl                                  ;Zeiger sichern
        ldi     xl, low(adrtab)
        ldi     xh, high(adrtab)                        ;Zeiger auf LCD-Adresstabelle
        clr     r21                                     ;H-Byte = 0
        add     xl, r20
        adc     xh, r21                                 ;Tabellenplatz ermitteln
        ld      r18, x                                  ;LCD-Adresse holen
        movw    xl, r2                                  ;Zeiger wiederherstellen
        clr     r17                                     ;LCD in Befehlsmode setzen
        rcall   lcdw40                                  ;ausgeben
;
        ldi     r17, 1                                  ;LCD in Datenmodus setzen
        lpm     r18, z+                                 ;Feuchtesensorname erstes Zeichen holen
        rcall   lcdw40                                  ;ausgeben
        lpm     r18, z+                                 ;Feuchtesensorname zweites Zeichen holen
        rcall   lcdw40                                  ;ausgeben
;
        ldi     r21, 3                                  ;3 Zeichen ausgeben (Sensorwert Luftfeuchtigkeit)
        adiw    xl, 2                                   ;ungenutzte Stellen berspringen
dse210: ld      r18, x+                                 ;Zeichen holen
        rcall   lcdw40                                  ;ausgeben
        dec     r21                                     ;alle Zeichen bearbeitet?
        brne    dse210                                  ;nein -> Schleife
;
        ldi     r21, 4                                  ;4 Zeichen ausgeben (Einheit)
dse220: lpm     r18, z+                                 ;Zeichen holen
        rcall   lcdw40                                  ;ausgeben
        dec     r21                                     ;alle Zeichen bearbeitet?
        brne    dse220                                  ;nein -> Schleife
;
        ldi     r18, ' '                                ;Leerzeichen laden
        bst     flags, 2                                ;Anzeige-Automatik gesperrt?
        brts    dse230                                  ;ja -> Leerzeichen ausgeben
        cp      r19, r22                                ;sonst: aktueller Sensor?
        brne    dse230                                  ;nein -> berspringen
        ldi     r18, 0                                  ;sonst Pfeil links laden
dse230: rcall   lcdw40                                  ;ausgeben
;
        adiw    zl, 4                                   ;Zeiger auf nchsten Sensortext setzen
        inc     r19                                     ;Sensorzhler erhhen
        inc     r20                                     ;Wertezhler erhhen
        cpi     r20, 8                                  ;alle Daten ausgegeben?
        brne    dse200                                  ;nein -> Schleife
        rjmp    dsymbl                                  ;Manuell-Symbol ausgeben
;
; Alarmanzeige
; Register: r16, r17, r18, r19, r20, r21, r22, r23, x, z
;
dalarm: ldi     xl, low(alstat)
        ldi     xh, high(alstat)                        ;Zeiger auf Alarm-Status
        clr     r21                                     ;Alarmzhler = 0 setzen
;
        ldi     r23, 0xff                               ;letzten Alarm nicht zeigen
        bst     flags, 2                                ;Automatik-Anzeige gesperrt?
        brts    dal010                                  ;ja -> weiter
        lds     r16, intc16                             ;Interrupt-Zhler L holen
        lds     r17, intc16 + 1                         ;Interrupt-Zhler H holen
        or      r16, r17                                ;Zhler = 0?
        breq    dal010                                  ;ja -> weiter
        lds     r23, lastal                             ;sonst letzten Alarm holen
;
dal010: movw    r2, xl                                  ;Zeiger sichern
        ldi     xl, low(adrtab)
        ldi     xh, high(adrtab)                        ;Zeiger auf LCD-Adresstabelle
        clr     r22                                     ;H-Byte = 0
        add     xl, r21
        adc     xh, r22                                 ;Tabellenplatz ermitteln
        ld      r18, x                                  ;LCD-Adresse holen
        movw    xl, r2                                  ;Zeiger wiederherstellen
        clr     r17                                     ;LCD in Befehlsmode setzen
        rcall   lcdw40                                  ;ausgeben
;
        ldi     r20, 20                                 ;Zeichenzhler setzen
        ldi     r19, 20                                 ;Textnummer auf Alarmtext 1 setzen
        add     r19, r21                                ;Alarmzhler addieren
        rcall   lcdtxt                                  ;Alarmtext ausgeben
        ldi     r18, ' '                                ;Leerzeichen laden
        rcall   lcdw40                                  ;ausgeben
        dec     r20                                     ;Zeichenzhler - 1
;
        ldi     r18, 2                                  ;Sonderzeichen "offen"
        ld      r16, x+                                 ;Alarm-Flags holen
        bst     r16, 4                                  ;Alarm aktiv?
        brtc    dal030                                  ;nein -> weiter
        ldi     r18, 3                                  ;Sonderzeichen "geschlossen"
dal030: rcall   lcdw40                                  ;ausgeben
        dec     r20                                     ;Zeichenzhler - 1
;
        movw    r2, zl                                  ;Zeiger sichern
        cp      r21, r23                                ;Alarm krzlich ausgelst?
        brne    dal040                                  ;nein -> weiter
        ldi     r19, 7                                  ;sonst Text 7 (Pfeil links)
        rcall   lcdtxt                                  ;Text ausgeben
dal040: movw    zl, r2                                  ;Zeiger wiederherstellen
        rcall   lcdlin                                  ;Zeilenende lschen
;
        inc     r21                                     ;nchster Alarm
        cpi     r21, 4                                  ;alle Alarme bearbeitet?
        brne    dal010                                  ;nein -> Schleife
        rjmp    dsymbl                                  ;Manuell-Symbol ausgeben
;
; Statusanzeige: Versionsnummer, Anzahl der Sensoren, Sensorbelegung, Modus (Direkt, Speicher)
; Register: r16, r17, r18, r19, r20, x, z
;
dstatu: lds     r18, adrtab                             ;LCD-Adresse Zeile 1
        clr     r17                                     ;LCD in Befehlsmodus setzen
        rcall   lcdw40                                  ;Adresse setzen
        ldi     r20, 20                                 ;Zeichenzhler setzen
        ldi     r19, 0                                  ;Text 0 (Versionsnummer)
        rcall   lcdtxt                                  ;ausgeben
        rcall   lcdlin                                  ;alle Zeichen bis Zeilenende lschen
;
        lds     r18, adrtab + 1                         ;LCD-Adresse Zeile 2
        clr     r17                                     ;LCD in Befehlsmodus setzen
        rcall   lcdw40                                  ;Adresse setzen
        ldi     r20, 20                                 ;Zeichenzhler setzen
        ldi     r19, 1                                  ;Text 1 (Sensoren:)
        rcall   lcdtxt                                  ;ausgeben
;
        ldi     r18, 'T'                                ;Zeichen "T" (Temperatursensoren)
        rcall   lcdw40                                  ;ausgeben
        dec     r20                                     ;Zeichenzhler - 1
;
        lds     xl, temsen                              ;Anzahl der Temperatursensoren holen
        clr     xh                                      ;High-Byte lschen
        rcall   dezasc                                  ;in Dezimal-ASII umwandeln
        mov     r19, r16                                ;Einer-Stelle sichern
        cpi     r17, '0'                                ;Zehner-Stelle = 0?
        breq    dst010                                  ;ja -> nur Einer ausgeben
        mov     r18, r17                                ;sonst Zehner-Stelle holen
        rcall   lcdw40                                  ;ausgeben
        dec     r20                                     ;Zeichenzhler - 1
dst010: mov     r18, r19                                ;Einer-Stelle holen
        rcall   lcdw40                                  ;ausgeben
        dec     r20                                     ;Zeichenzhler - 1
;
        ldi     r18, ' '                                ;Leerzeichen
        rcall   lcdw40                                  ;ausgeben
        dec     r20                                     ;Zeichenzhler - 1
        ldi     r18, 'P'                                ;Zeichen "P" (Luftdrucksensoren)
        rcall   lcdw40                                  ;ausgeben
        dec     r20                                     ;Zeichenzhler - 1
        lds     r18, prssen                             ;Anzahl der Luftdrucksensoren holen
        ori     r18, 0x30                               ;in ASCII wandeln
        rcall   lcdw40                                  ;ausgeben
        dec     r20                                     ;Zeichenzhler - 1
;
        ldi     r18, ' '                                ;Leerzeichen
        rcall   lcdw40                                  ;ausgeben
        dec     r20                                     ;Zeichenzhler - 1
        ldi     r18, 'H'                                ;Zeichen "H" (Luftfeuchtigkeitssensoren)
        rcall   lcdw40                                  ;ausgeben
        dec     r20                                     ;Zeichenzhler - 1
        lds     r18, humsen                             ;Anzahl der Luftfeuchtigkeitssensoren holen
        ori     r18, 0x30                               ;in ASCII wandeln
        rcall   lcdw40                                  ;ausgeben
        dec     r20                                     ;Zeichenzhler - 1
        rcall   lcdlin                                  ;alle Zeichen bis Zeilenende lschen
;
        lds     r18, adrtab + 2                         ;LCD-Adresse Zeile 3
        clr     r17                                     ;LCD in Befehlsmodus setzen
        rcall   lcdw40                                  ;Adresse setzen
        ldi     r20, 20                                 ;Zeichenzhler setzen
        ldi     r19, 2                                  ;Text 2 (Belegung:)
        rcall   lcdtxt                                  ;ausgeben
;
        ldi     zl, low(senmap)
        ldi     zh, high(senmap)                        ;Zeiger auf Sensorbelegung
        ldi     r19, 8                                  ;8 Zeichen ausgeben
        ldi     r17, 1                                  ;LCD in Datenmodus setzen
dst020: ld      r18, z+                                 ;Zeichen holen
        rcall   lcdw40                                  ;ausgeben
        dec     r20                                     ;Zeichenzhler - 1
        dec     r19                                     ;alle Zeichen ausgegeben?
        brne    dst020                                  ;nein -> Schleife
        rcall   lcdlin                                  ;Zeilenende lschen
;
        lds     r18, adrtab + 3                         ;LCD-Adresse Zeile 4
        clr     r17                                     ;LCD in Befehlsmodus setzen
        rcall   lcdw40                                  ;Adresse setzen
        ldi     r20, 20                                 ;Zeichenzhler setzen
        ldi     r19, 3                                  ;Text 3 (Modus:)
        rcall   lcdtxt                                  ;ausgeben
        lds     r16, modflg                             ;Modus-Flag holen
        ldi     r19, 5                                  ;Text 5 (Speicher)
        tst     r16                                     ;Speicher-Modus?
        breq    dst030                                  ;ja -> Text ausgeben
        ldi     r19, 26                                 ;Text 26 (Update)
        cpi     r16, 1                                  ;Update-Modus?
        breq    dst030                                  ;ja -> Text ausgeben
        ldi     r19, 4                                  ;sonst Text 4 (Direkt)
dst030: rcall   lcdtxt                                  ;ausgeben
        rcall   lcdlin                                  ;Zeilenende lschen
;
; Alle manuellen Anzeigen bis Anzeige-Modus 5: Ausgabe eines Symbols unten rechts
;
dsymbl: bst     flags, 2                                ;Automatik-Anzeige gesperrt?
        brtc    dsy010                                  ;nein -> Ende
;
        lds     r18, adrtab + 7                         ;LCD-Adresse 4. Zeilenmitte
        ldi     r16, 9                                  ;Offset fr Zeilenende
        add     r18, r16                                ;LCD-Adresse fr Zeilenende
        clr     r17                                     ;LCD in Befehlsmodus setzen
        rcall   lcdw40                                  ;Adresse setzen
        ldi     r18, 4                                  ;Dreieck-Symbol laden
        ldi     r17, 1                                  ;LCD in Datenmodus setzen
        rcall   lcdw40                                  ;Zeichen ausgeben
dsy010: ret
;
; Anzeige der ROM-Codes aller 1-Wire-Sensoren, Anzeige-Modus 6-12
; Register: r16, r17, r18, r19, r20, r21, r22, x, z
;
drcode: clr     r22                                     ;Sensor-Zhler = 0 setzen
        clr     r21                                     ;Sensornummer = 0 setzen
        lds     r20, dimode                             ;Anzeige-Modus holen
        subi    r20, 6                                  ;Anzeige-Modus korrigieren (0 entspricht der ersten ROM-Code-Seite)
drc010: breq    drc020                                  ;richtige Seite erreicht? ja -> anzeigen
        subi    r21, -4                                 ;sonst Sensornummer um 4 erhhen
        dec     r20                                     ;Anzeige-Modus vermindern
        rjmp    drc010                                  ;erneut testen
;
drc020: ldi     xl, low(adrtab)
        ldi     xh, high(adrtab)                        ;Zeiger auf LCD-Adresstabelle
        clr     r16                                     ;H-Byte = 0
        add     xl, r22
        adc     xh, r16                                 ;Tabellenplatz ermitteln
        ld      r18, x                                  ;LCD-Adresse holen
        clr     r17                                     ;LCD in Befehlsmode setzen
        rcall   lcdw40                                  ;ausgeben
;
        lds     r20, temsen                             ;Anzahl der Temperatursensoren holen
        cp      r21, r20                                ;Sensornummer im Bereich der Temperatursensoren?
        brcs    drc030                                  ;ja -> Temperatursensor anzeigen
        lds     r19, humsen                             ;sonst Anzahl der Luftfeuchtigkeitssensoren holen
        add     r20, r19                                ;zur Anzahl der Temperatursensoren addieren
        cp      r21, r20                                ;Sensornummer im Bereich der Luftfeuchtigkeitssensoren?
        brcs    drc050                                  ;ja -> Luftfeuchtigkeitssensor anzeigen
        rjmp    drc100                                  ;sonst keine weiteren Sensoren anzeigen (Leerzeile ausgeben)
;
drc030: ldi     zl, low(tsenid)
        ldi     zh, high(tsenid)                        ;Zeiger auf ROM-Codes der Temperatursensoren
        ldi     xl, low(tsentb)
        ldi     xh, high(tsentb)                        ;Zeiger auf Busnummern-Tabelle der Temperatursensoren
;
        mov     r20, r21                                ;Sensornummer kopieren
        mov     r18, r21                                ;Sensornummer nochmals kopieren
        inc     r18                                     ;korrigieren (0-7 -> 1-8)
        subi    r18, -0x30                              ;in ASCII-Ziffer umwandeln
        cpi     r18, 0x39                               ;Ziffer > 8?
        brcs    drc040                                  ;nein -> weiter
        subi    r18, -8                                 ;sonst in ASCII-Grobuchstabe umwandeln (A-P)
drc040: ldi     r17, 1                                  ;LCD in Datenmodus setzen
        rcall   lcdw40                                  ;ausgeben
        rjmp    drc060                                  ;Busnummer und ROM-Code ausgeben
;
drc050: ldi     zl, low(hsenid)
        ldi     zh, high(hsenid)                        ;Zeiger auf ROM-Codes der Luftfeuchtigkeitssensoren
        ldi     xl, low(hsentb)
        ldi     xh, high(hsentb)                        ;Zeiger auf Busnummern-Tabelle der Luftfeuchtigkeitssensoren
;
        mov     r20, r21                                ;Sensornummer kopieren
        lds     r19, temsen                             ;Anzahl der Temperatursensoren holen und von der
        sub     r20, r19                                ;Sensornummer abziehen (ergibt Anzahl der Luftfeuchtigkeitssensoren)
        mov     r18, r20                                ;Ergebnis kopieren
        subi    r18, -0x51                              ;in ASCII-Grobuchstabe umwandeln (Q-S)
        ldi     r17, 1                                  ;LCD in Datenmodus setzen
        rcall   lcdw40                                  ;ausgeben
;
drc060: mov     r16, r20                                ;aufbereitete Sensornummer holen
        add     r16, r16
        add     r16, r16
        add     r16, r16                                ;mit 8 multiplizieren
        clr     r17
        add     zl, r16                                 ;ROM-Code-Tabellenplatz berechnen (Low)
        adc     zh, r17                                 ;ROM-Code-Tabellenplatz berechnen (High)
        add     xl, r20                                 ;Busnummer-Tabellenplatz berechnen (Low)
        adc     xh, r17                                 ;Busnummer-Tabellenplatz berechnen (High)
;
        ldi     r17, 1                                  ;LCD in Datenmodus setzen
        ldi     r18, '-'                                ;Minus-Zeichen
        rcall   lcdw40                                  ;ausgeben
        ld      r18, x                                  ;Busnummer aus Tabelle holen
        inc     r18                                     ;korrigieren (0-7 -> 1-8)
        ori     r18, 0x30                               ;in ASCII-Ziffer umwandeln
        rcall   lcdw40                                  ;ausgeben
        ldi     r18, ' '                                ;Leerzeichen
        rcall   lcdw40                                  ;ausgeben
;
        ldi     r20, 8                                  ;8 Bytes ROM-Code ausgeben
drc070: ld      r19, z+                                 ;Byte aus ROM-Code-Speicher holen
        mov     r18, r19                                ;Byte kopieren
        swap    r18                                     ;Nibbles tauschen (oberes Nibble nach unten holen)
        andi    r18, 0x0f                               ;unteres Nibble filtern
        ori     r18, 0x30                               ;in ASCII-Ziffer umwandeln
        cpi     r18, 0x3a                               ;ASCII-Wert > 9?
        brcs    drc080                                  ;nein -> weiter
        subi    r18, -7                                 ;sonst in ASCII-Grobuchstabe umwandeln
drc080: rcall   lcdw40                                  ;ausgeben
        mov     r18, r19                                ;Byte nochmals kopieren
        andi    r18, 0x0f                               ;unteres Nibble filtern
        ori     r18, 0x30                               ;in ASCII-Ziffer umwandeln
        cpi     r18, 0x3a                               ;ASCII-Wert > 9?
        brcs    drc090                                  ;nein -> weiter
        subi    r18, -7                                 ;sonst in ASCII-Grobuchstabe umwandeln
drc090: rcall   lcdw40                                  ;ausgeben
        dec     r20                                     ;alle Bytes ausgegeben?
        brne    drc070                                  ;nein -> Schleife
        rjmp    drc110
;
drc100: ldi     r20, 20                                 ;20 Leerzeichen
        rcall   lcdlin                                  ;ausgeben
drc110: inc     r21                                     ;Sensornummer erhhen
        inc     r22                                     ;Sensor-Zhler erhhen
        cpi     r22, 4                                  ;alle 4 Sensoren bearbeitet?
        brcc    drc120                                  ;ja -> Ende
        rjmp    drc020                                  ;sonst Schleife
drc120: ret
;
; ---------------
; 1-Wire Routinen
; ---------------
;
; 1-Wire, Einzelnes Bit in 1-Wire-Puffer einschieben
; Register: r16, r17, x
;
bitsto: ldi     xl, low(owbuff)
        ldi     xh, high(owbuff)                        ;Zeiger auf 1-Wire-Puffer
        ldi     r17, 8                                  ;8 Pufferbytes bearbeiten
        ror     r16                                     ;neues Bit in Carry schieben
bits10: ld      r16, x                                  ;Pufferbyte holen
        ror     r16                                     ;bertrag-Bit einschieben
        st      x+, r16                                 ;und wieder speichern
        dec     r17                                     ;alle Pufferbytes bearbeitet?
        brne    bits10                                  ;nein -> Schleife
        ret
;
; 1-Wire, Reset und Presence-Test
; Register: r16
;           r20   <- Busnummer
;           Carry -> 0 = ok, 1 = Fehler
;
owres:  cpi     r20, 0                                  ;Busnummer = 0?
        breq    ow0res                                  ;ja -> bearbeiten
        cpi     r20, 1                                  ;Busnummer = 1?
        breq    ow1res                                  ;ja -> bearbeiten
        cpi     r20, 2                                  ;Busnummer = 2?
        breq    ow2res                                  ;ja -> bearbeiten
        cpi     r20, 3                                  ;Busnummer = 3?
        breq    ow3res                                  ;ja -> bearbeiten
        cpi     r20, 4                                  ;Busnummer = 4?
        breq    ow4res                                  ;ja -> bearbeiten
        cpi     r20, 5                                  ;Busnummer = 5?
        breq    owrs10                                  ;ja -> bearbeiten
        cpi     r20, 6                                  ;Busnummer = 6?
        breq    owrs20                                  ;ja -> bearbeiten
        rjmp    ow7res                                  ;sonst Busnummer 7 bearbeiten
owrs10: rjmp    ow5res
owrs20: rjmp    ow6res
;
ow0res: cbi     portc, 0                                ;Strong-pull-up auschalten
        sbi     ddrc, 0                                 ;Port auf Ausgang Low setzen
        rcall   w480us                                  ;480s warten
        cbi     ddrc, 0                                 ;Port auf Eingang setzen
        rcall   w060us                                  ;60s warten (Presence Start)
        in      r16, pinc                               ;Port lesen
        andi    r16, 0x01                               ;Portpin selektieren
        breq    ow0s10                                  ;Presence ok? ja -> weiter
        sec                                             ;sonst Fehler
        ret
ow0s10: rcall   w420us                                  ;240s warten (Presence Ende)
        ret                                             ;Reset ok
;
ow1res: cbi     portc, 1                                ;Strong-pull-up auschalten
        sbi     ddrc, 1                                 ;Port auf Ausgang Low setzen
        rcall   w480us                                  ;480s warten
        cbi     ddrc, 1                                 ;Port auf Eingang setzen
        rcall   w060us                                  ;60s warten (Presence Start)
        in      r16, pinc                               ;Port lesen
        andi    r16, 0x02                               ;Portpin selektieren
        breq    ow1s10                                  ;Presence ok? ja -> weiter
        sec                                             ;sonst Fehler
        ret
ow1s10: rcall   w420us                                  ;240s warten (Presence Ende)
        ret                                             ;Reset ok
;
ow2res: cbi     portc, 2                                ;Strong-pull-up auschalten
        sbi     ddrc, 2                                 ;Port auf Ausgang Low setzen
        rcall   w480us                                  ;480s warten
        cbi     ddrc, 2                                 ;Port auf Eingang setzen
        rcall   w060us                                  ;60s warten (Presence Start)
        in      r16, pinc                               ;Port lesen
        andi    r16, 0x04                               ;Portpin selektieren
        breq    ow2s10                                  ;Presence ok? ja -> weiter
        sec                                             ;sonst Fehler
        ret
ow2s10: rcall   w420us                                  ;240s warten (Presence Ende)
        ret                                             ;Reset ok
;
ow3res: cbi     portc, 3                                ;Strong-pull-up auschalten
        sbi     ddrc, 3                                 ;Port auf Ausgang Low setzen
        rcall   w480us                                  ;480s warten
        cbi     ddrc, 3                                 ;Port auf Eingang setzen
        rcall   w060us                                  ;60s warten (Presence Start)
        in      r16, pinc                               ;Port lesen
        andi    r16, 0x08                               ;Portpin selektieren
        breq    ow3s10                                  ;Presence ok? ja -> weiter
        sec                                             ;sonst Fehler
        ret
ow3s10: rcall   w420us                                  ;240s warten (Presence Ende)
        ret                                             ;Reset ok
;
ow4res: cbi     portd, 4                                ;Strong-pull-up auschalten
        sbi     ddrd, 4                                 ;Port auf Ausgang Low setzen
        rcall   w480us                                  ;480s warten
        cbi     ddrd, 4                                 ;Port auf Eingang setzen
        rcall   w060us                                  ;60s warten (Presence Start)
        in      r16, pind                               ;Port lesen
        andi    r16, 0x10                               ;Portpin selektieren
        breq    ow4s10                                  ;Presence ok? ja -> weiter
        sec                                             ;sonst Fehler
        ret
ow4s10: rcall   w420us                                  ;240s warten (Presence Ende)
        ret                                             ;Reset ok
;
ow5res: cbi     portd, 5                                ;Strong-pull-up auschalten
        sbi     ddrd, 5                                 ;Port auf Ausgang Low setzen
        rcall   w480us                                  ;480s warten
        cbi     ddrd, 5                                 ;Port auf Eingang setzen
        rcall   w060us                                  ;60s warten (Presence Start)
        in      r16, pind                               ;Port lesen
        andi    r16, 0x20                               ;Portpin selektieren
        breq    ow5s10                                  ;Presence ok? ja -> weiter
        sec                                             ;sonst Fehler
        ret
ow5s10: rcall   w420us                                  ;240s warten (Presence Ende)
        ret                                             ;Reset ok
;
ow6res: cbi     portd, 6                                ;Strong-pull-up auschalten
        sbi     ddrd, 6                                 ;Port auf Ausgang Low setzen
        rcall   w480us                                  ;480s warten
        cbi     ddrd, 6                                 ;Port auf Eingang setzen
        rcall   w060us                                  ;60s warten (Presence Start)
        in      r16, pind                               ;Port lesen
        andi    r16, 0x40                               ;Portpin selektieren
        breq    ow6s10                                  ;Presence ok? ja -> weiter
        sec                                             ;sonst Fehler
        ret
ow6s10: rcall   w420us                                  ;240s warten (Presence Ende)
        ret                                             ;Reset ok
;
ow7res: cbi     portd, 7                                ;Strong-pull-up auschalten
        sbi     ddrd, 7                                 ;Port auf Ausgang Low setzen
        rcall   w480us                                  ;480s warten
        cbi     ddrd, 7                                 ;Port auf Eingang setzen
        rcall   w060us                                  ;60s warten (Presence Start)
        in      r16, pind                               ;Port lesen
        andi    r16, 0x80                               ;Portpin selektieren
        breq    ow7s10                                  ;Presence ok? ja -> weiter
        sec                                             ;sonst Fehler
        ret
ow7s10: rcall   w420us                                  ;240s warten (Presence Ende)
        ret                                             ;Reset ok
;
; 1-Wire, Schreiben eines Byte
; Register: r16, r18
;           r19 <- zu sendendes Byte
;           r20 <-Busnummer (0-7)
;
owwrit: ldi     r18, 8                                  ;8 Bits ausgeben
owwr10: ror     r19                                     ;LSB in Carry schieben
        rcall   owbwri                                  ;Bit ausgeben
        dec     r18                                     ;alle Bits gesendet?
        brne    owwr10                                  ;nein -> Schleife
        ret
;
; 1-Wire, Lesen eines Bytes
; Register: r16, r17, r18
;           r19 -> gelesenes Byte
;           r20 <- Busnummer (0-7)
;
owread: ldi     r18, 8                                  ;8 Bits ausgeben
owre10: rcall   owbrea                                  ;Bit einlesen
        ror     r19                                     ;Bit aus Carry einschieben
        dec     r18                                     ;alle Bits gelesen?
        brne    owre10                                  ;nein -> Schleife
        ret
;
; 1-Wire, Schreiben eines Bits
; Register: r16
;           r20   <- Busnummer (0-7)
;           Carry <- zu schreibendes Bit
;
owbwri: mov     r16, r20                                ;Busnummer kopieren
        tst     r16                                     ;Busnummer = 0?
        breq    ow0wri                                  ;ja -> bearbeiten
        dec     r16                                     ;Busnummer = 1?
        breq    ow1wri                                  ;ja -> bearbeiten
        dec     r16                                     ;Busnummer = 2?
        breq    ow2wri                                  ;ja -> bearbeiten
        dec     r16                                     ;Busnummer = 3?
        breq    ow3wri                                  ;ja -> bearbeiten
        dec     r16                                     ;Busnummer = 4?
        breq    ow4wri                                  ;ja -> bearbeiten
        dec     r16                                     ;Busnummer = 5?
        breq    ow5wri                                  ;ja -> bearbeiten
        dec     r16                                     ;Busnummer = 6?
        breq    ow6wri                                  ;ja -> bearbeiten
        rjmp    ow7wri                                  ;sonst Busnummer 7 bearbeiten
;
ow0wri: sbi     ddrc, 0                                 ;Port auf Ausgang Low setzen
        brcs    ow0w10                                  ;1-Bit? ja -> weiter
        rcall   w060us                                  ;60s warten (= Low)
        cbi     ddrc, 0                                 ;Port auf Eingang setzen
        rcall   w005us                                  ;4,5s warten (Pause)
        ret
ow0w10: rcall   w002us                                  ;2s warten (= High)
        cbi     ddrc, 0                                 ;Port auf Eingang setzen
        rcall   w060us                                  ;60s warten
        ret
;
ow1wri: sbi     ddrc, 1                                 ;Port auf Ausgang Low setzen
        brcs    ow1w10                                  ;1-Bit? ja -> weiter
        rcall   w060us                                  ;60s warten (= Low)
        cbi     ddrc, 1                                 ;Port auf Eingang setzen
        rcall   w005us                                  ;4,5s warten (Pause)
        ret
ow1w10: rcall   w002us                                  ;2s warten (= High)
        cbi     ddrc, 1                                 ;Port auf Eingang setzen
        rcall   w060us                                  ;60s warten
        ret
;
ow2wri: sbi     ddrc, 2                                 ;Port auf Ausgang Low setzen
        brcs    ow2w10                                  ;1-Bit? ja -> weiter
        rcall   w060us                                  ;60s warten (= Low)
        cbi     ddrc, 2                                 ;Port auf Eingang setzen
        rcall   w005us                                  ;4,5s warten (Pause)
        ret
ow2w10: rcall   w002us                                  ;2s warten (= High)
        cbi     ddrc, 2                                 ;Port auf Eingang setzen
        rcall   w060us                                  ;60s warten
        ret
;
ow3wri: sbi     ddrc, 3                                 ;Port auf Ausgang Low setzen
        brcs    ow3w10                                  ;1-Bit? ja -> weiter
        rcall   w060us                                  ;60s warten (= Low)
        cbi     ddrc, 3                                 ;Port auf Eingang setzen
        rcall   w005us                                  ;4,5s warten (Pause)
        ret
ow3w10: rcall   w002us                                  ;2s warten (= High)
        cbi     ddrc, 3                                 ;Port auf Eingang setzen
        rcall   w060us                                  ;60s warten
        ret
;
ow4wri: sbi     ddrd, 4                                 ;Port auf Ausgang Low setzen
        brcs    ow4w10                                  ;1-Bit? ja -> weiter
        rcall   w060us                                  ;60s warten (= Low)
        cbi     ddrd, 4                                 ;Port auf Eingang setzen
        rcall   w005us                                  ;4,5s warten (Pause)
        ret
ow4w10: rcall   w002us                                  ;2s warten (= High)
        cbi     ddrd, 4                                 ;Port auf Eingang setzen
        rcall   w060us                                  ;60s warten
        ret
;
ow5wri: sbi     ddrd, 5                                 ;Port auf Ausgang Low setzen
        brcs    ow5w10                                  ;1-Bit? ja -> weiter
        rcall   w060us                                  ;60s warten (= Low)
        cbi     ddrd, 5                                 ;Port auf Eingang setzen
        rcall   w005us                                  ;4,5s warten (Pause)
        ret
ow5w10: rcall   w002us                                  ;2s warten (= High)
        cbi     ddrd, 5                                 ;Port auf Eingang setzen
        rcall   w060us                                  ;60s warten
        ret
;
ow6wri: sbi     ddrd, 6                                 ;Port auf Ausgang Low setzen
        brcs    ow6w10                                  ;1-Bit? ja -> weiter
        rcall   w060us                                  ;60s warten (= Low)
        cbi     ddrd, 6                                 ;Port auf Eingang setzen
        rcall   w005us                                  ;4,5s warten (Pause)
        ret
ow6w10: rcall   w002us                                  ;2s warten (= High)
        cbi     ddrd, 6                                 ;Port auf Eingang setzen
        rcall   w060us                                  ;60s warten
        ret
;
ow7wri: sbi     ddrd, 7                                 ;Port auf Ausgang Low setzen
        brcs    ow7w10                                  ;1-Bit? ja -> weiter
        rcall   w060us                                  ;60s warten (= Low)
        cbi     ddrd, 7                                 ;Port auf Eingang setzen
        rcall   w005us                                  ;4,5s warten (Pause)
        ret
ow7w10: rcall   w002us                                  ;2s warten (= High)
        cbi     ddrd, 7                                 ;Port auf Eingang setzen
        rcall   w060us                                  ;60s warten
        ret
;
; 1-Wire, Lesen eines Bits
; Register: r16, r17
;           r20   <- Busnummer (0-7)
;           Carry -> gelesenes Bit
;
owbrea: mov     r16, r20                                ;Busnummer kopieren
        tst     r16                                     ;Busnummer = 0?
        breq    ow0rea                                  ;ja -> bearbeiten
        dec     r16                                     ;Busnummer = 1?
        breq    ow1rea                                  ;ja -> bearbeiten
        dec     r16                                     ;Busnummer = 2?
        breq    ow2rea                                  ;ja -> bearbeiten
        dec     r16                                     ;Busnummer = 3?
        breq    ow3rea                                  ;ja -> bearbeiten
        dec     r16                                     ;Busnummer = 4?
        breq    ow4rea                                  ;ja -> bearbeiten
        dec     r16                                     ;Busnummer = 5?
        breq    owbr10                                  ;ja -> bearbeiten
        dec     r16                                     ;Busnummer = 6?
        breq    owbr20                                  ;ja -> bearbeiten
        rjmp    ow7rea                                  ;sonst Busnummer 7 bearbeiten
owbr10: rjmp    ow5rea
owbr20: rjmp    ow6rea
;
ow0rea: sbi     ddrc, 0                                 ;Port auf Ausgang Low setzen
        rcall   w002us                                  ;2s warten
        cbi     ddrc, 0                                 ;Port auf Eingang setzen
        rcall   w005us                                  ;4,5s warten (Pause)
        rcall   w005us                                  ;4,5s warten (Pause)
        in      r17, pinc                               ;Port lesen
        andi    r17, 0x01                               ;Portpin selektieren
        clc                                             ;Carry lschen (= 0)
        breq    ow0r10                                  ;Port = 0? ja -> weiter
        sec                                             ;sonst Carry setzen (= 1)
ow0r10: rcall   w050us                                  ;50s warten
        ret
;
ow1rea: sbi     ddrc, 1                                 ;Port auf Ausgang Low setzen
        rcall   w002us                                  ;2s warten
        cbi     ddrc, 1                                 ;Port auf Eingang setzen
        rcall   w005us                                  ;4,5s warten (Pause)
        rcall   w005us                                  ;4,5s warten (Pause)
        in      r17, pinc                               ;Port lesen
        andi    r17, 0x02                               ;Portpin selektieren
        clc                                             ;Carry lschen (= 0)
        breq    ow1r10                                  ;Port = 0? ja -> weiter
        sec                                             ;sonst Carry setzen (= 1)
ow1r10: rcall   w050us                                  ;50s warten
        ret
;
ow2rea: sbi     ddrc, 2                                 ;Port auf Ausgang Low setzen
        rcall   w002us                                  ;2s warten
        cbi     ddrc, 2                                 ;Port auf Eingang setzen
        rcall   w005us                                  ;4,5s warten (Pause)
        rcall   w005us                                  ;4,5s warten (Pause)
        in      r17, pinc                               ;Port lesen
        andi    r17, 0x04                               ;Portpin selektieren
        clc                                             ;Carry lschen (= 0)
        breq    ow2r10                                  ;Port = 0? ja -> weiter
        sec                                             ;sonst Carry setzen (= 1)
ow2r10: rcall   w050us                                  ;50s warten
        ret
;
ow3rea: sbi     ddrc, 3                                 ;Port auf Ausgang Low setzen
        rcall   w002us                                  ;2s warten
        cbi     ddrc, 3                                 ;Port auf Eingang setzen
        rcall   w005us                                  ;4,5s warten (Pause)
        rcall   w005us                                  ;4,5s warten (Pause)
        in      r17, pinc                               ;Port lesen
        andi    r17, 0x08                               ;Portpin selektieren
        clc                                             ;Carry lschen (= 0)
        breq    ow3r10                                  ;Port = 0? ja -> weiter
        sec                                             ;sonst Carry setzen (= 1)
ow3r10: rcall   w050us                                  ;50s warten
        ret
;
ow4rea: sbi     ddrd, 4                                 ;Port auf Ausgang Low setzen
        rcall   w002us                                  ;2s warten
        cbi     ddrd, 4                                 ;Port auf Eingang setzen
        rcall   w005us                                  ;4,5s warten (Pause)
        rcall   w005us                                  ;4,5s warten (Pause)
        in      r17, pind                               ;Port lesen
        andi    r17, 0x10                               ;Portpin selektieren
        clc                                             ;Carry lschen (= 0)
        breq    ow4r10                                  ;Port = 0? ja -> weiter
        sec                                             ;sonst Carry setzen (= 1)
ow4r10: rcall   w050us                                  ;50s warten
        ret
;
ow5rea: sbi     ddrd, 5                                 ;Port auf Ausgang Low setzen
        rcall   w002us                                  ;2s warten
        cbi     ddrd, 5                                 ;Port auf Eingang setzen
        rcall   w005us                                  ;4,5s warten (Pause)
        rcall   w005us                                  ;4,5s warten (Pause)
        in      r17, pind                               ;Port lesen
        andi    r17, 0x20                               ;Portpin selektieren
        clc                                             ;Carry lschen (= 0)
        breq    ow5r10                                  ;Port = 0? ja -> weiter
        sec                                             ;sonst Carry setzen (= 1)
ow5r10: rcall   w050us                                  ;50s warten
        ret
;
ow6rea: sbi     ddrd, 6                                 ;Port auf Ausgang Low setzen
        rcall   w002us                                  ;2s warten
        cbi     ddrd, 6                                 ;Port auf Eingang setzen
        rcall   w005us                                  ;4,5s warten (Pause)
        rcall   w005us                                  ;4,5s warten (Pause)
        in      r17, pind                               ;Port lesen
        andi    r17, 0x40                               ;Portpin selektieren
        clc                                             ;Carry lschen (= 0)
        breq    ow6r10                                  ;Port = 0? ja -> weiter
        sec                                             ;sonst Carry setzen (= 1)
ow6r10: rcall   w050us                                  ;50s warten
        ret
;
ow7rea: sbi     ddrd, 7                                 ;Port auf Ausgang Low setzen
        rcall   w002us                                  ;2s warten
        cbi     ddrd, 7                                 ;Port auf Eingang setzen
        rcall   w005us                                  ;4,5s warten (Pause)
        rcall   w005us                                  ;4,5s warten (Pause)
        in      r17, pind                               ;Port lesen
        andi    r17, 0x80                               ;Portpin selektieren
        clc                                             ;Carry lschen (= 0)
        breq    ow7r10                                  ;Port = 0? ja -> weiter
        sec                                             ;sonst Carry setzen (= 1)
ow7r10: rcall   w050us                                  ;50s warten
        ret
;
; 1-Wire, Strong Pull-up einschalten, Bus fest auf High-Pegel legen whrend Temperaturmessung
; Register: r16
;           r20 <- Busnummer
;
owstpu: cpi     r20, 0                                  ;Busnummer = 0?
        breq    ow0stp                                  ;ja -> bearbeiten
        cpi     r20, 1                                  ;Busnummer = 1?
        breq    ow1stp                                  ;ja -> bearbeiten
        cpi     r20, 2                                  ;Busnummer = 2?
        breq    ow2stp                                  ;ja -> bearbeiten
        cpi     r20, 3                                  ;Busnummer = 3?
        breq    ow3stp                                  ;ja -> bearbeiten
        cpi     r20, 4                                  ;Busnummer = 4?
        breq    ow4stp                                  ;ja -> bearbeiten
        cpi     r20, 5                                  ;Busnummer = 5?
        breq    ow5stp                                  ;ja -> bearbeiten
        cpi     r20, 6                                  ;Busnummer = 6?
        breq    ow6stp                                  ;ja -> bearbeiten
        rjmp    ow7stp                                  ;sonst Busnummer 7 bearbeiten
;
ow0stp: sbi     portc, 0                                ;Portausgang auf High setzen
        sbi     ddrc, 0                                 ;Port auf Ausgang umschalten
        ret
;
ow1stp: sbi     portc, 1                                ;Portausgang auf High setzen
        sbi     ddrc, 1                                 ;Port auf Ausgang umschalten
        ret
;
ow2stp: sbi     portc, 2                                ;Portausgang auf High setzen
        sbi     ddrc, 2                                 ;Port auf Ausgang umschalten
        ret
;
ow3stp: sbi     portc, 3                                ;Portausgang auf High setzen
        sbi     ddrc, 3                                 ;Port auf Ausgang umschalten
        ret
;
ow4stp: sbi     portd, 4                                ;Portausgang auf High setzen
        sbi     ddrd, 4                                 ;Port auf Ausgang umschalten
        ret
;
ow5stp: sbi     portd, 5                                ;Portausgang auf High setzen
        sbi     ddrd, 5                                 ;Port auf Ausgang umschalten
        ret
;
ow6stp: sbi     portd, 6                                ;Portausgang auf High setzen
        sbi     ddrd, 6                                 ;Port auf Ausgang umschalten
        ret
;
ow7stp: sbi     portd, 7                                ;Portausgang auf High setzen
        sbi     ddrd, 7                                 ;Port auf Ausgang umschalten
        ret
;
; 1-Wire, Unterprogramm zum Senden des Kommandos "Match ROM" und der Sensor-ROM-ID
; Register: r16, r18, r19, r21
;           r20 <- Busnummer (0-7)
;           x   <- Zeiger auf Sensor-ROM-ID
;
owmaro: ldi     r19, 0x55                               ;ROM-Kommando "Match ROM"
        rcall   owwrit                                  ;senden
        ldi     r21, 8                                  ;8 Bytes ROM-ID senden
owma10: ld      r19, -x                                 ;Byte aus Tabelle holen
        rcall   owwrit                                  ;Byte senden
        dec     r21                                     ;alle Bytes gesendet?
        brne    owma10                                  ;nein -> Schleife
        ret
;
; CRC-Berechnung fr 1-Wire-Devices fr 8 Byte im 1-Wire-Puffer, nach Application Note 27 von Dallas/Maxim
; Register: r10, r11, r16, r18, r19, x
;           r17 -> ermittelter CRC-Wert
;
crc:    ldi     xl, low(owbuff + 8)
        ldi     xh, high(owbuff + 8)                    ;Zeiger auf 1-Wire-Puffer-Ende
        ldi     r19, 8                                  ;8 Bytes berechnen
        ldi     r16, 0x18                               ;Konstante fr Berechnung
        mov     r10, r16                                ;bereitlegen
        clr     r17                                     ;CRC-Startwert = 0 setzen
;
crc100: ld      r18, -x                                 ;Byte (LSB) aus Puffer holen
        mov     r11, r18                                ;Byte zwischenspeichern
        ldi     r16, 8                                  ;8 Verschiebungen
crc200: eor     r18, r17                                ;CRC berechnen
        ror     r18                                     ;Ergebnis in Carry schieben
        mov     r18, r17                                ;letzten CRC-Wert holen
        brcc    crc300                                  ;war Ergebnis = 0? ja -> weiter
        eor     r18, r10                                ;sonst CRC-Wert updaten
crc300: ror     r18                                     ;neuen CRC-Wert positionieren
        mov     r17, r18                                ;und speichern
        mov     r18, r11                                ;restliche Bits holen
        bst     r18, 0                                  ;LSB sichern, nchste Bitposition
        ror     r18                                     ;gesichertes Bit als MSB einsetzen
        bld     r18, 7                                  ;(Ersatz fr RR-Befehl)
        mov     r11, r18                                ;und Wert zwischenspeichern
        dec     r16                                     ;alle Verschiebungen erledigt?
        brne    crc200                                  ;nein -> Schleife
        dec     r19                                     ;alle Bytes bearbeitet?
        brne    crc100                                  ;nein -> Schleife
        ret
;
; ---------------
; RS-232 Routinen
; ---------------
;
; Alarmausgabe ber RS-232
; Register: r16
;           r19 <- Alarmwert (0/1)
;           r20 <- Alarmnummer (0-3)
;
rsaout: mov     r16, r20                                ;aktuelle Alarmnummer holen
        inc     r16                                     ;Nummer korrigieren
        ori     r16, 0x40                               ;in ASCII-Zeichen wandeln
        rcall   rssend                                  ;Zeichen senden
        ldi     r16, ':'                                ;Doppelpunkt laden
        rcall   rssend                                  ;Zeichen senden
        ldi     r16, '0'                                ;Zeichen '0' laden
        add     r16, r19                                ;Alarmwert addieren
        rcall   rssend                                  ;Zeichen senden
        ldi     r16, cr                                 ;CR-Zeichen laden
        rcall   rssend                                  ;Zeichen senden
        ret
;
; USART Senden (RS-232)
; Register: r16 <- zu sendendes Zeichen
;
rssend: sbis    ucsra, udre                             ;Sendepuffer leer?
        rjmp    rssend                                  ;nein -> warten
        out     udr, r16                                ;sonst Byte senden
        ret
;
; ------------
; LCD-Routinen
; ------------
;
; LCD-Ausgabe eines Textes unter Bercksichtigung der Men-Sprache
; Register: r16, r17, r18, z
;           r19 <- Textnummer
;           r20 <- Zeichenzhler (wird bei jedem Zeichen dekrementiert)
;
lcdtxt: ldi     r17, 1                                  ;LCD in Datenmodus setzen
        lds     r16, lang                               ;Men-Sprache holen
        tst     r16                                     ;Sprache Englisch?
        brne    lcdt10                                  ;nein -> weiter
        ldi     zl, low(entext * 2)                     ;sonst
        ldi     zh, high(entext * 2)                    ;Zeiger auf englische Texte
        rjmp    lcdt20
;
lcdt10: ldi     zl, low(getext * 2)
        ldi     zh, high(getext * 2)                    ;Zeiger auf deutsche Texte
;
lcdt20: lpm     r16, z+                                 ;Zeichen holen
        cpi     r16, 0xfe                               ;Text-Kennung gefunden?
        brne    lcdt20                                  ;nein -> weiter suchen
        lpm     r16, z+                                 ;Zeichen holen
        cp      r16, r19                                ;Text gefunden?
        breq    lcdt30                                  ;ja -> ausgeben
        cpi     r16, 0xfe                               ;Endezeichen gefunden?
        brne    lcdt20                                  ;nein -> weiter suchen
        ret                                             ;sonst Ende
;
lcdt30: lpm     r18, z+                                 ;Zeichen holen
        cpi     r18, 0xff                               ;Endezeichen?
        brne    lcdt40                                  ;nein -> weiter
        ret                                             ;sonst Ende
lcdt40: rcall   lcdw40                                  ;Zeichen ausgeben
        dec     r20                                     ;Zeichenzhler - 1
        rjmp    lcdt30                                  ;nchstes Zeichen bearbeiten
;
; LCD-Ausgabe, Auffllen einer Zeile mit Leerzeichen
; Register: r16, r17, r18
;           r20 <- Anzahl der Zeichen bis Zeilenende
;
lcdlin: ldi     r17, 1                                  ;LCD in Datenmodus setzen
lcdl10: tst     r20                                     ;Zeichenzhler = 0?
        brne    lcdl20                                  ;nein -> weiter
        ret                                             ;sonst Ende
;
lcdl20: ldi     r18, ' '                                ;Leerzeichen-Code laden
        rcall   lcdw40                                  ;ausgeben
        dec     r20                                     ;Zeichenzhler - 1
        rjmp    lcdl10                                  ;Schleife
;
; LCD-Ausgabe eines 4-stelligen Dezimalwertes
; Register: r16, r17, r18, r19, r21, x
;           r0, r1 <- 16-Bit-Wert
;           r20    <- Zeichenzhler (bei jedem Zeichen dekrementiert)
;
lcddez: clr     r21                                     ;Tausender = 0 setzen
        movw    r18, r0                                 ;Wert kopieren
        ldi     r16, low(1000)
        ldi     r17, high(1000)                         ;Konstante 1000 laden
lcdd10: sub     r18, r16                                ;Wert - 1000
        sbc     r19, r17                                ;Ergebnis negativ?
        brmi    lcdd20                                  ;ja -> weiter
        movw    r0, r18                                 ;sonst Ergebnis kopieren
        inc     r21                                     ;Tausender zhlen
        rjmp    lcdd10                                  ;wiederholen
;
lcdd20: tst     r21                                     ;Tausender = 0?
        breq    lcdd30                                  ;ja -> 0 durch Leerzeichen ersetzen
        ori     r21, 0x30                               ;Tausender in ASCII wandeln
        rjmp    lcdd40                                  ;und speichern
lcdd30: ldi     r21, ' '                                ;Leerzeichen laden
lcdd40: sts     dezval, r21                             ;speichern
        clr     r21                                     ;Hunderter = 0 setzen
        movw    r18, r0                                 ;letzten Wert kopieren
        ldi     r16, 100                                ;Konstante 100 laden
        clr     r17                                     ;High-Byte = 0
lcdd50: sub     r18, r16                                ;Wert - 100
        sbc     r19, r17                                ;Ergebnis negativ?
        brmi    lcdd60                                  ;ja -> weiter
        movw    r0, r18                                 ;sonst Ergebnis kopieren
        inc     r21                                     ;Hunderter zhlen
        rjmp    lcdd50                                  ;wiederholen
;
lcdd60: ori     r21, 0x30                               ;Hunderter in ASCII wandeln
        sts     dezval + 1, r21                         ;und speichern
        clr     r21                                     ;Zehner = 0 setzen
        movw    r18, r0                                 ;letzten Wert kopieren
        ldi     r16, 10                                 ;Konstante 10 laden
        clr     r17                                     ;High-Byte = 0
lcdd70: sub     r18, r16                                ;Wert - 10
        sbc     r19, r17                                ;Ergebnis negativ?
        brmi    lcdd80                                  ;ja -> weiter
        movw    r0, r18                                 ;sonst Ergebnis kopieren
        inc     r21                                     ;Zehner zhlen
        rjmp    lcdd70                                  ;wiederholen
;
lcdd80: ori     r21, 0x30                               ;Zehner in ASCII wandeln
        sts     dezval + 2, r21                         ;und speichern
        mov     r21, r0                                 ;Restwert (Einer) holen
        ori     r21, 0x30                               ;Einer in ASCII wandeln
        sts     dezval + 3, r21                         ;und speichern
;
        ldi     xl, low(dezval)
        ldi     xh, high(dezval)                        ;Zeiger auf Dezimalwert
        ldi     r19, 4                                  ;4 Zeichen ausgeben
        ldi     r17, 1                                  ;LCD in Datenmodus setzen
lcdd90: ld      r18, x+                                 ;Zeichen holen
        rcall   lcdw40                                  ;ausgeben
        dec     r20                                     ;Zeichenzhler - 1
        dec     r19                                     ;alle Zeichen ausgegeben?
        brne    lcdd90                                  ;nein -> Schleife
        ret
;
#if     sm2                                             ;--- Sensormodul2 ---
;
; LCD-Ausgabe eines Zeichens im 4-Bit Mode (Sensormodul2)
; Register: r16
;           r17 <- 0 = Befehlsbyte, 1 = Datenbyte
;           r18 <- LCD-Daten oder -Befehl
;
lcdout: tst     r17                                     ;Datenbyte?
        brne    lcdo10                                  ;ja -> weiter
        cbi     portc, 4                                ;LCD in Befehlsmodus setzen
        rjmp    lcdo20
lcdo10: sbi     portc, 4                                ;LCD in Datenmodus setzen
lcdo20: mov     r16, r18                                ;Datenbyte kopieren
        ror     r18                                     ;oberes Nibble an die richtige
        ror     r18                                     ;Position schieben
        andi    r18, 0x3c                               ;Nibble filtern
        ori     r18, 0x03                               ;Pullup fr Al1 und Al2 setzen
        out     portb, r18                              ;Daten/Befehlsbyte ausgeben
        sbi     portc, 5                                ;E auf High setzen
        cbi     portc, 5                                ;E wieder auf Low setzen
;
        mov     r18, r16                                ;Datenbyte kopieren
        rol     r18                                     ;unteres Nibble an die richtige
        rol     r18                                     ;Position schieben
        andi    r18, 0x3c                               ;Nibble filtern
        ori     r18, 0x03                               ;Pullup fr Al1 und Al2 setzen
        out     portb, r18                              ;Daten/Befehlsbyte ausgeben
        sbi     portc, 5                                ;E auf High setzen
        cbi     portc, 5                                ;E wieder auf Low setzen
        ret
;
; LCD-Ausgabe eines Zeichens im 4-Bit Mode und Warteschleife 40s
; Register: r16
;           r17 <- 0 = Befehlsbyte, 1 = Datenbyte
;           r18 <- LCD-Daten/Befehlsbyte
;
lcdw40: rcall   lcdout                                  ;Zeichen ausgeben
        rjmp    w040us                                  ;Warteschleife 40s
;
; LCD-Ausgabe eines Zeichens im 8-Bit Mode (Sensormodul2)
; Register: r16
;           r17 <- 0=Befehlsbyte, 1=Datenbyte
;           r18 <- LCD-Daten oder -Befehl
;
lcd8bt: tst     r17                                     ;Datenbyte?
        brne    lcd810                                  ;ja -> weiter
        cbi     portc, 4                                ;LCD in Befehlsmodus setzen
        rjmp    lcd820
lcd810: sbi     portc, 4                                ;LCD in Datenmodus setzen
lcd820: ror     r18                                     ;oberes Nibble an die richtige
        ror     r18                                     ;Position schieben
        andi    r18, 0x3c                               ;Nibble filtern
        ori     r18, 0x03                               ;Pullup fr Al1 und Al2 setzen
        out     portb, r18                              ;Daten/Befehlsbyte ausgeben
        sbi     portc, 5                                ;E auf High setzen
        cbi     portc, 5                                ;E wieder auf Low setzen
        ret
;
#endif
;
#if     sm3                                             ;--- Sensormodul3 ---
;
; LCD-Ausgabe eines Zeichens im 4-Bit Mode (Sensormodul3)
; Register: r16
;           r17 <- 0 = Befehlsbyte, 1 = Datenbyte
;           r18 <- LCD-Daten oder -Befehl
;
lcdout: tst     r17                                     ;Datenbyte?
        brne    lcdo10                                  ;ja -> weiter
        cbi     portd, 0                                ;LCD in Befehlsmodus setzen
        rjmp    lcdo20
lcdo10: sbi     portd, 0                                ;LCD in Datenmodus setzen
lcdo20: mov     r16, r18                                ;Datenbyte kopieren
        ror     r18                                     ;oberes Nibble an die richtige
        ror     r18                                     ;Position schieben
        andi    r18, 0x3c                               ;Nibble filtern
        ori     r18, 0x03                               ;Pull-up fr Al1 und Al2 setzen
        out     portb, r18                              ;Daten/Befehlsbyte ausgeben
        sbi     portc, 5                                ;E auf High setzen
        cbi     portc, 5                                ;E wieder auf Low setzen
;
        mov     r18, r16                                ;Datenbyte kopieren
        rol     r18                                     ;unteres Nibble an die richtige
        rol     r18                                     ;Position schieben
        andi    r18, 0x3c                               ;Nibble filtern
        ori     r18, 0x03                               ;Pull-up fr Al1 und Al2 setzen
        out     portb, r18                              ;Daten/Befehlsbyte ausgeben
        sbi     portc, 5                                ;E auf High setzen
        cbi     portc, 5                                ;E wieder auf Low setzen
        ret
;
; LCD-Ausgabe eines Zeichens im 4-Bit Mode und Warteschleife 40s
; Register: r16
;           r17 <- 0 = Befehlsbyte, 1 = Datenbyte
;           r18 <- LCD-Daten/Befehlsbyte
;
lcdw40: rcall   lcdout                                  ;Zeichen ausgeben
        rjmp    w040us                                  ;Warteschleife 40s
;
; LCD-Ausgabe eines Zeichens im 8-Bit Mode (Sensormodul3)
; Register: r16
;           r17 <- 0 = Befehlsbyte, 1 = Datenbyte
;           r18 <- LCD-Daten oder -Befehl
;
lcd8bt: tst     r17                                     ;Datenbyte?
        brne    lcd810                                  ;ja -> weiter
        cbi     portd, 0                                ;LCD in Befehlsmodus setzen
        rjmp    lcd820
lcd810: sbi     portd, 0                                ;LCD in Datenmodus setzen
lcd820: ror     r18                                     ;oberes Nibble an die richtige
        ror     r18                                     ;Position schieben
        andi    r18, 0x3c                               ;Nibble filtern
        ori     r18, 0x03                               ;Pull-up fr Al1 und Al2 setzen
        out     portb, r18                              ;Daten/Befehlsbyte ausgeben
        sbi     portc, 5                                ;E auf High setzen
        cbi     portc, 5                                ;E wieder auf Low setzen
        ret
;
#endif
;
; --------------
; Warteschleifen
; --------------
;
; Warteschleife ca. 50ms
; Register: r15, r16
;
w050ms: clr     r15                                     ;Zeitkonstante uere Schleife
        clr     r16                                     ;Zeitkonstante innere Schleife
w050mx: dec     r16
        brne    w050mx                                  ;innere Schleife bearbeiten
        dec     r15
        brne    w050mx                                  ;uere Schleife bearbeiten
        ret
;
; Warteschleife 16ms
; Register: r15, r16
;
w016ms: ldi     r16, 84                                 ;Zeitkonstante uere Schleife
        mov     r15, r16                                ;kopieren
        ldi     r16, 26                                 ;Zeitkonstante innere Schleife
        rjmp    w050mx
;
; Warteschleife 10ms
; Register: r15, r16
;
w010ms: ldi     r16, 52                                 ;Zeitkonstante uere Schleife
        mov     r15, r16                                ;kopieren
        ldi     r16, 225                                ;Zeitkonstante innere Schleife
        rjmp    w050mx
;
; Warteschleife 4,1ms
; Register: r15, r16
;
w004ms: ldi     r16, 22                                 ;Zeitkonstante uere Schleife
        mov     r15, r16                                ;kopieren
        ldi     r16, 73                                 ;Zeitkonstante innere Schleife
        rjmp    w050mx
;
; Warteschleife 1ms
; Register: r15, r16
;
w001ms: ldi     r16, 6                                  ;Zeitkonstante uere Schleife
        mov     r15, r16                                ;kopieren
        ldi     r16, 45                                 ;Zeitkonstante innere Schleife
        rjmp    w050mx
;
; Warteschleife 480s
; Register: r15, r16
;
w480us: ldi     r16, 3                                  ;Zeitkonstante uere Schleife
        mov     r15, r16                                ;kopieren
        ldi     r16, 123                                ;Zeitkonstante innere Schleife
        rjmp    w050mx
;
; Warteschleife 420s
; Register: r15, r16
;
w420us: ldi     r16, 3                                  ;Zeitkonstante uere Schleife
        mov     r15, r16                                ;kopieren
        ldi     r16, 42                                 ;Zeitkonstante innere Schleife
        rjmp    w050mx
;
; Warteschleife 100s
; Register: r15, r16
;
w100us: ldi     r16, 1                                  ;Zeitkonstante uere Schleife
        mov     r15, r16                                ;kopieren
        ldi     r16, 129                                ;Zeitkonstante innere Schleife
        rjmp    w050mx
;
; Warteschleife 60s
; Register: r16
;
w060us: ldi     r16, 76                                 ;Zeitkonstante laden
w060xx: dec     r16
        brne    w060xx                                  ;Schleife bearbeiten
        ret
;
; Warteschleife 50s
; Register: r16
;
w050us: ldi     r16, 65                                 ;Zeitkonstante laden
w050xx: dec     r16
        brne    w050xx                                  ;Schleife bearbeiten
        ret
;
; Warteschleife 40s
; Register: r16
;
w040us: ldi     r16, 49                                 ;Zeitkonstante laden
w040xx: dec     r16
        brne    w040xx                                  ;Schleife bearbeiten
        ret
;
; Warteschleife 4,5s
; Register: r16
;
w005us: ldi     r16, 4                                  ;Zeitkonstante laden
w005xx: dec     r16
        brne    w005xx                                  ;Schleife bearbeiten
        ret
;
; Warteschleife 2s
; Register: keine
;
w002us: nop
        ret
;
; ---------------
; EEPROM-Routinen
; ---------------
;
; Lesen eines Bytes aus dem EEPROM
; Register: r16 <- EEPROM-Adresse Low
;           r17 <- EEPROM-Adresse High
;           r18 -> EEPROM-Daten
;
eeread: sbic    eecr, eewe                              ;luft ein Schreibzyklus?
        rjmp    eeread                                  ;ja -> warten
        out     eearl, r16                              ;EEPROM-Adresse Low setzen
        out     eearh, r17                              ;EEPROM-Adresse High setzen
        sbi     eecr, eere                              ;EEPROM lesen aktivieren
        in      r18, eedr                               ;Daten lesen
        ret
;
; Schreiben eines Bytes in das EEPROM
; Register: r16 <- EEPROM-Adresse Low
;           r17 <- EEPROM-Adresse High
;           r18 <- EEPROM-Daten
;
eewrit: sbic    eecr, eewe                              ;luft ein Schreibzyklus?
        rjmp    eewrit                                  ;ja -> warten
        out     eearl, r16                              ;EEPROM-Adresse Low setzen
        out     eearh, r17                              ;EEPROM-Adresse High setzen
        out     eedr, r18                               ;EEPROM-Datenbyte setzen
        sbi     eecr, eemwe                             ;Masterwrite-Bit setzen
        sbi     eecr, eewe                              ;Write-Bit setzen
        ret
;
; ----------------
; Diverse Routinen
; ----------------
;
; Anzahl der Temperatur- und Luftfeuchtigkeitssensoren ermitteln und speichern
; Register: r16, r17, r18, z
;
getsen: ldi     zl, low(tsentb)
        ldi     zh, high(tsentb)                        ;Zeiger auf Temperatursensortabelle
        ldi     r18, 24                                 ;24 Werte durchsuchen
        clr     r17                                     ;Sensorzhler = 0 setzen
gets10: ld      r16, z+                                 ;Wert aus Tabelle lesen
        cpi     r16, 8                                  ;gltiger Wert?
        brcc    gets20                                  ;nein -> weiter
        inc     r17                                     ;sonst Sensorzhler erhhen
gets20: dec     r18                                     ;alle Sensoren bearbeitet?
        brne    gets10                                  ;nein -> Schleife
        sts     temsen, r17                             ;Anzahl der Sensoren speichern
;
        ldi     zl, low(hsentb)
        ldi     zh, high(hsentb)                        ;Zeiger auf Luftfeuchtigkeitssensortabelle
        ldi     r18, 3                                  ;3 Werte durchsuchen
        clr     r17                                     ;Sensorzhler = 0 setzen
gets30: ld      r16, z+                                 ;Wert aus Tabelle lesen
        cpi     r16, 8                                  ;gltiger Wert?
        brcc    gets40                                  ;nein -> weiter
        inc     r17                                     ;sonst Sensorzhler erhhen
gets40: dec     r18                                     ;alle Sensoren bearbeitet?
        brne    gets30                                  ;nein -> Schleife
        sts     humsen, r17                             ;Anzahl der Sensoren speichern
        ret
;
; Sensorbelegung ermitteln und speichern
; Register: r16, r17, r18, r19, x, z
;
getmap: ldi     r19, 0                                  ;Busnummer = 0 setzen
        ldi     xl, low(senmap)
        ldi     xh, high(senmap)                        ;Zeiger auf Sensorbelegung
;
getm10: ldi     r17, '0'                                ;Sensorzhler = 0 setzen (ASCII)
        ldi     zl, low(tsentb)
        ldi     zh, high(tsentb)                        ;Zeiger auf Temperatursensortabelle
        ldi     r18, 24                                 ;24 Werte durchsuchen
getm20: ld      r16, z+                                 ;Wert aus Tabelle lesen
        cp      r16, r19                                ;Sensor am aktuellen Bus?
        brne    getm30                                  ;nein -> weiter
        inc     r17                                     ;sonst Sensorzhler erhhen
getm30: dec     r18                                     ;alle Tabellenpltze gelesen?
        brne    getm20                                  ;nein -> Schleife
;
        ldi     zl, low(hsentb)
        ldi     zh, high(hsentb)                        ;Zeiger auf Luftfeuchtigkeitssensortabelle
        ldi     r18, 3                                  ;3 Werte durchsuchen
getm40: ld      r16, z+                                 ;Wert aus Tabelle lesen
        cp      r16, r19                                ;Sensor am aktuellen Bus?
        brne    getm50                                  ;nein -> weiter
        inc     r17                                     ;sonst Sensorzhler erhhen
getm50: dec     r18                                     ;alle Tabellenpltze gelesen?
        brne    getm40                                  ;nein -> Schleife
;
        cpi     r17, '9' + 1                            ;mehr als 9 Sensoren am Bus?
        brcs    getm60                                  ;nein -> weiter
        subi    r17, -7                                 ;sonst Zahl in Buchstaben wandeln, 10 = A, 11 = B, 12 = C usw.
getm60: st      x+, r17                                 ;Sensoranzahl speichern
        inc     r19                                     ;nchsten Bus testen
        cpi     r19, 8                                  ;alle Busse getestet?
        brne    getm10                                  ;nein -> Schleife
        ret
;
; Temperatursensordaten im EEPROM speichern
; Register: r16, r17, r18, r19, r20, x
;
savsen: ldi     r19, 24 * 8                             ;Daten von 24 Sensoren schreiben, jeweils 8 Bytes
        ldi     r16, low(etseid)
        ldi     r17, high(etseid)                       ;EEPROM-Adresse der Temperatursensor-IDs
        ldi     xl, low(tsenid)
        ldi     xh, high(tsenid)                        ;Zeiger auf Temperatursensor-IDs im RAM
satid1: ld      r18, x+                                 ;Byte aus RAM holen
        rcall   eewrit                                  ;und im EEPROM speichern
        inc     r16                                     ;nchste EEPROM-Adresse Low
        brne    satid2                                  ;bertrag? nein -> weiter
        inc     r17                                     ;nchste EEPROM-Adresse High
satid2: dec     r19                                     ;alle Sensoren gelesen?
        brne    satid1                                  ;nein -> Schleife
;
        ldi     r19, 24                                 ;Daten von 24 Sensoren schreiben
        ldi     r16, low(etsetb)
        ldi     r17, high(etsetb)                       ;EEPROM-Adresse der Temperatursensor-Tabelle
        ldi     xl, low(tsentb)
        ldi     xh, high(tsentb)                        ;Zeiger auf Sensor-Tabelle im RAM
sattb1: ld      r18, x+                                 ;Byte aus RAM holen
        rcall   eewrit                                  ;und im EEPROM speichern
        inc     r16                                     ;nchste EEPROM-Adresse Low
        brne    sattb2                                  ;bertrag? nein -> weiter
        inc     r17                                     ;nchste EEPROM-Adresse High
sattb2: dec     r19                                     ;alle Sensoren gelesen?
        brne    sattb1                                  ;nein -> Schleife
;
; Luftfeuchtigkeitssensordaten im EEPROM speichern
;
        ldi     r19, 3 * 8                              ;Daten von 3 Sensoren schreiben, jeweils 8 Bytes
        ldi     r16, low(ehseid)
        ldi     r17, high(ehseid)                       ;EEPROM-Adresse der Luftfeuchtigkeitssensor-IDs
        ldi     xl, low(hsenid)
        ldi     xh, high(hsenid)                        ;Zeiger auf Luftfeuchtigkeitssensor-IDs im RAM
sahid1: ld      r18, x+                                 ;Byte aus RAM holen
        rcall   eewrit                                  ;und im EEPROM speichern
        inc     r16                                     ;nchste EEPROM-Adresse Low
        brne    sahid2                                  ;bertrag? nein -> weiter
        inc     r17                                     ;nchste EEPROM-Adresse High
sahid2: dec     r19                                     ;alle Sensoren gelesen?
        brne    sahid1                                  ;nein -> Schleife
;
        ldi     r19, 3                                  ;Daten von 3 Sensoren schreiben
        ldi     r16, low(ehsetb)
        ldi     r17, high(ehsetb)                       ;EEPROM-Adresse der Luftfeuchtigkeitssensor-Tabelle
        ldi     xl, low(hsentb)
        ldi     xh, high(hsentb)                        ;Zeiger auf Sensor-Tabelle im RAM
sahtb1: ld      r18, x+                                 ;Byte aus RAM holen
        rcall   eewrit                                  ;und im EEPROM speichern
        inc     r16                                     ;nchste EEPROM-Adresse Low
        brne    sahtb2                                  ;bertrag? nein -> weiter
        inc     r17                                     ;nchste EEPROM-Adresse High
sahtb2: dec     r19                                     ;alle Sensoren gelesen?
        brne    sahtb1                                  ;nein -> Schleife
;
; Anzahl der Luftdrucksensoren speichern
;
        ldi     r16, low(eprsen)
        ldi     r17, high(eprsen)                       ;EEPROM-Adresse Anzahl der Luftdrucksensoren
        lds     r18, prssen                             ;Anzahl der Luftdrucksensoren holen
        rcall   eewrit                                  ;im EEPROM speichern
        ret
;
; 16-Bit Interrupt-Zhler auf Anzeige-Timeout setzen (60s)
; Register: r16,r17
;
setinc: ldi     r16, low(setrtn)
        ldi     r17, high(setrtn)
        sts     intc16, r16                             ;Interrupt-Zhler auf Rckkehr-Zeit in den
        sts     intc16 + 1, r17                         ;Automatik-Modus setzen
        ret
;
; 16-Bit-Wert in Dezimal-ASCII umwandeln
; Register: x                  <- 16-Bit-Wert
;           r19, r18, r17, r16 -> Dezimalwert-ASCII
;
dezasc: clr     r16                                     ;0-Wert fr 16-Bit Subtraktion
        clr     r18                                     ;ASCII Stelle 2 lschen
        clr     r19                                     ;ASCII Stelle 1 lschen
deza10: mov     r17, xl                                 ;Binrzahl zwischenspeichern L
        subi    xl, 100                                 ;100 subtrahieren
        sbc     xh, r16                                 ;bertrag subtrahieren
        brcs    deza20                                  ;war Zahl>99? nein -> weiter
        inc     r18                                     ;ASCII Stelle 2 erhhen
        cpi     r18, 10                                 ;bertrag?
        brcs    deza10                                  ;nein -> Schleife
        clr     r18                                     ;sonst ASCII Stelle 2 lschen
        inc     r19                                     ;und ASCII Stelle 1 erhhen
        rjmp    deza10                                  ;Schleife
;
deza20: mov     xl, r17                                 ;letzten Binrwert L holen
        clr     r17                                     ;ASCII Stelle 3 lschen
deza30: cpi     xl, 10                                  ;Wert>9?
        brcs    deza40                                  ;nein -> Rest auswerten
        inc     r17                                     ;sonst ASCII Stelle 3 erhhen
        subi    xl, 10                                  ;Wert um 10 vermindern
        rjmp    deza30                                  ;Schleife
deza40: mov     r16, xl                                 ;Rest in ASCII Stelle 4 legen
        ori     r16, 0x30                               ;Ergebnis in ASCII wandeln
        ori     r17, 0x30
        ori     r18, 0x30
        ori     r19, 0x30
        ret
;
; Multiplikation ohne Vorzeichen 16 Bit x 16 Bit = 32 Bit aus "Application Note AVR201" von Atmel
; Register: r0, r1, r2
;           r20, r21           <- Faktor 1
;           r4, r5             <- Faktor 2
;           r19, r18, r17, r16 -> Produkt
;
umul16: clr     r2
        mul     r5, r21                                 ;Faktor 1 H * Faktor 2 H
        movw    r18, r0
        mul     r4, r20                                 ;Faktor 1 L * Faktor 2 L
        movw    r16, r0
        mul     r5, r20                                 ;Faktor 1 H * Faktor 2 L
        add     r17, r0
        adc     r18, r1
        adc     r19, r2
        mul     r21, r4                                 ;Faktor 2 H * Faktor 1 L
        add     r17, r0
        adc     r18, r1
        adc     r19, r2
        ret
;
; Division ohne Vorzeichen 24 Bit / 16 Bit = 24 Bit von Peter Dannegger (www.avrfreaks.net)
; Register: r18, r17, r16  <- Divident
;           r5, r4         <- Divisor
;           r21, r20, r19  -> Quotient
;
udiv24: clr     r19
        clr     r20
        ldi     r21, 24
udiv10: lsl     r16
        rol     r17
        rol     r18
        rol     r19
        rol     r20
        brcs    udiv20
        cp      r19, r4
        cpc     r20, r5
        brcs    udiv30
udiv20: sub     r19, r4
        sbc     r20, r5
        inc     r16
udiv30: dec     r21
        brne    udiv10
        mov     r4, r19
        mov     r5, r20
        ret
;
; ------------------
; Interrupt-Routinen
; ------------------
;
; Timer1 Interrupt, wird alle 50ms aufgerufen
; Register: isreg, itemp1, itemp2, y
;
; Abfrage der Alarmeingnge
;
timer1: in      isreg, sreg                             ;SREG sichern
        ldi     yl, low(alstat)
        ldi     yh, high(alstat)                        ;Zeiger auf Alarmstatus
        ld      itemp2, y                               ;Alarmstatus holen
        in      itemp1, pinb                            ;Alarmeingnge 1 und 2 lesen
        com     itemp1                                  ;und invertieren
        bst     itemp1, 0                               ;Alarmbit 1 holen
        bld     itemp2, 0                               ;und in Status einsetzen
        st      y+, itemp2                              ;Alarmstatus speichern
        ld      itemp2, y                               ;nchsten Alarmstatus holen
        bst     itemp1, 1                               ;Alarmbit 2 holen
        bld     itemp2, 0                               ;und in Status einsetzen
        st      y+, itemp2                              ;Alarmstatus speichern
        ld      itemp2, y                               ;nchsten Alarmstatus holen
        in      itemp1, pind                            ;Alarmeingnge 3 und 4 lesen
        com     itemp1                                  ;und invertieren
        bst     itemp1, 2                               ;Alarmbit 3 holen
        bld     itemp2, 0                               ;und in Status einsetzen
        st      y+, itemp2                              ;Alarmstatus speichern
        ld      itemp2, y                               ;nchsten Alarmstatus holen
        bst     itemp1, 3                               ;Alarmbit 2 holen
        bld     itemp2, 0                               ;und in Status einsetzen
        st      y+, itemp2                              ;Alarmstatus speichern
;
; Steuerung der Alarme (Flags, Alarmzhler)
;
        ldi     yl, low(alstat)
        ldi     yh, high(alstat)                        ;Zeiger auf Alarmstatus
        lds     itemp2, altime                          ;Alarm-Zeit laden
        mov     itemp1, itemp2                          ;und vorbereiten
tim100: ld      itemp2, y                               ;Alarmstatus holen
        andi    itemp2, 0b00000011                      ;nur Statusbits 0/1 bewerten
        cpi     itemp2, 0b00000000                      ;Status gleich geblieben?
        breq    tim130                                  ;ja -> berspringen
        cpi     itemp2, 0b00000011                      ;Status gleich geblieben?
        breq    tim130                                  ;ja -> berspringen
        ld      itemp2, y                               ;Alarmstatus nochmals holen
        bst     itemp2, 2                               ;Bearbeitungs-Bit gesetzt?
        brts    tim110                                  ;ja -> weiter
        sbr     itemp2, 0b00000100                      ;sonst Bearbeitungs-Flag setzen
        st      y, itemp2                               ;Alarmstatus wieder speichern
        std     y+4, itemp1                             ;Alarmzhler neu laden
tim110: ldd     itemp2, y + 4                           ;Alarmzhler holen
        cpi     itemp2, 0                               ;Zhler abgelaufen?
        brne    tim120                                  ;nein -> berspringen
        ld      itemp2, y                               ;Alarmstatus nochmals holen
        bst     itemp2, 0                               ;aktuelles Statusbit holen
        bld     itemp2, 1                               ;und als neuen Status setzen
        cbr     itemp2, 0b00000100                      ;Bearbeitungs-Flag lschen
        sbr     itemp2, 0b00001000                      ;Melde-Flag setzen
        bld     itemp2, 4                               ;Melde-Status setzen
        st      y, itemp2                               ;neuen Alarmstatus speichern
        rjmp    tim140
tim120: dec     itemp2                                  ;sonst Zhler vermindern
        std     y + 4, itemp2                           ;und wieder speichern
        rjmp    tim140
tim130: ld      itemp2, y                               ;Alarmstatus nochmals holen
        cbr     itemp2, 0b00000100                      ;Bearbeitungs-Flag lschen
        st      y, itemp2                               ;neuen Alarmstatus speichern
tim140: adiw    yl, 1                                   ;Zeiger erhhen
        cpi     yl, low(alstat + 4)                     ;alle Werte bearbeitet?
        brne    tim100                                  ;nein -> Schleife
;
; Steuerung des Tasters, kurze und lange Bettigung (unterschiedlich bei Sensormodul2 und Sensormodul3)
;
        bst     flags, 3                                ;letzten Taster-Status holen
        bld     itemp2, 0                               ;in Arbeitsregister legen
;
#if     sm2                                             ;--- Sensormodul2 ---
;
        in      itemp1, pind                            ;Tastereingang beim Sensormodul2 direkt lesen
#endif
;
#if     sm3                                             ;--- Sensormodul3 ---
;
        sbi     portd, 0                                ;PortD0 auf High setzen (Pull-up vorbereiten)
        cbi     ddrd, 0                                 ;PortD0 auf Eingang umschalten
        nop
        nop
        nop
        nop                                             ;kurze Wartezeit (1s)
        in      itemp1, pind                            ;Tastereingang lesen
        sbi     ddrd, 0                                 ;PortD0 wieder auf Ausgang umschalten
#endif
;
        com     itemp1                                  ;Taster-Bit invertieren
        bst     itemp1, 0                               ;Taster-Bit kopieren und
        bld     flags, 3                                ;als neuen Status speichern
        eor     itemp2, itemp1                          ;Statusnderung?
        andi    itemp2, 0x01                            ;Ergebnis filtern
        brne    tim300                                  ;Statusnderung? ja -> weiter
        bld     flags, 4                                ;sonst Entprellt-Status setzen
        lds     itemp2, keycnt                          ;Taster-Timeout-Zhler holen
        brts    tim200                                  ;Gedrckt-Status? ja -> weiter
        cbr     flags, 0b11110000                       ;sonst alle Taster-Flags lschen
        ldi     itemp2, keytio                          ;und Timeout-Zhler neu setzen
        rjmp    tim220
tim200: tst     itemp2                                  ;Timeout-Zhler=0?
        brne    tim210                                  ;nein -> weiter, sonst
        sbr     flags, 0b01000000                       ;Flag: Taster lang gedrckt
        rjmp    tim220
tim210: dec     itemp2                                  ;Timeout-Zhler vermindern
tim220: sts     keycnt, itemp2                          ;neuen Zhlerstand speichern
;
; 16-Bit Interrupt-Zhler vermindern und Interrupt-Flag setzen
;
tim300: lds     itemp2, intc16                          ;Interrupt-Zhler L holen
        lds     itemp1, intc16 + 1                      ;Interrupt-Zhler H holen
        or      itemp1, itemp2                          ;verknpfen, Zhlerstand=0?
        breq    tim500                                  ;ja -> Ende
;
        subi    itemp2, 1                               ;Interrupt-Zhler L-1
        sts     intc16, itemp2                          ;Ergebnis L speichern
        lds     itemp2, intc16 + 1                      ;Interrupt-Zhler H holen
        sbci    itemp2, 0                               ;bertrag subtrahieren
        sts     intc16 + 1, itemp2                      ;Ergebnis H speichern
;
tim500: sbr     flags, 0b00000001                       ;Interrupt-Flag setzen
        out     sreg, isreg                             ;SREG wieder herstellen
        reti
;
; ------------------
; Tabellen und Texte
; ------------------
;
; LCD-Adress-Tabelle
;
lcdtab: .db     0x80, 0xc0, 0x94, 0xd4                          ;LCD-Adressen fr Werte 1-4 (Zeilenanfang 1-4)
        .db     0x8a, 0xca, 0x9e, 0xde                          ;LCD-Adressen fr Werte 5-8 (Zeilenmitte 1-4)
;
; LCD Sonderzeichen
;
xchars: .db     2, 6, 0x0e, 0x1e, 0x0e, 6, 2, 0                 ;Pfeil nach links
        .db     0x10, 0x18, 0x1c, 0x1e, 0x1c, 0x18, 0x10, 0     ;Pfeil nach rechts
        .db     0x1f, 0x11, 0x11, 0x11, 0x11, 0x11, 0x1f, 0     ;Alarmkontakt offen
        .db     0x1f, 0x1f, 0x1f, 0x1f, 0x1f, 0x1f, 0x1f, 0     ;Alarmkontakt geschlossen
        .db     0, 0, 1, 3, 7, 0x0f, 0x1f, 0                    ;Dreieck-Symbol
;
; Deutsche Texte
;
getext: .db     0xfe, 0                                         ;Beginn von Text 0 (Modulname und Versionsnummer)
;
#if     sm2
        .db     "Sensormodul2"                                  ;Text fr Sensormodul2
#endif
#if     sm3
        .db     "Sensormodul3"                                  ;Text fr Sensormodul3
#endif
;
        .db     version, 0xff, 0xff                             ;Versionsnummer
        .db     0xfe, 1, "Sensoren: ", 0xff, 0xff
        .db     0xfe, 2, "Belegung: ", 0xff, 0xff
        .db     0xfe, 3, "Modus   : ", 0xff, 0xff
        .db     0xfe, 4, "Direkt", 0xff, 0xff
        .db     0xfe, 5, "Speicher", 0xff, 0xff
        .db     0xfe, 6, 1, " ", 0xff, 0xff                     ;Pfeil nach rechts
        .db     0xfe, 7, " ", 0, 0xff, 0xff                     ;Pfeil nach links
        .db     0xfe, 8, "Lang ~ ", 0xe1, "ndern", 0xff
        .db     0xfe, 9, "Kurz ~ weiter", 0xff
        .db     0xfe, 10, "Kurz ~ Ende", 0xff
        .db     0xfe, 11, "Modus: ", 0xff
        .db     0xfe, 12, "Sende-Intervall:", 0xff, 0xff
        .db     0xfe, 13, "Alarm-Zeit:", 0xff
        .db     0xfe, 14, "Sprache/Language:", 0xff
        .db     0xfe, 15, "Deutsch", 0xff
        .db     0xfe, 16, " ms", 0xff                           ;Millisekunden
        .db     0xfe, 17, "Sensor-Reset", 0xff, 0xff
        .db     0xfe, 20, "Alarm-Eingang 1 ", 0xff, 0xff
        .db     0xfe, 21, "Alarm-Eingang 2 ", 0xff, 0xff
        .db     0xfe, 22, "Alarm-Eingang 3 ", 0xff, 0xff
        .db     0xfe, 23, "Alarm-Eingang 4 ", 0xff, 0xff
        .db     0xfe, 24, "Luftdruck ", 0xff, 0xff
        .db     0xfe, 25, " hPa", 0xff, 0xff                    ;Hektopascal
        .db     0xfe, 26, "Update", 0xff, 0xff
;
        .db     0xfe, 0xfe                                      ;Ende des Textbereiches
;
; English texts
;
entext: .db     0xfe, 0                                         ;Start of text 0 (module name and version number)
;
#if     sm2
        .db     "Sensor module2"                                ;Text for sensor module2
#endif
#if     sm3
        .db     "Sensor module3"                                ;Text for sensor module3
#endif
;
        .db     version, 0xff, 0xff                             ;Version number
        .db     0xfe, 1, "Sensors: ", 0xff
        .db     0xfe, 2, "Map    : ", 0xff
        .db     0xfe, 3, "Mode   : ", 0xff
        .db     0xfe, 4, "Direct", 0xff, 0xff
        .db     0xfe, 5, "Memory", 0xff, 0xff
        .db     0xfe, 6, 1, " ", 0xff, 0xff                     ;Right arrow
        .db     0xfe, 7, " ", 0, 0xff, 0xff                     ;Left arrow
        .db     0xfe, 8, "Long  ~ change", 0xff, 0xff
        .db     0xfe, 9, "Short ~ next", 0xff, 0xff
        .db     0xfe, 10, "Short ~ end", 0xff
        .db     0xfe, 11, "Mode: ", 0xff, 0xff
        .db     0xfe, 12, "Transmit interval:", 0xff, 0xff
        .db     0xfe, 13, "Alarm time:", 0xff
        .db     0xfe, 14, "Language/Sprache:", 0xff
        .db     0xfe, 15, "English", 0xff
        .db     0xfe, 16, " ms", 0xff                           ;Milliseconds
        .db     0xfe, 17, "Sensor reset", 0xff, 0xff
        .db     0xfe, 20, "Alarm input 1   ", 0xff, 0xff
        .db     0xfe, 21, "Alarm input 2   ", 0xff, 0xff
        .db     0xfe, 22, "Alarm input 3   ", 0xff, 0xff
        .db     0xfe, 23, "Alarm input 4   ", 0xff, 0xff
        .db     0xfe, 24, "Air press.", 0xff, 0xff
        .db     0xfe, 25, " hPa", 0xff, 0xff                    ;Hectopascal
        .db     0xfe, 26, "Update", 0xff, 0xff
;
        .db     0xfe, 0xfe                                      ;End of text area
;
; Sensortexte: Diese knnen gendert werden. Jeder Name muss 4 Zeichen lang sein. Umlaute mssen entsprechend des LCD-Zeichensatzes kodiert werden (im Zweifelsfall bitte im Datenblatt nachschauen):
; = 0xe1, = 0xef, = 0xf5, = 0xe2
; Sensor texts: These can be modified. Each name must have a length of 4 characters:
;
temtxt: .db     "Tem1", "Tem2", "Tem3", "Tem4"                  ;Temperatursensornamen, Name of temperature sensor 1-4
        .db     "Tem5", "Tem6", "Tem7", "Tem8"                  ;Temperatursensornamen, Name of temperature sensor 5-8
        .db     "TemA", "TemB", "TemC", "TemD"                  ;Temperatursensornamen, Name of temperature sensor a-d (9-12)
        .db     "TemE", "TemF", "TemG", "TemH"                  ;Temperatursensornamen, Name of temperature sensor e-h (13-16)
        .db     "TemI", "TemJ", "TemK", "TemL"                  ;Temperatursensornamen, Name of temperature sensor i-l (17-20)
        .db     "TemM", "TemN", "TemO", "TemP"                  ;Temperatursensornamen, Name of temperature sensor m-p (21-24)
;
; Weitere Sensortexte / Further sensor texts (4 + 2 + 4 Zeichen / characters):
;
humtxt: .db     "TemQ", "HX", " %RH"                            ;Luftfeuchtigkeitssensor 1, Name of humidity sensor 1 (26/29)
        .db     "TemR", "HY", " %RH"                            ;Luftfeuchtigkeitssensor 2, Name of humidity sensor 2 (27/30)
        .db     "TemS", "HZ", " %RH"                            ;Luftfeuchtigkeitssensor 3, Name of humidity sensor 3 (28/31)
;
;
; ------------
; EEPROM-Daten
; ------------
;
.eseg
;
dummy:  .byte   1                                               ;Byte 0 wird nicht genutzt
emodfl: .byte   1                                               ;Modus-Flag, siehe Variable "modflg"
eintva: .byte   1                                               ;Sende-Intervall, siehe Variable "intval"
ealtim: .byte   1                                               ;Alarm-Zeit, siehe Variable "altime"
elang:  .byte   1                                               ;Men-Sprache, siehe Variable "lang"
etseid: .byte   24 * 8                                          ;8-Byte-ROM-IDs, siehe Variable "tsenid"
etsetb: .byte   24                                              ;Sensor-Tabelle, siehe Variable "tsentb"
ehseid: .byte   3 * 8                                           ;8-Byte-ROM-IDs, siehe Variable "hsenid"
ehsetb: .byte   3                                               ;Sensor-Tabelle, siehe Variable "hsentb"
eprsen: .byte   1                                               ;Anzahl der Luftdrucksensoren, siehe Variable "prssen"
;
